@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * Copyright © 2012, CohortFS, LLC.
  * Author: Adam C. Emerson <aemerson@linuxbox.com>
@@ -36,6 +37,7 @@
 #include <sys/sysmacros.h> /* for makedev(3) */
 #endif
 #include <fcntl.h>
+#include <sys/xattr.h>
 #include <cephfs/libcephfs.h>
 #include "fsal.h"
 #include "fsal_types.h"
@@ -47,12 +49,23 @@
 #include "statx_compat.h"
 #include "nfs_core.h"
 #include "linux/falloc.h"
+#ifdef USE_LTTNG
+#include "gsh_lttng/fsal_ceph.h"
+#endif
 
 typedef enum {
 	CEPH_AS_CALLER,
 	CEPH_AS_ROOT,
 } ceph_access_t;
 
+/*
+ * If the inode has a mode that doesn't allow writes, then the follow-on
+ * setxattr for setting the security context can fail. We set this flag
+ * in op_ctx->fsal_private after a create to indicate that the setxattr
+ * should be done as root.
+ */
+#define CEPH_SETXATTR_AS_ROOT	((void *)(-1UL))
+
 /**
  * @brief Release an object
  *
@@ -89,7 +102,7 @@ static void ceph_fsal_release(struct fsal_obj_handle *obj_pub)
 static fsal_status_t ceph_fsal_lookup(struct fsal_obj_handle *dir_pub,
 				      const char *path,
 				      struct fsal_obj_handle **obj_pub,
-				      struct attrlist *attrs_out)
+				      struct fsal_attrlist *attrs_out)
 {
 	/* Generic status return */
 	int rc = 0;
@@ -105,8 +118,13 @@ static fsal_status_t ceph_fsal_lookup(struct fsal_obj_handle *dir_pub,
 
 	LogFullDebug(COMPONENT_FSAL, "Lookup %s", path);
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_lookup, __func__, __LINE__,
+		   path, NULL, 0);
+#endif
+
 	rc = fsal_ceph_ll_lookup(export->cmount, dir->i, path, &i, &stx,
-					!!attrs_out, op_ctx->creds);
+					!!attrs_out, &op_ctx->creds);
 	if (rc < 0)
 		return ceph2fsal_error(rc);
 
@@ -117,11 +135,16 @@ static fsal_status_t ceph_fsal_lookup(struct fsal_obj_handle *dir_pub,
 
 	*obj_pub = &obj->handle;
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_lookup, __func__, __LINE__,
+		   path, &obj->handle, stx.stx_ino);
+#endif
+
 	return fsalstat(0, 0);
 }
 
 static int
-ceph_fsal_get_sec_label(struct ceph_handle *handle, struct attrlist *attrs)
+ceph_fsal_get_sec_label(struct ceph_handle *handle, struct fsal_attrlist *attrs)
 {
 	int rc = 0;
 	struct ceph_export *export =
@@ -200,9 +223,12 @@ static fsal_status_t ceph_fsal_readdir(struct fsal_obj_handle *dir_pub,
 	unsigned int want = attrmask2ceph_want(attrmask);
 	/* Return status */
 	fsal_status_t fsal_status = { ERR_FSAL_NO_ERROR, 0 };
+	/* local rfiles in target dir */
+	uint64_t rfiles = 0;
+
 
 	rc = fsal_ceph_ll_opendir(export->cmount, dir->i, &dir_desc,
-				  op_ctx->creds);
+				  &op_ctx->creds);
 	if (rc < 0)
 		return ceph2fsal_error(rc);
 
@@ -218,13 +244,13 @@ static fsal_status_t ceph_fsal_readdir(struct fsal_obj_handle *dir_pub,
 
 		rc = fsal_ceph_readdirplus(export->cmount, dir_desc, dir->i,
 					   &de, &stx, want, 0, &i,
-					   op_ctx->creds);
+					   &op_ctx->creds);
 		if (rc < 0) {
 			fsal_status = ceph2fsal_error(rc);
 			goto closedir;
 		} else if (rc == 1) {
 			struct ceph_handle *obj;
-			struct attrlist attrs;
+			struct fsal_attrlist attrs;
 			enum fsal_dir_result cb_rc;
 
 			/* skip . and .. */
@@ -261,6 +287,7 @@ static fsal_status_t ceph_fsal_readdir(struct fsal_obj_handle *dir_pub,
 			/* Read ahead not supported by this FSAL. */
 			if (cb_rc >= DIR_READAHEAD)
 				goto closedir;
+			rfiles += 1;
 
 		} else if (rc == 0) {
 			*eof = true;
@@ -270,6 +297,11 @@ static fsal_status_t ceph_fsal_readdir(struct fsal_obj_handle *dir_pub,
 		}
 	}
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_readdir, __func__, __LINE__,
+		   dir_pub->fileid, rfiles);
+#endif
+
  closedir:
 
 	rc = ceph_ll_releasedir(export->cmount, dir_desc);
@@ -300,9 +332,9 @@ static fsal_status_t ceph_fsal_readdir(struct fsal_obj_handle *dir_pub,
  */
 
 static fsal_status_t ceph_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
-				const char *name, struct attrlist *attrib,
+				const char *name, struct fsal_attrlist *attrib,
 				struct fsal_obj_handle **new_obj,
-				struct attrlist *attrs_out)
+				struct fsal_attrlist *attrs_out)
 {
 	/* Generic status return */
 	int rc = 0;
@@ -322,14 +354,14 @@ static fsal_status_t ceph_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 
 	LogFullDebug(COMPONENT_FSAL,
 		     "mode = %o uid=%d gid=%d",
-		     attrib->mode, (int) op_ctx->creds->caller_uid,
-		     (int) op_ctx->creds->caller_gid);
+		     attrib->mode, (int) op_ctx->creds.caller_uid,
+		     (int) op_ctx->creds.caller_gid);
 
 	unix_mode = fsal2unix_mode(attrib->mode)
 		& ~op_ctx->fsal_export->exp_ops.fs_umask(op_ctx->fsal_export);
 
 	rc = fsal_ceph_ll_mkdir(export->cmount, dir->i, name, unix_mode, &i,
-			&stx, !!attrs_out, op_ctx->creds);
+			&stx, !!attrs_out, &op_ctx->creds);
 	if (rc < 0)
 		return ceph2fsal_error(rc);
 
@@ -343,9 +375,16 @@ static fsal_status_t ceph_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 	if (attrib->valid_mask) {
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
+		 *
+		 * Must use root creds to override some permissions checks
+		 * when the mode is not writeable (e.g. when setxattr'ing
+		 * security labels).
 		 */
+		op_ctx->fsal_private = CEPH_SETXATTR_AS_ROOT;
 		status = (*new_obj)->obj_ops->setattr2(*new_obj, false, NULL,
 						      attrib);
+		op_ctx->fsal_private = NULL;
+
 		if (FSAL_IS_ERROR(status)) {
 			/* Release the handle we just allocated. */
 			LogFullDebug(COMPONENT_FSAL,
@@ -376,6 +415,11 @@ static fsal_status_t ceph_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 
 	FSAL_SET_MASK(attrib->valid_mask, ATTR_MODE);
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_mkdir, __func__, __LINE__,
+		   name, &obj->handle, stx.stx_ino);
+#endif
+
 	return status;
 }
 
@@ -403,9 +447,9 @@ static fsal_status_t ceph_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 static fsal_status_t ceph_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 				      const char *name,
 				      object_file_type_t nodetype,
-				      struct attrlist *attrib,
+				      struct fsal_attrlist *attrib,
 				      struct fsal_obj_handle **new_obj,
-				      struct attrlist *attrs_out)
+				      struct fsal_attrlist *attrs_out)
 {
 #ifdef USE_FSAL_CEPH_MKNOD
 	/* Generic status return */
@@ -451,7 +495,7 @@ static fsal_status_t ceph_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 	}
 
 	rc = fsal_ceph_ll_mknod(export->cmount, dir->i, name, unix_mode,
-			unix_dev, &i, &stx, !!attrs_out, op_ctx->creds);
+			unix_dev, &i, &stx, !!attrs_out, &op_ctx->creds);
 	if (rc < 0)
 		return ceph2fsal_error(rc);
 
@@ -466,8 +510,10 @@ static fsal_status_t ceph_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
+		op_ctx->fsal_private = CEPH_SETXATTR_AS_ROOT;
 		status = (*new_obj)->obj_ops->setattr2(*new_obj, false, NULL,
 						      attrib);
+		op_ctx->fsal_private = NULL;
 		if (FSAL_IS_ERROR(status)) {
 			/* Release the handle we just allocated. */
 			LogFullDebug(COMPONENT_FSAL,
@@ -490,6 +536,11 @@ static fsal_status_t ceph_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 
 	FSAL_SET_MASK(attrib->valid_mask, ATTR_MODE);
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_mknod, __func__, __LINE__,
+		   name, nodetype, &obj->handle, stx.stx_ino);
+#endif
+
 	return status;
 #else
 	return fsalstat(ERR_FSAL_NOTSUPP, ENOTSUP);
@@ -518,9 +569,9 @@ static fsal_status_t ceph_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 
 static fsal_status_t ceph_fsal_symlink(struct fsal_obj_handle *dir_hdl,
 				  const char *name, const char *link_path,
-				  struct attrlist *attrib,
+				  struct fsal_attrlist *attrib,
 				  struct fsal_obj_handle **new_obj,
-				  struct attrlist *attrs_out)
+				  struct fsal_attrlist *attrs_out)
 {
 	/* Generic status return */
 	int rc = 0;
@@ -538,7 +589,7 @@ static fsal_status_t ceph_fsal_symlink(struct fsal_obj_handle *dir_hdl,
 	fsal_status_t status;
 
 	rc = fsal_ceph_ll_symlink(export->cmount, dir->i, name, link_path,
-			      &i, &stx, !!attrs_out, op_ctx->creds);
+			      &i, &stx, !!attrs_out, &op_ctx->creds);
 	if (rc < 0)
 		return ceph2fsal_error(rc);
 
@@ -553,8 +604,10 @@ static fsal_status_t ceph_fsal_symlink(struct fsal_obj_handle *dir_hdl,
 		/* Now per support_ex API, if there are any other attributes
 		 * set, go ahead and get them set now.
 		 */
+		op_ctx->fsal_private = CEPH_SETXATTR_AS_ROOT;
 		status = (*new_obj)->obj_ops->setattr2(*new_obj, false, NULL,
 						      attrib);
+		op_ctx->fsal_private = NULL;
 		if (FSAL_IS_ERROR(status)) {
 			/* Release the handle we just allocated. */
 			LogFullDebug(COMPONENT_FSAL,
@@ -610,7 +663,7 @@ static fsal_status_t ceph_fsal_readlink(struct fsal_obj_handle *link_pub,
 	char content[PATH_MAX];
 
 	rc = fsal_ceph_ll_readlink(export->cmount, link->i, content,
-				   PATH_MAX, op_ctx->creds);
+				   PATH_MAX, &op_ctx->creds);
 	if (rc < 0)
 		return ceph2fsal_error(rc);
 
@@ -635,7 +688,7 @@ static fsal_status_t ceph_fsal_readlink(struct fsal_obj_handle *link_pub,
  */
 
 static fsal_status_t ceph_fsal_getattrs(struct fsal_obj_handle *handle_pub,
-					struct attrlist *attrs)
+					struct fsal_attrlist *attrs)
 {
 	/* Generic status return */
 	int rc = 0;
@@ -647,9 +700,13 @@ static fsal_status_t ceph_fsal_getattrs(struct fsal_obj_handle *handle_pub,
 		container_of(handle_pub, struct ceph_handle, handle);
 	/* Stat buffer */
 	struct ceph_statx stx;
+#ifdef CEPHFS_POSIX_ACL
+	/* Object file type */
+	bool is_dir;
+#endif				/* CEPHFS_POSIX_ACL */
 
 	rc = fsal_ceph_ll_getattr(export->cmount, handle->i, &stx,
-				CEPH_STATX_ATTR_MASK, op_ctx->creds);
+				CEPH_STATX_ATTR_MASK, &op_ctx->creds);
 	if (rc < 0)
 		goto out_err;
 
@@ -657,6 +714,22 @@ static fsal_status_t ceph_fsal_getattrs(struct fsal_obj_handle *handle_pub,
 	if (rc < 0)
 		goto out_err;
 
+#ifdef CEPHFS_POSIX_ACL
+	if (attrs->request_mask & ATTR_ACL) {
+		is_dir = (bool)(handle_pub->type == DIRECTORY);
+		rc = ceph_get_acl(export, handle, is_dir, attrs);
+		if (rc < 0) {
+			LogDebug(COMPONENT_FSAL, "failed to get acl: %d", rc);
+			goto out_err;
+		}
+	}
+#endif				/* CEPHFS_POSIX_ACL */
+
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_getattrs, __func__, __LINE__,
+		   stx.stx_ino, stx.stx_size, stx.stx_mode);
+#endif
+
 	ceph2fsal_attributes(&stx, attrs);
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 out_err:
@@ -697,7 +770,7 @@ static fsal_status_t ceph_fsal_link(struct fsal_obj_handle *handle_pub,
 		container_of(destdir_pub, struct ceph_handle, handle);
 
 	rc = fsal_ceph_ll_link(export->cmount, handle->i, destdir->i, name,
-				op_ctx->creds);
+				&op_ctx->creds);
 	if (rc < 0)
 		return ceph2fsal_error(rc);
 
@@ -737,7 +810,7 @@ static fsal_status_t ceph_fsal_rename(struct fsal_obj_handle *obj_hdl,
 		container_of(newdir_pub, struct ceph_handle, handle);
 
 	rc = fsal_ceph_ll_rename(export->cmount, olddir->i, old_name,
-					newdir->i, new_name, op_ctx->creds);
+					newdir->i, new_name, &op_ctx->creds);
 	if (rc < 0) {
 		/*
 		 * RFC5661, section 18.26.3 - renaming on top of a non-empty
@@ -783,10 +856,10 @@ static fsal_status_t ceph_fsal_unlink(struct fsal_obj_handle *dir_pub,
 
 	if (obj_pub->type != DIRECTORY) {
 		rc = fsal_ceph_ll_unlink(export->cmount, dir->i, name,
-					op_ctx->creds);
+					&op_ctx->creds);
 	} else {
 		rc = fsal_ceph_ll_rmdir(export->cmount, dir->i, name,
-					op_ctx->creds);
+					&op_ctx->creds);
 	}
 
 	if (rc < 0) {
@@ -796,6 +869,11 @@ static fsal_status_t ceph_fsal_unlink(struct fsal_obj_handle *dir_pub,
 		return ceph2fsal_error(rc);
 	}
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_unlink, __func__, __LINE__,
+		   name, object_file_type_to_str(obj_pub->type));
+#endif
+
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
 
@@ -830,11 +908,10 @@ static fsal_status_t ceph_open_my_fd(struct ceph_handle *myself,
 		.caller_garray = NULL
 	};
 	struct user_cred *creds =
-		access == CEPH_AS_ROOT ? &root_creds : op_ctx->creds;
+		access == CEPH_AS_ROOT ? &root_creds : &op_ctx->creds;
 
 	LogFullDebug(COMPONENT_FSAL,
-		     "my_fd = %p my_fd->fd = %p openflags = %x,"
-		     " posix_flags = %x, access = %d",
+		     "my_fd = %p my_fd->fd = %p openflags = %x, posix_flags = %x, access = %d",
 		     my_fd, my_fd->fd, openflags, posix_flags, access);
 
 	assert(my_fd->fd == NULL
@@ -960,6 +1037,11 @@ static fsal_status_t ceph_fsal_close(struct fsal_obj_handle *obj_hdl)
 
 	PTHREAD_RWLOCK_unlock(&obj_hdl->obj_lock);
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_close, __func__, __LINE__,
+		   obj_hdl->fileid);
+#endif
+
 	return status;
 }
 
@@ -1047,11 +1129,7 @@ static fsal_status_t ceph_fsal_merge(struct fsal_obj_handle *orig_hdl,
 		dupe = container_of(dupe_hdl, struct ceph_handle, handle);
 
 		/* This can block over an I/O operation. */
-		PTHREAD_RWLOCK_wrlock(&orig_hdl->obj_lock);
-
-		status = merge_share(&orig->share, &dupe->share);
-
-		PTHREAD_RWLOCK_unlock(&orig_hdl->obj_lock);
+		status = merge_share(orig_hdl, &orig->share, &dupe->share);
 	}
 
 	return status;
@@ -1126,10 +1204,10 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
 				 fsal_openflags_t openflags,
 				 enum fsal_create_mode createmode,
 				 const char *name,
-				 struct attrlist *attrib_set,
+				 struct fsal_attrlist *attrib_set,
 				 fsal_verifier_t verifier,
 				 struct fsal_obj_handle **new_obj,
-				 struct attrlist *attrs_out,
+				 struct fsal_attrlist *attrs_out,
 				 bool *caller_perm_check)
 {
 	int posix_flags = 0;
@@ -1161,11 +1239,12 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
 
 	if (createmode >= FSAL_EXCLUSIVE) {
 		/* Now fixup attrs for verifier if exclusive create */
-		set_common_verifier(attrib_set, verifier);
+		set_common_verifier(attrib_set, verifier, false);
 	}
 
 	if (name == NULL) {
 		/* This is an open by handle */
+		/* **TODO** split open by handle to another function */
 		if (state != NULL) {
 			/* Prepare to take the share reservation, but only if we
 			 * are called with a valid state (if state is NULL the
@@ -1222,7 +1301,7 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
 			/* Refresh the attributes */
 			retval = fsal_ceph_ll_getattr(export->cmount,
 					myself->i, &stx, !!attrs_out,
-					op_ctx->creds);
+					&op_ctx->creds);
 
 			if (retval == 0) {
 				LogFullDebug(COMPONENT_FSAL,
@@ -1349,6 +1428,11 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
 				     fsal_err_txt(status));
 		}
 
+#ifdef USE_LTTNG
+		tracepoint(fsalceph, ceph_open, __func__, __LINE__,
+			   name, *new_obj, stx.stx_ino, "opend");
+#endif
+
 		return status;
 	}
 
@@ -1384,7 +1468,7 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
 
 	retval = fsal_ceph_ll_create(export->cmount,  myself->i, name,
 				unix_mode, posix_flags, &i, &fd, &stx,
-				!!attrs_out, op_ctx->creds);
+				!!attrs_out, &op_ctx->creds);
 
 	if (retval < 0) {
 		LogFullDebug(COMPONENT_FSAL,
@@ -1406,7 +1490,7 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
 		posix_flags &= ~O_EXCL;
 		retval = fsal_ceph_ll_create(export->cmount,  myself->i,
 				name, unix_mode, posix_flags, &i, &fd,
-				&stx, !!attrs_out, op_ctx->creds);
+				&stx, !!attrs_out, &op_ctx->creds);
 		if (retval < 0) {
 			LogFullDebug(COMPONENT_FSAL,
 				     "Non-exclusive Create %s failed with %s",
@@ -1469,6 +1553,11 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
 
 	*new_obj = &hdl->handle;
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_open, __func__, __LINE__,
+		   name, &hdl->handle, stx.stx_ino, "created");
+#endif
+
 	if (created && attrib_set->valid_mask != 0) {
 		/* Set attributes using our newly opened file descriptor as the
 		 * share_fd if there are any left to set (mode and truncate
@@ -1477,10 +1566,12 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
 		 * Note that we only set the attributes if we were responsible
 		 * for creating the file and we have attributes to set.
 		 */
+		op_ctx->fsal_private = CEPH_SETXATTR_AS_ROOT;
 		status = (*new_obj)->obj_ops->setattr2(*new_obj,
 						      false,
 						      state,
 						      attrib_set);
+		op_ctx->fsal_private = NULL;
 
 		if (FSAL_IS_ERROR(status))
 			goto fileerr;
@@ -1534,7 +1625,7 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
 	if (created) {
 		/* Remove the file we just created */
 		fsal_ceph_ll_unlink(export->cmount, myself->i, name,
-					op_ctx->creds);
+					&op_ctx->creds);
 	}
 
 	return status;
@@ -1544,7 +1635,7 @@ static fsal_status_t ceph_fsal_open2(struct fsal_obj_handle *obj_hdl,
  * @brief Return open status of a state.
  *
  * This function returns open flags representing the current open
- * status for a state. The state_lock must be held.
+ * status for a state. The st_lock must be held.
  *
  * @param[in] obj_hdl     File on which to operate
  * @param[in] state       File state to interrogate
@@ -1564,7 +1655,7 @@ static fsal_openflags_t ceph_fsal_status2(struct fsal_obj_handle *obj_hdl,
  * @brief Re-open a file that may be already opened
  *
  * This function supports changing the access mode of a share reservation and
- * thus should only be called with a share state. The state_lock must be held.
+ * thus should only be called with a share state. The st_lock must be held.
  *
  * This MAY be used to open a file the first time if there is no need for
  * open by name or create semantics. One example would be 9P lopen.
@@ -1760,6 +1851,11 @@ static void ceph_fsal_read2(struct fsal_obj_handle *obj_hdl, bool bypass,
 		offset += nb_read;
 	}
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_read, __func__, __LINE__,
+		   obj_hdl->fileid, nb_read);
+#endif
+
 #if 0
 	/** @todo
 	 *
@@ -1871,6 +1967,11 @@ static void ceph_fsal_write2(struct fsal_obj_handle *obj_hdl, bool bypass,
 		}
 	}
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_write, __func__, __LINE__,
+		   obj_hdl->fileid, nb_written);
+#endif
+
  out:
 
 	if (ceph_fd)
@@ -1918,6 +2019,12 @@ static fsal_status_t ceph_fsal_commit2(struct fsal_obj_handle *obj_hdl,
 	 * altogether.
 	 */
 	retval = ceph_ll_sync_inode(export->cmount, myself->i, 0);
+
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_commit, __func__, __LINE__,
+		   obj_hdl->fileid);
+#endif
+
 	return ceph2fsal_error(retval);
 }
 #else
@@ -1936,8 +2043,7 @@ static fsal_status_t ceph_fsal_commit2(struct fsal_obj_handle *obj_hdl,
 	bool closefd = false;
 	struct ceph_export *export =
 		container_of(op_ctx->fsal_export, struct ceph_export, export);
-	struct user_cred root_creds = {};
-	struct user_cred *saved_creds = op_ctx->creds;
+	struct user_cred saved_creds = op_ctx->creds;
 
 
 	/*
@@ -1946,7 +2052,7 @@ static fsal_status_t ceph_fsal_commit2(struct fsal_obj_handle *obj_hdl,
 	 * permissions since it was opened by the writer, so open the file
 	 * with root creds here since we're just doing a fsync.
 	 */
-	op_ctx->creds = &root_creds;
+	memset(&op_ctx->creds, 0, sizeof(op_ctx->creds));
 	status = fsal_reopen_obj(obj_hdl, false, false, FSAL_O_WRITE,
 				 (struct fsal_fd *)&myself->fd, &myself->share,
 				 ceph_open_func, ceph_close_func,
@@ -1961,6 +2067,11 @@ static fsal_status_t ceph_fsal_commit2(struct fsal_obj_handle *obj_hdl,
 			status = ceph2fsal_error(retval);
 	}
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_commit, __func__, __LINE__,
+		   obj_hdl->fileid);
+#endif
+
 	if (closefd)
 		(void) ceph_ll_close(export->cmount, out_fd->fd);
 
@@ -2145,6 +2256,11 @@ static fsal_status_t ceph_fsal_lock_op2(struct fsal_obj_handle *obj_hdl,
 
 	/* Fall through (retval == 0) */
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_lock, __func__, __LINE__,
+		   obj_hdl->fileid, lock_op);
+#endif
+
  err:
 
 	if (ceph_fd)
@@ -2239,6 +2355,11 @@ static fsal_status_t ceph_fsal_lease_op2(struct fsal_obj_handle *obj_hdl,
 	retval = ceph_ll_delegation(export->cmount, my_fd, cmd, ceph_deleg_cb,
 				    obj_hdl);
 
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_lease, __func__, __LINE__,
+		   obj_hdl->fileid, cmd);
+#endif
+
 	if (ceph_fd)
 		PTHREAD_RWLOCK_unlock(&ceph_fd->fdlock);
 
@@ -2267,7 +2388,7 @@ static fsal_status_t ceph_fsal_lease_op2(struct fsal_obj_handle *obj_hdl,
  */
 static fsal_status_t ceph_fsal_setattr2(struct fsal_obj_handle *obj_hdl,
 					bool bypass, struct state_t *state,
-					struct attrlist *attrib_set)
+					struct fsal_attrlist *attrib_set)
 {
 	struct ceph_handle *myself =
 		container_of(obj_hdl, struct ceph_handle, handle);
@@ -2299,6 +2420,26 @@ static fsal_status_t ceph_fsal_setattr2(struct fsal_obj_handle *obj_hdl,
 		attrib_set->mode &=
 		    ~op_ctx->fsal_export->exp_ops.fs_umask(op_ctx->fsal_export);
 
+#ifdef CEPHFS_POSIX_ACL
+	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_ACL)) {
+		status = ceph_set_acl(export, myself, false, attrib_set);
+		if (FSAL_IS_ERROR(status)) {
+			LogMajor(COMPONENT_FSAL, "set access acl status = %s",
+							fsal_err_txt(status));
+			goto out;
+		}
+
+		if (obj_hdl->type == DIRECTORY) {
+			status = ceph_set_acl(export, myself, true, attrib_set);
+			if (FSAL_IS_ERROR(status)) {
+				LogWarn(COMPONENT_FSAL,
+						"set default acl status = %s",
+						fsal_err_txt(status));
+			}
+		}
+	}
+#endif				/* CEPHFS_POSIX_ACL */
+
 	/* Test if size is being set, make sure file is regular and if so,
 	 * require a read/write file descriptor.
 	 */
@@ -2408,7 +2549,13 @@ static fsal_status_t ceph_fsal_setattr2(struct fsal_obj_handle *obj_hdl,
 #endif
 
 	rc = fsal_ceph_ll_setattr(export->cmount, myself->i, &stx, mask,
-					op_ctx->creds);
+					&op_ctx->creds);
+
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_setattrs, __func__, __LINE__,
+		   stx.stx_ino, stx.stx_size, stx.stx_mode);
+#endif
+
 	if (rc < 0) {
 		LogDebug(COMPONENT_FSAL,
 			 "setattrx returned %s (%d)",
@@ -2417,11 +2564,16 @@ static fsal_status_t ceph_fsal_setattr2(struct fsal_obj_handle *obj_hdl,
 	}
 
 	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR4_SEC_LABEL)) {
+		struct user_cred creds = op_ctx->creds;
+
+		if (op_ctx->fsal_private == CEPH_SETXATTR_AS_ROOT)
+			memset(&creds, 0, sizeof(creds));
+
 		rc = fsal_ceph_ll_setxattr(export->cmount, myself->i,
 				export->sec_label_xattr,
 				attrib_set->sec_label.slai_data.slai_data_val,
 				attrib_set->sec_label.slai_data.slai_data_len,
-				0, op_ctx->creds);
+				0, &creds);
 		if (rc < 0) {
 			status = ceph2fsal_error(rc);
 			goto out;
@@ -2567,6 +2719,7 @@ static fsal_status_t ceph_fsal_fallocate(struct fsal_obj_handle *obj_hdl,
 	Fh *my_fd = NULL;
 	bool has_lock = false;
 	bool closefd = false;
+	int mode;
 	fsal_openflags_t openflags = FSAL_O_WRITE;
 	struct ceph_export *export =
 		container_of(op_ctx->fsal_export, struct ceph_export, export);
@@ -2591,10 +2744,10 @@ static fsal_status_t ceph_fsal_fallocate(struct fsal_obj_handle *obj_hdl,
 		goto out;
 	}
 
-	retval = ceph_ll_fallocate(export->cmount, my_fd,
-				   allocate ? 0 :
-				   FALLOC_FL_KEEP_SIZE | FALLOC_FL_PUNCH_HOLE,
+	mode = allocate ? 0 : FALLOC_FL_KEEP_SIZE | FALLOC_FL_PUNCH_HOLE;
+	retval = ceph_ll_fallocate(export->cmount, my_fd, mode,
 				   offset, length);
+
 	if (retval < 0) {
 		status = ceph2fsal_error(retval);
 		goto out;
@@ -2603,6 +2756,12 @@ static fsal_status_t ceph_fsal_fallocate(struct fsal_obj_handle *obj_hdl,
 	retval = ceph_ll_fsync(export->cmount, my_fd, false);
 	if (retval < 0)
 		status = ceph2fsal_error(retval);
+
+#ifdef USE_LTTNG
+	tracepoint(fsalceph, ceph_falloc, __func__, __LINE__,
+		   obj_hdl->fileid, mode, offset, length);
+#endif
+
  out:
 	if (ceph_fd)
 		PTHREAD_RWLOCK_unlock(&ceph_fd->fdlock);
@@ -2617,6 +2776,222 @@ static fsal_status_t ceph_fsal_fallocate(struct fsal_obj_handle *obj_hdl,
 }
 #endif
 
+static fsal_status_t ceph_fsal_getxattrs(struct fsal_obj_handle *handle_pub,
+					 xattrkey4 *xa_name,
+					 xattrvalue4 *xa_value)
+{
+	int rc = 0;
+	fsal_status_t status;
+	struct ceph_export *export =
+		container_of(op_ctx->fsal_export, struct ceph_export, export);
+	const struct ceph_handle *handle =
+		container_of(handle_pub, const struct ceph_handle, handle);
+	char name[sizeof("user.") + NAME_MAX];
+
+	/*
+	 * The nfs client only deals with user.* xattrs, but doesn't send
+	 * the namespace on the wire. We have to add it in here.
+	 */
+	rc = snprintf(name, sizeof(name), "user.%.*s",
+		      xa_name->utf8string_len, xa_name->utf8string_val);
+	if (rc >= sizeof(name))
+		return ceph2fsal_error(-ENAMETOOLONG);
+
+	rc = fsal_ceph_ll_getxattr(export->cmount, handle->i, name,
+			      xa_value->utf8string_val,
+			      xa_value->utf8string_len,
+			      &op_ctx->creds);
+
+	if (rc < 0) {
+		LogDebug(COMPONENT_FSAL, "GETXATTRS returned rc %d", rc);
+
+		if (rc == -ERANGE) {
+			status = fsalstat(ERR_FSAL_XATTR2BIG, 0);
+			goto out;
+		}
+		if (rc == -ENODATA) {
+			status = fsalstat(ERR_FSAL_NOXATTR, 0);
+			goto out;
+		}
+		status = ceph2fsal_error(rc);
+		goto out;
+	}
+	xa_value->utf8string_len = rc;
+
+	LogDebug(COMPONENT_FSAL,
+		 "GETXATTRS %s is '%.*s'", name,
+		 xa_value->utf8string_len,
+		 xa_value->utf8string_val);
+
+	status = fsalstat(ERR_FSAL_NO_ERROR, 0);
+out:
+	return status;
+}
+
+static fsal_status_t ceph_fsal_setxattrs(struct fsal_obj_handle *handle_pub,
+					 setxattr_option4 option,
+					 xattrkey4 *xa_name,
+					 xattrvalue4 *xa_value)
+{
+	int rc = 0;
+	int flags;
+	fsal_status_t status = {0, 0};
+	struct ceph_export *export =
+		container_of(op_ctx->fsal_export, struct ceph_export, export);
+	const struct ceph_handle *handle =
+		container_of(handle_pub, const struct ceph_handle, handle);
+	char name[sizeof("user.") + NAME_MAX];
+
+	/*
+	 * The nfs client only deals with user.* xattrs, but doesn't send
+	 * the namespace on the wire. We have to add it in here.
+	 */
+	rc = snprintf(name, sizeof(name), "user.%.*s",
+		      xa_name->utf8string_len, xa_name->utf8string_val);
+	if (rc >= sizeof(name))
+		return ceph2fsal_error(-ENAMETOOLONG);
+
+	switch (option) {
+	case SETXATTR4_EITHER:
+		flags = 0;
+		break;
+	case SETXATTR4_CREATE:
+		flags = XATTR_CREATE;
+		break;
+	case SETXATTR4_REPLACE:
+		flags = XATTR_REPLACE;
+		break;
+	default:
+		return ceph2fsal_error(-EINVAL);
+	}
+
+	LogDebug(COMPONENT_FSAL,
+			"SETXATTR of %s to %*.s", name,
+			xa_value->utf8string_len,
+			xa_value->utf8string_val);
+	rc = fsal_ceph_ll_setxattr(export->cmount, handle->i, name,
+			      xa_value->utf8string_val,
+			      xa_value->utf8string_len,
+			      flags, &op_ctx->creds);
+	if (rc < 0) {
+		LogDebug(COMPONENT_FSAL,
+			 "SETXATTRS returned rc %d", rc);
+		if (rc == -ERANGE) {
+			status = fsalstat(ERR_FSAL_XATTR2BIG, 0);
+			goto out;
+		}
+		if (rc == -ENODATA) {
+			status = fsalstat(ERR_FSAL_NOXATTR, 0);
+			goto out;
+		}
+		status = ceph2fsal_error(rc);
+		goto out;
+	}
+	status = fsalstat(ERR_FSAL_NO_ERROR, 0);
+out:
+	return status;
+}
+
+static fsal_status_t ceph_fsal_removexattrs(struct fsal_obj_handle *handle_pub,
+					    xattrkey4 *xa_name)
+{
+	int rc = 0;
+	fsal_status_t status = {0, 0};
+	struct ceph_export *export =
+		container_of(op_ctx->fsal_export, struct ceph_export, export);
+	const struct ceph_handle *handle =
+		container_of(handle_pub, const struct ceph_handle, handle);
+	char name[sizeof("user.") + NAME_MAX];
+
+	/*
+	 * The nfs client only deals with user.* xattrs, but doesn't send
+	 * the namespace on the wire. We have to add it in here.
+	 */
+	rc = snprintf(name, sizeof(name), "user.%.*s",
+		      xa_name->utf8string_len, xa_name->utf8string_val);
+	if (rc >= sizeof(name))
+		return ceph2fsal_error(-ENAMETOOLONG);
+
+	rc = fsal_ceph_ll_removexattr(export->cmount, handle->i, name,
+				      &op_ctx->creds);
+	if (rc < 0) {
+		if (rc == -ERANGE) {
+			status = fsalstat(ERR_FSAL_XATTR2BIG, 0);
+			goto out;
+		}
+		if (rc == -ENODATA) {
+			status = fsalstat(ERR_FSAL_NOXATTR, 0);
+			goto out;
+		}
+		LogDebug(COMPONENT_FSAL,
+			 "REMOVEXATTR returned rc %d", rc);
+		status = ceph2fsal_error(rc);
+		goto out;
+	}
+	status = fsalstat(ERR_FSAL_NO_ERROR, 0);
+out:
+	return status;
+}
+
+static fsal_status_t ceph_fsal_listxattrs(struct fsal_obj_handle *handle_pub,
+					  uint32_t maxbytes,
+					  nfs_cookie4 *lxa_cookie,
+					  bool_t *lxr_eof,
+					  xattrlist4 *lxr_names)
+{
+	char *buf = NULL;
+	int rc, loop;
+	size_t listlen = 0;
+	struct ceph_export *export =
+		container_of(op_ctx->fsal_export, struct ceph_export, export);
+	const struct ceph_handle *handle =
+		container_of(handle_pub, const struct ceph_handle, handle);
+	UserPerm *perms = user_cred2ceph(&op_ctx->creds);
+	fsal_status_t status;
+
+	if (!perms)
+		return fsalstat(ERR_FSAL_NOMEM, ENOMEM);
+
+	/* Log Message */
+	LogFullDebug(COMPONENT_FSAL,
+			"in cookie %llu length %d",
+			(unsigned long long)lxa_cookie, maxbytes);
+
+	/* Get a listing, but give up if we keep getting ERANGE back. */
+	loop = 0;
+	do {
+		rc = ceph_ll_listxattr(export->cmount, handle->i, NULL, 0,
+					&listlen, perms);
+		if (rc < 0) {
+			status = ceph2fsal_error(rc);
+			goto out;
+		}
+
+		gsh_free(buf);
+		buf = gsh_malloc(listlen);
+		rc = ceph_ll_listxattr(export->cmount, handle->i, buf, listlen,
+					&listlen, perms);
+	} while (rc == -ERANGE && loop++ < 5);
+
+	if (rc < 0) {
+		LogDebug(COMPONENT_FSAL, "ceph_ll_listxattr returned rc %d",
+				rc);
+		if (rc == -ERANGE) {
+			status = fsalstat(ERR_FSAL_SERVERFAULT, 0);
+			goto out;
+		}
+		status = ceph2fsal_error(rc);
+		goto out;
+	}
+
+	status = fsal_listxattr_helper(buf, listlen, maxbytes, lxa_cookie,
+				       lxr_eof, lxr_names);
+out:
+	gsh_free(buf);
+	ceph_userperm_destroy(perms);
+	return status;
+}
+
 /**
  * @brief Override functions in ops vector
  *
@@ -2665,4 +3040,8 @@ void handle_ops_init(struct fsal_obj_ops *ops)
 #ifdef USE_CEPH_LL_FALLOCATE
 	ops->fallocate = ceph_fsal_fallocate;
 #endif
+	ops->getxattrs = ceph_fsal_getxattrs;
+	ops->setxattrs = ceph_fsal_setxattrs;
+	ops->listxattrs = ceph_fsal_listxattrs;
+	ops->removexattrs = ceph_fsal_removexattrs;
 }
