@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /** @file export.c
  *  @brief GPFS FSAL module export functions.
  *
@@ -42,6 +43,7 @@
 #include "fsal_convert.h"
 #include "FSAL/fsal_commonlib.h"
 #include "FSAL/fsal_config.h"
+#include "FSAL/fsal_localfs.h"
 #include "gpfs_methods.h"
 #include "nfs_exports.h"
 #include "export_mgr.h"
@@ -91,8 +93,8 @@ static fsal_status_t get_dynamic_info(struct fsal_export *exp_hdl,
 	infop->avail_files = buffstatgpfs.f_ffree;
 	infop->maxread = buffstatgpfs.f_bsize;
 	infop->maxwrite = buffstatgpfs.f_bsize;
-	infop->time_delta.tv_sec = 1;
-	infop->time_delta.tv_nsec = 0;
+	infop->time_delta.tv_sec = 0;
+	infop->time_delta.tv_nsec = FSAL_DEFAULT_TIME_DELTA_NSEC;
 
  out:
 	return fsalstat(fsal_error, 0);
@@ -158,10 +160,11 @@ get_quota(struct fsal_export *exp_hdl, const char *filepath, int quota_type,
 	args.cmd = GPFS_QCMD(Q_GETQUOTA, quota_type);
 	args.qid = quota_id;
 	args.bufferP = &gpfs_quota;
+	args.cli_ip = NULL;
 	if (op_ctx && op_ctx->client)
 		args.cli_ip = op_ctx->client->hostaddr_str;
 
-	fsal_set_credentials(op_ctx->creds);
+	fsal_set_credentials(&op_ctx->creds);
 	if (gpfs_ganesha(OPENHANDLE_QUOTA, &args) < 0)
 		retval = errno;
 	fsal_restore_ganesha_credentials();
@@ -225,10 +228,11 @@ set_quota(struct fsal_export *exp_hdl, const char *filepath, int quota_type,
 	args.cmd = GPFS_QCMD(Q_SETQUOTA, quota_type);
 	args.qid = quota_id;
 	args.bufferP = &gpfs_quota;
+	args.cli_ip = NULL;
 	if (op_ctx && op_ctx->client)
 		args.cli_ip = op_ctx->client->hostaddr_str;
 
-	fsal_set_credentials(op_ctx->creds);
+	fsal_set_credentials(&op_ctx->creds);
 	if (gpfs_ganesha(OPENHANDLE_QUOTA, &args) < 0)
 		retval = errno;
 	fsal_restore_ganesha_credentials();
@@ -371,7 +375,7 @@ gpfs_free_state(struct fsal_export *exp_hdl, struct state_t *state)
 
 /**
  *  @brief overwrite vector entries with the methods that we support
- *  @param ops tpye of struct export_ops
+ *  @param ops type of struct export_ops
  */
 void gpfs_export_ops_init(struct export_ops *ops)
 {
@@ -397,7 +401,7 @@ static void free_gpfs_filesystem(struct gpfs_filesystem *gpfs_fs)
 }
 
 /**
- *  @brief Extract major from from fsid
+ *  @brief Extract major from fsid
  *  @param fh GPFS file handle
  *  @param fsid FSAL ID
  */
@@ -469,7 +473,9 @@ errout:
  *  @param exp FSAL export
  *  @return 0(zero) on success, otherwise error.
  */
-int gpfs_claim_filesystem(struct fsal_filesystem *fs, struct fsal_export *exp)
+int gpfs_claim_filesystem(struct fsal_filesystem *fs,
+			  struct fsal_export *exp,
+			  void **private_data)
 {
 	struct gpfs_filesystem *gpfs_fs;
 	int retval;
@@ -510,18 +516,18 @@ int gpfs_claim_filesystem(struct fsal_filesystem *fs, struct fsal_export *exp)
 	glist_add_tail(&map->exp->filesystems, &map->on_filesystems);
 	PTHREAD_MUTEX_unlock(&gpfs_fs->upvector_mutex);
 
-	map->exp->export_fd = open(op_ctx->ctx_export->fullpath,
-						O_RDONLY | O_DIRECTORY);
+	map->exp->export_fd = open(CTX_FULLPATH(op_ctx),
+				   O_RDONLY | O_DIRECTORY);
 	if (map->exp->export_fd < 0) {
 		retval = errno;
 		LogMajor(COMPONENT_FSAL,
 			"Could not open GPFS export point %s: rc = %s (%d)",
-			op_ctx->ctx_export->fullpath, strerror(retval), retval);
+			CTX_FULLPATH(op_ctx), strerror(retval), retval);
 		goto errout;
 	}
 
 	LogFullDebug(COMPONENT_FSAL, "export_fd %d path %s",
-			map->exp->export_fd, op_ctx->ctx_export->fullpath);
+			map->exp->export_fd, CTX_FULLPATH(op_ctx));
 
 	/* We have set up the export. If the file system is already claimed,
 	 * we are done.
@@ -565,7 +571,7 @@ int gpfs_claim_filesystem(struct fsal_filesystem *fs, struct fsal_export *exp)
 		goto errout;
 	}
 
-	fs->private_data = gpfs_fs;
+	*private_data = gpfs_fs;
 
 	return 0;
 
@@ -669,7 +675,7 @@ void gpfs_unexport_filesystems(struct gpfs_fsal_export *exp)
 			LogInfo(COMPONENT_FSAL,
 				"GPFS is no longer exporting filesystem %s",
 				map->fs->fs->path);
-			unclaim_fs(map->fs->fs);
+			//unclaim_fs(map->fs->fs);
 		}
 
 		/* And free it */
@@ -726,8 +732,7 @@ gpfs_create_export(struct fsal_module *fsal_hdl, void *parse_node,
 
 	status.minor = fsal_internal_version();
 	LogInfo(COMPONENT_FSAL, "GPFS get version is %d options 0x%X id %d",
-		status.minor,
-		op_ctx->export_perms ?  op_ctx->export_perms->options : 0,
+		status.minor, op_ctx->export_perms.options,
 		op_ctx->ctx_export->export_id);
 
 	fsal_export_init(exp);
@@ -742,7 +747,7 @@ gpfs_create_export(struct fsal_module *fsal_hdl, void *parse_node,
 	if (rc != 0) {
 		LogCrit(COMPONENT_FSAL,
 			"Incorrect or missing parameters for export %s",
-			op_ctx->ctx_export->fullpath);
+			CTX_FULLPATH(op_ctx));
 		status.major = ERR_FSAL_INVAL;
 		goto free;
 	}
@@ -757,7 +762,7 @@ gpfs_create_export(struct fsal_module *fsal_hdl, void *parse_node,
 	exp->up_ops = up_ops;
 	op_ctx->fsal_export = exp;
 
-	status.minor = resolve_posix_filesystem(op_ctx->ctx_export->fullpath,
+	status.minor = resolve_posix_filesystem(CTX_FULLPATH(op_ctx),
 						fsal_hdl, exp,
 						gpfs_claim_filesystem,
 						gpfs_unclaim_filesystem,
@@ -766,7 +771,7 @@ gpfs_create_export(struct fsal_module *fsal_hdl, void *parse_node,
 	if (status.minor != 0) {
 		LogCrit(COMPONENT_FSAL,
 			"resolve_posix_filesystem(%s) returned %s (%d)",
-			op_ctx->ctx_export->fullpath,
+			CTX_FULLPATH(op_ctx),
 			strerror(status.minor), status.minor);
 		status.major = posix2fsal_error(status.minor);
 		goto detach;
@@ -799,8 +804,9 @@ gpfs_create_export(struct fsal_module *fsal_hdl, void *parse_node,
 	if (gpfs_exp->pnfs_ds_enabled) {
 		struct fsal_pnfs_ds *pds = NULL;
 
-		status = fsal_hdl->m_ops.fsal_pnfs_ds(fsal_hdl, parse_node,
-						      &pds);
+		status = fsal_hdl->m_ops.create_fsal_pnfs_ds(fsal_hdl,
+							     parse_node,
+							     &pds);
 
 		if (status.major != ERR_FSAL_NO_ERROR)
 			goto unexport;
@@ -815,14 +821,15 @@ gpfs_create_export(struct fsal_module *fsal_hdl, void *parse_node,
 				"Server id %d already in use.",
 				pds->id_servers);
 			status.major = ERR_FSAL_EXIST;
-			fsal_pnfs_ds_fini(pds);
-			gsh_free(pds);
+
+			/* Return the ref taken by create_fsal_pnfs_ds */
+			pnfs_ds_put(pds);
 			goto unexport;
 		}
 
 		LogInfo(COMPONENT_FSAL,
 			"gpfs_fsal_create: pnfs ds was enabled for [%s]",
-			op_ctx->ctx_export->fullpath);
+			CTX_FULLPATH(op_ctx));
 		export_ops_pnfs(&exp->exp_ops);
 	}
 	gpfs_exp->use_acl =
