@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * Copyright (C) 2019 Skytechnology sp. z o.o.
  *
@@ -55,7 +56,7 @@ static void lzfs_fsal_release(struct fsal_obj_handle *obj_hdl)
 static fsal_status_t lzfs_fsal_lookup(struct fsal_obj_handle *dir_hdl,
 				      const char *path,
 				      struct fsal_obj_handle **obj_hdl,
-				      struct attrlist *attrs_out)
+				      struct fsal_attrlist *attrs_out)
 {
 	struct lzfs_fsal_export *lzfs_export;
 	struct lzfs_fsal_handle *lzfs_obj, *lzfs_dir;
@@ -69,7 +70,7 @@ static fsal_status_t lzfs_fsal_lookup(struct fsal_obj_handle *dir_hdl,
 
 	LogFullDebug(COMPONENT_FSAL, "path=%s", path);
 
-	rc = liz_cred_lookup(lzfs_export->lzfs_instance, op_ctx->creds,
+	rc = liz_cred_lookup(lzfs_export->lzfs_instance, &op_ctx->creds,
 			     lzfs_dir->inode, path, &node);
 
 	if (rc < 0) {
@@ -104,7 +105,7 @@ static fsal_status_t lzfs_fsal_readdir(struct fsal_obj_handle *dir_hdl,
 	struct lzfs_fsal_handle *lzfs_dir, *lzfs_obj;
 	struct liz_direntry buffer[kBatchSize];
 	struct liz_fileinfo *dir_desc;
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 	off_t direntry_offset = 2;
 	enum fsal_dir_result cb_rc;
 	int rc;
@@ -115,7 +116,7 @@ static fsal_status_t lzfs_fsal_readdir(struct fsal_obj_handle *dir_hdl,
 	lzfs_dir = container_of(dir_hdl, struct lzfs_fsal_handle, handle);
 
 	liz_context_t *ctx = lzfs_fsal_create_context(
-			lzfs_export->lzfs_instance, op_ctx->creds);
+			lzfs_export->lzfs_instance, &op_ctx->creds);
 	dir_desc = liz_opendir(
 			lzfs_export->lzfs_instance, ctx, lzfs_dir->inode);
 	if (!dir_desc) {
@@ -182,9 +183,9 @@ static fsal_status_t lzfs_fsal_readdir(struct fsal_obj_handle *dir_hdl,
  */
 static fsal_status_t lzfs_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 				     const char *name,
-				     struct attrlist *attrib,
+				     struct fsal_attrlist *attrib,
 				     struct fsal_obj_handle **new_obj,
-				     struct attrlist *attrs_out)
+				     struct fsal_attrlist *attrs_out)
 {
 	struct lzfs_fsal_export *lzfs_export;
 	struct lzfs_fsal_handle *lzfs_dir, *lzfs_obj;
@@ -206,7 +207,7 @@ static fsal_status_t lzfs_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 	unix_mode = fsal2unix_mode(attrib->mode) &
 		~op_ctx->fsal_export->exp_ops.fs_umask(op_ctx->fsal_export);
 
-	rc = liz_cred_mkdir(lzfs_export->lzfs_instance, op_ctx->creds,
+	rc = liz_cred_mkdir(lzfs_export->lzfs_instance, &op_ctx->creds,
 			    lzfs_dir->inode, name, unix_mode, &dir_entry);
 
 	if (rc < 0) {
@@ -245,9 +246,9 @@ static fsal_status_t lzfs_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 static fsal_status_t lzfs_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 				      const char *name,
 				      object_file_type_t nodetype,
-				      struct attrlist *attrib,
+				      struct fsal_attrlist *attrib,
 				      struct fsal_obj_handle **new_obj,
-				      struct attrlist *attrs_out)
+				      struct fsal_attrlist *attrs_out)
 {
 	struct lzfs_fsal_export *lzfs_export;
 	struct lzfs_fsal_handle *lzfs_dir, *lzfs_obj;
@@ -291,7 +292,7 @@ static fsal_status_t lzfs_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 		return fsalstat(ERR_FSAL_INVAL, EINVAL);
 	}
 
-	rc = liz_cred_mknod(lzfs_export->lzfs_instance, op_ctx->creds,
+	rc = liz_cred_mknod(lzfs_export->lzfs_instance, &op_ctx->creds,
 			    lzfs_dir->inode, name, unix_mode, unix_dev,
 			    &node_entry);
 	if (rc < 0) {
@@ -333,9 +334,9 @@ static fsal_status_t lzfs_fsal_mknode(struct fsal_obj_handle *dir_hdl,
 static fsal_status_t lzfs_fsal_symlink(struct fsal_obj_handle *dir_hdl,
 				       const char *name,
 				       const char *link_path,
-				       struct attrlist *attrib,
+				       struct fsal_attrlist *attrib,
 				       struct fsal_obj_handle **new_obj,
-				       struct attrlist *attrs_out)
+				       struct fsal_attrlist *attrs_out)
 {
 	struct lzfs_fsal_export *lzfs_export;
 	struct lzfs_fsal_handle *lzfs_dir, *lzfs_obj;
@@ -351,7 +352,7 @@ static fsal_status_t lzfs_fsal_symlink(struct fsal_obj_handle *dir_hdl,
 		     PRIu32 " name=%s", lzfs_export->export.export_id,
 		     lzfs_dir->inode, name);
 
-	rc = liz_cred_symlink(lzfs_export->lzfs_instance, op_ctx->creds,
+	rc = liz_cred_symlink(lzfs_export->lzfs_instance, &op_ctx->creds,
 			      link_path, lzfs_dir->inode, name, &node_entry);
 	if (rc < 0) {
 		return lzfs_fsal_last_err();
@@ -405,7 +406,7 @@ static fsal_status_t lzfs_fsal_readlink(struct fsal_obj_handle *link_hdl,
 	LogFullDebug(COMPONENT_FSAL, "export=%" PRIu16 " inode=%" PRIu32,
 		     lzfs_export->export.export_id, lzfs_link->inode);
 
-	rc = liz_cred_readlink(lzfs_export->lzfs_instance, op_ctx->creds,
+	rc = liz_cred_readlink(lzfs_export->lzfs_instance, &op_ctx->creds,
 			       lzfs_link->inode, result,
 			       LIZARDFS_MAX_READLINK_LENGTH);
 	if (rc < 0) {
@@ -423,7 +424,7 @@ static fsal_status_t lzfs_fsal_readlink(struct fsal_obj_handle *link_hdl,
  * \see fsal_api.h for more information
  */
 static fsal_status_t lzfs_fsal_getattrs(struct fsal_obj_handle *obj_hdl,
-					struct attrlist *attrs)
+					struct fsal_attrlist *attrs)
 {
 	struct lzfs_fsal_export *lzfs_export;
 	struct lzfs_fsal_handle *lzfs_obj;
@@ -438,7 +439,7 @@ static fsal_status_t lzfs_fsal_getattrs(struct fsal_obj_handle *obj_hdl,
 	LogFullDebug(COMPONENT_FSAL, "export=%" PRIu16 " inode=%" PRIu32,
 		     lzfs_export->export.export_id, lzfs_obj->inode);
 
-	rc = liz_cred_getattr(lzfs_export->lzfs_instance, op_ctx->creds,
+	rc = liz_cred_getattr(lzfs_export->lzfs_instance, &op_ctx->creds,
 			      lzfs_obj->inode, &lzfs_attrs);
 
 	if (rc < 0) {
@@ -492,7 +493,7 @@ static fsal_status_t lzfs_fsal_rename(struct fsal_obj_handle *obj_hdl,
 		     lzfs_export->export.export_id, lzfs_olddir->inode,
 		     lzfs_newdir->inode, old_name, new_name);
 
-	rc = liz_cred_rename(lzfs_export->lzfs_instance, op_ctx->creds,
+	rc = liz_cred_rename(lzfs_export->lzfs_instance, &op_ctx->creds,
 			     lzfs_olddir->inode, old_name, lzfs_newdir->inode,
 			     new_name);
 
@@ -526,10 +527,10 @@ static fsal_status_t lzfs_fsal_unlink(struct fsal_obj_handle *dir_hdl,
 		     object_file_type_to_str(obj_hdl->type));
 
 	if (obj_hdl->type != DIRECTORY) {
-		rc = liz_cred_unlink(lzfs_export->lzfs_instance, op_ctx->creds,
+		rc = liz_cred_unlink(lzfs_export->lzfs_instance, &op_ctx->creds,
 				     lzfs_dir->inode, name);
 	} else {
-		rc = liz_cred_rmdir(lzfs_export->lzfs_instance, op_ctx->creds,
+		rc = liz_cred_rmdir(lzfs_export->lzfs_instance, &op_ctx->creds,
 				    lzfs_dir->inode, name);
 	}
 
@@ -556,8 +557,9 @@ static fsal_status_t lzfs_fsal_handle_to_wire(
 	liz_inode_t inode = lzfs_obj->inode;
 
 	if (fh_desc->len < sizeof(liz_inode_t)) {
-		LogMajor(COMPONENT_FSAL, "Space too small for handle. Need "
-			 "%zu, have %zu", sizeof(liz_inode_t), fh_desc->len);
+		LogMajor(COMPONENT_FSAL,
+			 "Space too small for handle. Need  %zu, have %zu",
+			 sizeof(liz_inode_t), fh_desc->len);
 		return fsalstat(ERR_FSAL_TOOSMALL, 0);
 	}
 
@@ -599,15 +601,15 @@ static fsal_status_t lzfs_int_open_fd(struct lzfs_fsal_handle *lzfs_obj,
 				   struct lzfs_fsal_export,
 				   export);
 
-	LogFullDebug(COMPONENT_FSAL, "fd = %p fd->fd = %p openflags = %x, "
-		     "posix_flags = %x", lzfs_fd, lzfs_fd->fd, openflags,
-		     posix_flags);
+	LogFullDebug(COMPONENT_FSAL,
+		     "fd = %p fd->fd = %p openflags = %x, posix_flags = %x",
+		     lzfs_fd, lzfs_fd->fd, openflags, posix_flags);
 
 	assert(lzfs_fd->fd == NULL &&
 	       lzfs_fd->openflags == FSAL_O_CLOSED &&
 	       openflags != 0);
 
-	lzfs_fd->fd = liz_cred_open(lzfs_export->lzfs_instance, op_ctx->creds,
+	lzfs_fd->fd = liz_cred_open(lzfs_export->lzfs_instance, &op_ctx->creds,
 				    lzfs_obj->inode, posix_flags);
 
 	if (!lzfs_fd->fd) {
@@ -645,7 +647,7 @@ static fsal_status_t lzfs_int_open_by_handle(struct fsal_obj_handle *obj_hdl,
 					     fsal_openflags_t openflags,
 					     enum fsal_create_mode createmode,
 					     fsal_verifier_t verifier,
-					     struct attrlist *attrs_out,
+					     struct fsal_attrlist *attrs_out,
 					     bool *caller_perm_check,
 					     bool after_mknod)
 {
@@ -703,7 +705,7 @@ static fsal_status_t lzfs_int_open_by_handle(struct fsal_obj_handle *obj_hdl,
 		int rc;
 
 		rc = liz_cred_getattr(lzfs_export->lzfs_instance,
-				      op_ctx->creds,
+				      &op_ctx->creds,
 				      lzfs_hdl->inode,
 				      &lzfs_attrs);
 
@@ -717,7 +719,7 @@ static fsal_status_t lzfs_int_open_by_handle(struct fsal_obj_handle *obj_hdl,
 		if (!FSAL_IS_ERROR(status) &&
 		    createmode >= FSAL_EXCLUSIVE &&
 		    createmode != FSAL_EXCLUSIVE_9P &&
-		    !check_verifier_stat(&lzfs_attrs.attr, verifier)) {
+		    !check_verifier_stat(&lzfs_attrs.attr, verifier, false)) {
 			// Verifier didn't match, return EEXIST
 			status = fsalstat(posix2fsal_error(EEXIST), EEXIST);
 		}
@@ -759,7 +761,7 @@ static fsal_status_t lzfs_int_open_by_name(struct fsal_obj_handle *obj_hdl,
 					   fsal_openflags_t openflags,
 					   const char *name,
 					   fsal_verifier_t verifier,
-					   struct attrlist *attrs_out,
+					   struct fsal_attrlist *attrs_out,
 					   bool *caller_perm_check)
 {
 	struct fsal_obj_handle *temp = NULL;
@@ -795,10 +797,10 @@ static fsal_status_t lzfs_fsal_open2(struct fsal_obj_handle *obj_hdl,
 				     fsal_openflags_t openflags,
 				     enum fsal_create_mode createmode,
 				     const char *name,
-				     struct attrlist *attr_set,
+				     struct fsal_attrlist *attr_set,
 				     fsal_verifier_t verifier,
 				     struct fsal_obj_handle **new_obj,
-				     struct attrlist *attrs_out,
+				     struct fsal_attrlist *attrs_out,
 				     bool *caller_perm_check)
 {
 	struct lzfs_fsal_export *lzfs_export;
@@ -810,7 +812,7 @@ static fsal_status_t lzfs_fsal_open2(struct fsal_obj_handle *obj_hdl,
 	LogAttrlist(COMPONENT_FSAL, NIV_FULL_DEBUG, "attrs ", attr_set, false);
 
 	if (createmode >= FSAL_EXCLUSIVE) {
-		set_common_verifier(attr_set, verifier);
+		set_common_verifier(attr_set, verifier, false);
 	}
 
 	if (name == NULL) {
@@ -841,7 +843,7 @@ static fsal_status_t lzfs_fsal_open2(struct fsal_obj_handle *obj_hdl,
 
 	struct liz_entry lzfs_attrs;
 
-	rc = liz_cred_mknod(lzfs_export->lzfs_instance, op_ctx->creds,
+	rc = liz_cred_mknod(lzfs_export->lzfs_instance, &op_ctx->creds,
 			    lzfs_obj->inode, name, unix_mode, 0, &lzfs_attrs);
 
 	if (rc < 0 && liz_last_err() == LIZARDFS_ERROR_EEXIST &&
@@ -894,7 +896,7 @@ fileerr:
 	(*new_obj)->obj_ops->release(*new_obj);
 	*new_obj = NULL;
 
-	rc = liz_cred_unlink(lzfs_export->lzfs_instance, op_ctx->creds,
+	rc = liz_cred_unlink(lzfs_export->lzfs_instance, &op_ctx->creds,
 			     lzfs_obj->inode, name);
 
 	return status;
@@ -1064,7 +1066,7 @@ static void lzfs_fsal_read2(struct fsal_obj_handle *obj_hdl,
 
 	for (i = 0; i < read_arg->iov_count; i++) {
 		nb_read = liz_cred_read(lzfs_export->lzfs_instance,
-					op_ctx->creds,
+					&op_ctx->creds,
 					lzfs_fd.fd, offset,
 					read_arg->iov[i].iov_len,
 					read_arg->iov[i].iov_base);
@@ -1137,7 +1139,7 @@ struct lzfs_fsal_fd lzfs_fd;
 
 	for (i = 0; i < write_arg->iov_count; i++) {
 		nb_written = liz_cred_write(lzfs_export->lzfs_instance,
-					    op_ctx->creds, lzfs_fd.fd, offset,
+					    &op_ctx->creds, lzfs_fd.fd, offset,
 					    write_arg->iov[i].iov_len,
 					    write_arg->iov[i].iov_base);
 
@@ -1149,7 +1151,7 @@ struct lzfs_fsal_fd lzfs_fd;
 			if (write_arg->fsal_stable) {
 				int rc = liz_cred_fsync(
 						lzfs_export->lzfs_instance,
-						op_ctx->creds, lzfs_fd.fd);
+						&op_ctx->creds, lzfs_fd.fd);
 
 				if (rc < 0) {
 					status = lzfs_fsal_last_err();
@@ -1204,7 +1206,7 @@ static fsal_status_t lzfs_fsal_commit2(struct fsal_obj_handle *obj_hdl,
 
 	if (!FSAL_IS_ERROR(status)) {
 		int rc = liz_cred_fsync(lzfs_export->lzfs_instance,
-					op_ctx->creds, out_fd->fd);
+					&op_ctx->creds, out_fd->fd);
 
 		if (rc < 0) {
 			status = lzfs_fsal_last_err();
@@ -1274,11 +1276,9 @@ static fsal_status_t lzfs_fsal_merge(struct fsal_obj_handle *orig_hdl,
 			     lzfs_orig->unique_key.export_id, lzfs_orig->inode,
 			     lzfs_dupe->inode);
 
-		PTHREAD_RWLOCK_wrlock(&orig_hdl->obj_lock);
-
-		status = merge_share(&lzfs_orig->share, &lzfs_dupe->share);
-
-		PTHREAD_RWLOCK_unlock(&orig_hdl->obj_lock);
+		/* This can block over an I/O operation. */
+		status = merge_share(orig_hdl, &lzfs_orig->share,
+				     &lzfs_dupe->share);
 	}
 
 	return status;
@@ -1291,7 +1291,7 @@ static fsal_status_t lzfs_fsal_merge(struct fsal_obj_handle *orig_hdl,
 static fsal_status_t lzfs_fsal_setattr2(struct fsal_obj_handle *obj_hdl,
 					bool bypass,
 					struct state_t *state,
-					struct attrlist *attrib_set)
+					struct fsal_attrlist *attrib_set)
 {
 	struct lzfs_fsal_export *lzfs_export;
 	struct lzfs_fsal_handle *lzfs_obj;
@@ -1379,7 +1379,7 @@ static fsal_status_t lzfs_fsal_setattr2(struct fsal_obj_handle *obj_hdl,
 	}
 
 	liz_attr_reply_t reply;
-	int rc = liz_cred_setattr(lzfs_export->lzfs_instance, op_ctx->creds,
+	int rc = liz_cred_setattr(lzfs_export->lzfs_instance, &op_ctx->creds,
 				  lzfs_obj->inode, &attr, mask, &reply);
 
 	if (rc < 0) {
@@ -1470,8 +1470,8 @@ fsal_status_t lzfs_fsal_lock_op2(struct fsal_obj_handle *obj_hdl,
 	} else if (lock_op == FSAL_OP_UNLOCK) {
 		openflags = FSAL_O_ANY;
 	} else {
-		LogFullDebug(COMPONENT_FSAL, "ERROR: Lock operation requested "
-			     "was not TEST, READ, or WRITE.");
+		LogFullDebug(COMPONENT_FSAL,
+			     "ERROR: Lock operation requested was not TEST, READ, or WRITE.");
 		return fsalstat(ERR_FSAL_NOTSUPP, 0);
 	}
 
@@ -1485,8 +1485,8 @@ fsal_status_t lzfs_fsal_lock_op2(struct fsal_obj_handle *obj_hdl,
 	} else if (request_lock->lock_type == FSAL_LOCK_W) {
 		lock_info.l_type = F_WRLCK;
 	} else {
-		LogFullDebug(COMPONENT_FSAL, "ERROR: The requested lock type "
-			     "was not read or write.");
+		LogFullDebug(COMPONENT_FSAL,
+			     "ERROR: The requested lock type was not read or write.");
 		return fsalstat(ERR_FSAL_NOTSUPP, 0);
 	}
 
@@ -1509,8 +1509,8 @@ fsal_status_t lzfs_fsal_lock_op2(struct fsal_obj_handle *obj_hdl,
 					  true);
 	}
 	if (FSAL_IS_ERROR(status)) {
-		LogCrit(COMPONENT_FSAL, "Unable to find fd for lock "
-			"operation");
+		LogCrit(COMPONENT_FSAL,
+			"Unable to find fd for lock  operation");
 		return status;
 	}
 
@@ -1518,10 +1518,10 @@ fsal_status_t lzfs_fsal_lock_op2(struct fsal_obj_handle *obj_hdl,
 	liz_set_lock_owner(fileinfo, (uint64_t)owner);
 	if (lock_op == FSAL_OP_LOCKT) {
 		retval = liz_cred_getlk(lzfs_export->lzfs_instance,
-					op_ctx->creds, fileinfo, &lock_info);
+					&op_ctx->creds, fileinfo, &lock_info);
 	} else {
 		retval = liz_cred_setlk(lzfs_export->lzfs_instance,
-					op_ctx->creds, fileinfo, &lock_info);
+					&op_ctx->creds, fileinfo, &lock_info);
 	}
 
 	if (retval < 0) {
@@ -1583,7 +1583,7 @@ struct lzfs_fsal_export *lzfs_export;
 		     lzfs_obj->inode, lzfs_destdir->inode, name);
 
 	liz_entry_t result;
-	int rc = liz_cred_link(lzfs_export->lzfs_instance, op_ctx->creds,
+	int rc = liz_cred_link(lzfs_export->lzfs_instance, &op_ctx->creds,
 			       lzfs_obj->inode, lzfs_destdir->inode, name,
 			       &result);
 	if (rc < 0) {
