@@ -1,7 +1,8 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:shiftwidth=8:tabstop=8:
  *
- * Copyright 2017-2019 Red Hat, Inc.
+ * Copyright 2017-2021 Red Hat, Inc.
  * Author: Daniel Gryniewicz  dang@redhat.com
  *
  *
@@ -292,6 +293,7 @@ mem_dirent_lookup(struct mem_fsal_obj_handle *dir, const char *name)
 	struct mem_dirent key;
 	struct avltree_node *node;
 
+	memset(&key, 0, sizeof(key));
 	key.d_name = name;
 
 	node = avltree_lookup(&key.avl_n, &dir->mh_dir.avl_name);
@@ -355,6 +357,7 @@ mem_readdir_seekloc(struct mem_fsal_obj_handle *dir, fsal_cookie_t seekloc)
 	}
 
 
+	memset(&key, 0, sizeof(key));
 	key.d_index = seekloc;
 	node = avltree_lookup(&key.avl_i, &dir->mh_dir.avl_index);
 	if (!node) {
@@ -404,7 +407,7 @@ static void mem_remove_dirent_locked(struct mem_fsal_obj_handle *parent,
 	avltree_remove(&dirent->avl_i, &parent->mh_dir.avl_index);
 
 	/* Take the child lock, to remove from the child.  This should not race
-	 * with @r mem_insert_obj since that takes the locks seqentially */
+	 * with @r mem_insert_obj since that takes the locks sequentially */
 	child = dirent->hdl;
 	PTHREAD_RWLOCK_wrlock(&child->obj_handle.obj_lock);
 	glist_del(&dirent->dlist);
@@ -479,25 +482,21 @@ void mem_clean_export(struct mem_fsal_obj_handle *root)
 /**
  * @brief Remove all children from a directory's tree
  *
- * @param[in] parent	Directroy to clean
+ * @param[in] parent	Directory to clean
  */
 void mem_clean_all_dirents(struct mem_fsal_obj_handle *parent)
 {
 	struct avltree_node *node;
 	struct mem_dirent *dirent;
 
-	PTHREAD_RWLOCK_wrlock(&parent->obj_handle.obj_lock);
-
 	while ((node = avltree_first(&parent->mh_dir.avl_name))) {
 		dirent = avltree_container_of(node, struct mem_dirent, avl_n);
 		mem_remove_dirent_locked(parent, dirent);
 	}
-
-	PTHREAD_RWLOCK_unlock(&parent->obj_handle.obj_lock);
 }
 
-static void mem_copy_attrs_mask(struct attrlist *attrs_in,
-				struct attrlist *attrs_out)
+static void mem_copy_attrs_mask(struct fsal_attrlist *attrs_in,
+				struct fsal_attrlist *attrs_out)
 {
 	/* Use full timer resolution */
 	now(&attrs_out->ctime);
@@ -615,15 +614,8 @@ static fsal_status_t mem_open_func(struct fsal_obj_handle *obj_hdl,
 static fsal_status_t mem_close_func(struct fsal_obj_handle *obj_hdl,
 				    struct fsal_fd *fd)
 {
-	fsal_status_t status;
-
-	PTHREAD_RWLOCK_wrlock(&obj_hdl->obj_lock);
-
-	status = mem_close_my_fd(fd);
-
-	PTHREAD_RWLOCK_unlock(&obj_hdl->obj_lock);
-
-	return status;
+	(void) obj_hdl;
+	return mem_close_my_fd(fd);
 }
 
 #define mem_alloc_handle(p, n, t, e, a) \
@@ -643,7 +635,7 @@ _mem_alloc_handle(struct mem_fsal_obj_handle *parent,
 		  const char *name,
 		  object_file_type_t type,
 		  struct mem_fsal_export *mfe,
-		  struct attrlist *attrs,
+		  struct fsal_attrlist *attrs,
 		  const char *func, int line)
 {
 	struct mem_fsal_obj_handle *hdl;
@@ -689,12 +681,12 @@ _mem_alloc_handle(struct mem_fsal_obj_handle *parent,
 	if ((attrs && attrs->valid_mask & ATTR_OWNER) != 0)
 		hdl->attrs.owner = attrs->owner;
 	else
-		hdl->attrs.owner = op_ctx->creds->caller_uid;
+		hdl->attrs.owner = op_ctx->creds.caller_uid;
 
 	if ((attrs && attrs->valid_mask & ATTR_GROUP) != 0)
 		hdl->attrs.group = attrs->group;
 	else
-		hdl->attrs.group = op_ctx->creds->caller_gid;
+		hdl->attrs.group = op_ctx->creds.caller_gid;
 
 	/* Use full timer resolution */
 	now(&hdl->attrs.ctime);
@@ -823,9 +815,9 @@ static fsal_status_t _mem_int_lookup(struct mem_fsal_obj_handle *dir,
 static fsal_status_t mem_create_obj(struct mem_fsal_obj_handle *parent,
 				    object_file_type_t type,
 				    const char *name,
-				    struct attrlist *attrs_in,
+				    struct fsal_attrlist *attrs_in,
 				    struct fsal_obj_handle **new_obj,
-				    struct attrlist *attrs_out)
+				    struct fsal_attrlist *attrs_out)
 {
 	struct mem_fsal_export *mfe = container_of(op_ctx->fsal_export,
 						   struct mem_fsal_export,
@@ -883,7 +875,7 @@ static fsal_status_t mem_create_obj(struct mem_fsal_obj_handle *parent,
 static fsal_status_t mem_lookup(struct fsal_obj_handle *parent,
 				const char *path,
 				struct fsal_obj_handle **handle,
-				struct attrlist *attrs_out)
+				struct fsal_attrlist *attrs_out)
 {
 	struct mem_fsal_obj_handle *myself, *hdl = NULL;
 	fsal_status_t status;
@@ -944,7 +936,7 @@ static fsal_status_t mem_readdir(struct fsal_obj_handle *dir_hdl,
 	struct mem_fsal_obj_handle *myself;
 	struct mem_dirent *dirent, *dirent_next;
 	fsal_cookie_t cookie = 0;
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 	enum fsal_dir_result cb_rc;
 	int count = 0;
 
@@ -1039,9 +1031,9 @@ static fsal_status_t mem_readdir(struct fsal_obj_handle *dir_hdl,
  */
 static fsal_status_t mem_mkdir(struct fsal_obj_handle *dir_hdl,
 			       const char *name,
-			       struct attrlist *attrs_in,
+			       struct fsal_attrlist *attrs_in,
 			       struct fsal_obj_handle **new_obj,
-			       struct attrlist *attrs_out)
+			       struct fsal_attrlist *attrs_out)
 {
 	struct mem_fsal_obj_handle *parent =
 		container_of(dir_hdl, struct mem_fsal_obj_handle, obj_handle);
@@ -1071,9 +1063,9 @@ static fsal_status_t mem_mkdir(struct fsal_obj_handle *dir_hdl,
  */
 static fsal_status_t mem_mknode(struct fsal_obj_handle *dir_hdl,
 				const char *name, object_file_type_t nodetype,
-				struct attrlist *attrs_in,
+				struct fsal_attrlist *attrs_in,
 				struct fsal_obj_handle **new_obj,
-				struct attrlist *attrs_out)
+				struct fsal_attrlist *attrs_out)
 {
 	struct mem_fsal_obj_handle *hdl, *parent =
 		container_of(dir_hdl, struct mem_fsal_obj_handle, obj_handle);
@@ -1108,9 +1100,9 @@ static fsal_status_t mem_mknode(struct fsal_obj_handle *dir_hdl,
  */
 static fsal_status_t mem_symlink(struct fsal_obj_handle *dir_hdl,
 				 const char *name, const char *link_path,
-				 struct attrlist *attrs_in,
+				 struct fsal_attrlist *attrs_in,
 				 struct fsal_obj_handle **new_obj,
-				 struct attrlist *attrs_out)
+				 struct fsal_attrlist *attrs_out)
 {
 	struct mem_fsal_obj_handle *hdl, *parent =
 		container_of(dir_hdl, struct mem_fsal_obj_handle, obj_handle);
@@ -1166,7 +1158,7 @@ static fsal_status_t mem_readlink(struct fsal_obj_handle *obj_hdl,
  * @return FSAL status
  */
 static fsal_status_t mem_getattrs(struct fsal_obj_handle *obj_hdl,
-				  struct attrlist *outattrs)
+				  struct fsal_attrlist *outattrs)
 {
 	struct mem_fsal_obj_handle *myself =
 		container_of(obj_hdl, struct mem_fsal_obj_handle, obj_handle);
@@ -1217,7 +1209,7 @@ static fsal_status_t mem_getattrs(struct fsal_obj_handle *obj_hdl,
 fsal_status_t mem_setattr2(struct fsal_obj_handle *obj_hdl,
 			   bool bypass,
 			   struct state_t *state,
-			   struct attrlist *attrs_set)
+			   struct fsal_attrlist *attrs_set)
 {
 	struct mem_fsal_obj_handle *myself =
 		container_of(obj_hdl, struct mem_fsal_obj_handle, obj_handle);
@@ -1496,10 +1488,10 @@ fsal_status_t mem_open2(struct fsal_obj_handle *obj_hdl,
 			fsal_openflags_t openflags,
 			enum fsal_create_mode createmode,
 			const char *name,
-			struct attrlist *attrs_set,
+			struct fsal_attrlist *attrs_set,
 			fsal_verifier_t verifier,
 			struct fsal_obj_handle **new_obj,
-			struct attrlist *attrs_out,
+			struct fsal_attrlist *attrs_out,
 			bool *caller_perm_check)
 {
 	fsal_status_t status = {0, 0};
@@ -1508,7 +1500,7 @@ fsal_status_t mem_open2(struct fsal_obj_handle *obj_hdl,
 	bool truncated;
 	bool setattrs = attrs_set != NULL;
 	bool created = false;
-	struct attrlist verifier_attr;
+	struct fsal_attrlist verifier_attr;
 
 	if (state != NULL)
 		my_fd = (struct fsal_fd *)(state + 1);
@@ -1531,7 +1523,7 @@ fsal_status_t mem_open2(struct fsal_obj_handle *obj_hdl,
 			memset(&verifier_attr, 0, sizeof(verifier_attr));
 		}
 
-		set_common_verifier(attrs_set, verifier);
+		set_common_verifier(attrs_set, verifier, false);
 	}
 
 	if (name == NULL) {
@@ -1584,7 +1576,7 @@ fsal_status_t mem_open2(struct fsal_obj_handle *obj_hdl,
 		 */
 		if (createmode >= FSAL_EXCLUSIVE &&
 		    createmode != FSAL_EXCLUSIVE_9P &&
-		    !check_verifier_attrlist(&myself->attrs, verifier)) {
+		    !check_verifier_attrlist(&myself->attrs, verifier, false)) {
 			/* Verifier didn't match, return EEXIST */
 			status = fsalstat(posix2fsal_error(EEXIST), EEXIST);
 		}
@@ -1720,7 +1712,7 @@ fsal_status_t mem_open2(struct fsal_obj_handle *obj_hdl,
  * @brief Re-open a file that may be already opened
  *
  * This function supports changing the access mode of a share reservation and
- * thus should only be called with a share state. The state_lock must be held.
+ * thus should only be called with a share state. The st_lock must be held.
  *
  * This MAY be used to open a file the first time if there is no need for
  * open by name or create semantics. One example would be 9P lopen.
@@ -1786,7 +1778,7 @@ static void
 mem_async_complete(struct fridgethr_context *ctx)
 {
 	struct mem_async_arg *async_arg = ctx->arg;
-	struct root_op_context root_op_context;
+	struct req_op_context op_context;
 	struct mem_fsal_export *mem_export =
 	   container_of(async_arg->fsal_export, struct mem_fsal_export, export);
 	uint32_t async_delay = atomic_fetch_uint32_t(&mem_export->async_delay);
@@ -1802,15 +1794,17 @@ mem_async_complete(struct fridgethr_context *ctx)
 		usleep(async_delay);
 	}
 
-	/* Need an op context for the call back */
-	init_root_op_context(&root_op_context, async_arg->ctx_export,
-			     async_arg->fsal_export, 0, 0,
-			     UNKNOWN_REQUEST);
+	/* Get a ref to the async_arg->ctx_export and initialize op_context for
+	 * the call back
+	 */
+	get_gsh_export_ref(async_arg->ctx_export);
+	init_op_context_simple(&op_context, async_arg->ctx_export,
+			       async_arg->fsal_export);
 
 	async_arg->done_cb(async_arg->obj_hdl, fsalstat(ERR_FSAL_NO_ERROR, 0),
 			   async_arg->io_arg, async_arg->caller_arg);
 
-	release_root_op_context();
+	release_op_context();
 
 	gsh_free(async_arg);
 }
@@ -1916,8 +1910,9 @@ void mem_read2(struct fsal_obj_handle *obj_hdl,
 	if (has_lock)
 		PTHREAD_RWLOCK_unlock(&obj_hdl->obj_lock);
 
-	if (async_type > MEM_RANDOM_OR_INLINE ||
-	    ((async_type == MEM_RANDOM_OR_INLINE) && ((random() % 1) == 1))) {
+	if (MEM.async_threads > 0 &&
+	    (async_type > MEM_RANDOM_OR_INLINE ||
+	    ((async_type == MEM_RANDOM_OR_INLINE) && ((random() % 1) == 1)))) {
 		struct mem_async_arg *async_arg;
 
 		/* Was MEM_FIXED, MEM_RANDOM, or MEM_RANDOM_OR_INLINE and we
@@ -2053,8 +2048,9 @@ void mem_write2(struct fsal_obj_handle *obj_hdl,
 	if (has_lock)
 		PTHREAD_RWLOCK_unlock(&obj_hdl->obj_lock);
 
-	if (async_type > MEM_RANDOM_OR_INLINE ||
-	    ((async_type == MEM_RANDOM_OR_INLINE) && ((random() % 1) == 1))) {
+	if (MEM.async_threads > 0 &&
+	    (async_type > MEM_RANDOM_OR_INLINE ||
+	    ((async_type == MEM_RANDOM_OR_INLINE) && ((random() % 1) == 1)))) {
 		struct mem_async_arg *async_arg;
 
 		/* Was MEM_FIXED, MEM_RANDOM, or MEM_RANDOM_OR_INLINE and we
@@ -2394,12 +2390,8 @@ static fsal_status_t mem_merge(struct fsal_obj_handle *old_hdl,
 				    obj_handle);
 
 		/* This can block over an I/O operation. */
-		PTHREAD_RWLOCK_wrlock(&old_hdl->obj_lock);
-
-		status = merge_share(&old->mh_file.share,
+		status = merge_share(old_hdl, &old->mh_file.share,
 				     &new->mh_file.share);
-
-		PTHREAD_RWLOCK_unlock(&old_hdl->obj_lock);
 	}
 
 	return status;
@@ -2447,10 +2439,10 @@ void mem_handle_ops_init(struct fsal_obj_ops *ops)
 fsal_status_t mem_lookup_path(struct fsal_export *exp_hdl,
 			      const char *path,
 			      struct fsal_obj_handle **obj_hdl,
-			      struct attrlist *attrs_out)
+			      struct fsal_attrlist *attrs_out)
 {
 	struct mem_fsal_export *mfe;
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 
 	mfe = container_of(exp_hdl, struct mem_fsal_export, export);
 
@@ -2463,7 +2455,7 @@ fsal_status_t mem_lookup_path(struct fsal_export *exp_hdl,
 	}
 
 	attrs.valid_mask = ATTR_MODE;
-	attrs.mode = 0755;
+	attrs.mode = 0777;
 
 	if (mfe->root_handle == NULL) {
 		mfe->root_handle = mem_alloc_handle(NULL,
@@ -2483,7 +2475,7 @@ fsal_status_t mem_lookup_path(struct fsal_export *exp_hdl,
 
 /* create_handle
  * Does what original FSAL_ExpandHandle did (sort of)
- * returns a ref counted handle to be later used in cache_inode etc.
+ * returns a ref counted handle to be later used in mdcache etc.
  * NOTE! you must release this thing when done with it!
  * BEWARE! Thanks to some holes in the *AT syscalls implementation,
  * we cannot get an fd on an AF_UNIX socket, nor reliably on block or
@@ -2496,7 +2488,7 @@ fsal_status_t mem_lookup_path(struct fsal_export *exp_hdl,
 fsal_status_t mem_create_handle(struct fsal_export *exp_hdl,
 				struct gsh_buffdesc *hdl_desc,
 				struct fsal_obj_handle **obj_hdl,
-				struct attrlist *attrs_out)
+				struct fsal_attrlist *attrs_out)
 {
 	struct glist_head *glist;
 	struct fsal_obj_handle *hdl;
