@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -44,9 +45,6 @@
  * released.
  *
  * @param[in,out] export The export to be released
- *
- * @retval ERR_FSAL_NO_ERROR on success.
- * @retval ERR_FSAL_BUSY if the export is in use.
  */
 
 static void release(struct fsal_export *export_pub)
@@ -88,7 +86,7 @@ static void release(struct fsal_export *export_pub)
 static fsal_status_t lookup_path(struct fsal_export *export_pub,
 				 const char *path,
 				 struct fsal_obj_handle **pub_handle,
-				 struct attrlist *attrs_out)
+				 struct fsal_attrlist *attrs_out)
 {
 	/* The 'private' full export handle */
 	struct rgw_export *export =
@@ -258,7 +256,7 @@ static fsal_status_t wire_to_host(struct fsal_export *exp_hdl,
 static fsal_status_t create_handle(struct fsal_export *export_pub,
 				   struct gsh_buffdesc *desc,
 				   struct fsal_obj_handle **pub_handle,
-				   struct attrlist *attrs_out)
+				   struct fsal_attrlist *attrs_out)
 {
 	/* Full 'private' export structure */
 	struct rgw_export *export =
@@ -347,8 +345,8 @@ static fsal_status_t get_fs_dynamic_info(struct fsal_export *export_pub,
 	info->total_files = vfs_st.f_files;
 	info->free_files = vfs_st.f_ffree;
 	info->avail_files = vfs_st.f_favail;
-	info->time_delta.tv_sec = 1;
-	info->time_delta.tv_nsec = 0;
+	info->time_delta.tv_sec = 0;
+	info->time_delta.tv_nsec = FSAL_DEFAULT_TIME_DELTA_NSEC;
 
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
