@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -75,7 +76,7 @@ static void release(struct fsal_obj_handle *obj_hdl)
 static fsal_status_t lookup_int(struct fsal_obj_handle *dir_hdl,
 				const char *path,
 				struct fsal_obj_handle **obj_hdl,
-				struct attrlist *attrs_out,
+				struct fsal_attrlist *attrs_out,
 				struct stat *rcb_st,
 				uint32_t rcb_st_mask,
 				uint32_t flags)
@@ -120,7 +121,7 @@ static fsal_status_t lookup_int(struct fsal_obj_handle *dir_hdl,
 
 static fsal_status_t lookup(struct fsal_obj_handle *dir_hdl,
 			const char *path, struct fsal_obj_handle **obj_hdl,
-			struct attrlist *attrs_out)
+			struct fsal_attrlist *attrs_out)
 {
 	return lookup_int(dir_hdl, path, obj_hdl, attrs_out, NULL, 0,
 			RGW_LOOKUP_FLAG_NONE);
@@ -139,7 +140,7 @@ static bool rgw_cb(const char *name, void *arg, uint64_t offset,
 	struct rgw_cb_arg *rgw_cb_arg = arg;
 	struct fsal_obj_handle *obj = NULL;
 	fsal_status_t status;
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 	enum fsal_dir_result cb_rc;
 
 	fsal_prepare_attrs(&attrs, rgw_cb_arg->attrmask);
@@ -320,9 +321,10 @@ int rgw_fsal_dirent_cmp(
  */
 
 static fsal_status_t rgw_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
-				const char *name, struct attrlist *attrs_in,
+				const char *name,
+				struct fsal_attrlist *attrs_in,
 				struct fsal_obj_handle **obj_hdl,
-				struct attrlist *attrs_out)
+				struct fsal_attrlist *attrs_out)
 {
 	int rc;
 	struct rgw_file_handle *rgw_fh;
@@ -340,8 +342,8 @@ static fsal_status_t rgw_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
 
 	memset(&st, 0, sizeof(struct stat));
 
-	st.st_uid = op_ctx->creds->caller_uid;
-	st.st_gid = op_ctx->creds->caller_gid;
+	st.st_uid = op_ctx->creds.caller_uid;
+	st.st_gid = op_ctx->creds.caller_gid;
 	st.st_mode = fsal2unix_mode(attrs_in->mode)
 	    & ~op_ctx->fsal_export->exp_ops.fs_umask(op_ctx->fsal_export);
 
@@ -379,7 +381,7 @@ static fsal_status_t rgw_fsal_mkdir(struct fsal_obj_handle *dir_hdl,
  * @return FSAL status.
  */
 static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl,
-			struct attrlist *attrs)
+			struct fsal_attrlist *attrs)
 {
 	int rc;
 	struct stat st;
@@ -425,7 +427,7 @@ static fsal_status_t getattrs(struct fsal_obj_handle *obj_hdl,
 fsal_status_t rgw_fsal_setattr2(struct fsal_obj_handle *obj_hdl,
 				bool bypass,
 				struct state_t *state,
-				struct attrlist *attrib_set)
+				struct fsal_attrlist *attrib_set)
 {
 
 	fsal_status_t status = {0, 0};
@@ -698,11 +700,7 @@ fsal_status_t rgw_merge(struct fsal_obj_handle *orig_hdl,
 		dupe = container_of(dupe_hdl, struct rgw_handle, handle);
 
 		/* This can block over an I/O operation. */
-		PTHREAD_RWLOCK_wrlock(&orig_hdl->obj_lock);
-
-		status = merge_share(&orig->share, &dupe->share);
-
-		PTHREAD_RWLOCK_unlock(&orig_hdl->obj_lock);
+		status = merge_share(orig_hdl, &orig->share, &dupe->share);
 	}
 
 	return status;
@@ -758,10 +756,10 @@ fsal_status_t rgw_fsal_open2(struct fsal_obj_handle *obj_hdl,
 			fsal_openflags_t openflags,
 			enum fsal_create_mode createmode,
 			const char *name,
-			struct attrlist *attrib_set,
+			struct fsal_attrlist *attrib_set,
 			fsal_verifier_t verifier,
 			struct fsal_obj_handle **new_obj,
-			struct attrlist *attrs_out,
+			struct fsal_attrlist *attrs_out,
 			bool *caller_perm_check)
 {
 	int posix_flags = 0;
@@ -772,7 +770,7 @@ fsal_status_t rgw_fsal_open2(struct fsal_obj_handle *obj_hdl,
 	bool truncated;
 	bool setattrs = attrib_set != NULL;
 	bool created = false;
-	struct attrlist verifier_attr;
+	struct fsal_attrlist verifier_attr;
 	struct rgw_open_state *open_state = NULL;
 	struct rgw_file_handle *rgw_fh;
 	struct rgw_handle *obj;
@@ -806,7 +804,7 @@ fsal_status_t rgw_fsal_open2(struct fsal_obj_handle *obj_hdl,
 			memset(&verifier_attr, 0, sizeof(verifier_attr));
 		}
 
-		set_common_verifier(attrib_set, verifier);
+		set_common_verifier(attrib_set, verifier, false);
 	}
 
 	if (!name) {
@@ -1024,8 +1022,8 @@ fsal_status_t rgw_fsal_open2(struct fsal_obj_handle *obj_hdl,
 
 	memset(&st, 0, sizeof(struct stat)); /* XXX needed? */
 
-	st.st_uid = op_ctx->creds->caller_uid;
-	st.st_gid = op_ctx->creds->caller_gid;
+	st.st_uid = op_ctx->creds.caller_uid;
+	st.st_gid = op_ctx->creds.caller_gid;
 	st.st_mode = unix_mode;
 
 	uint32_t create_mask =
@@ -1194,7 +1192,7 @@ fsal_status_t rgw_fsal_open2(struct fsal_obj_handle *obj_hdl,
  * @brief Return open status of a state.
  *
  * This function returns open flags representing the current open
- * status for a state. The state_lock must be held.
+ * status for a state. The st_lock must be held.
  *
  * @param[in] obj_hdl     File on which to operate
  * @param[in] state       File state to interrogate
@@ -1217,7 +1215,7 @@ fsal_openflags_t rgw_fsal_status2(struct fsal_obj_handle *obj_hdl,
  * @brief Re-open a file that may be already opened
  *
  * This function supports changing the access mode of a share reservation and
- * thus should only be called with a share state. The state_lock must be held.
+ * thus should only be called with a share state. The st_lock must be held.
  *
  * This MAY be used to open a file the first time if there is no need for
  * open by name or create semantics. One example would be 9P lopen.
@@ -1667,6 +1665,264 @@ static void handle_to_key(struct fsal_obj_handle *obj_hdl,
 	fh_desc->len = sizeof(struct rgw_fh_hk);
 }
 
+#ifdef USE_FSAL_RGW_XATTRS
+
+static int getxattr_cb(rgw_xattrlist *attrs, void *arg, uint32_t flags)
+{
+	xattrvalue4 *cb_arg = (xattrvalue4 *)arg;
+
+	cb_arg->utf8string_val =
+		gsh_strldup(attrs->xattrs->val.val, attrs->xattrs->val.len,
+		&cb_arg->utf8string_len);
+	cb_arg->utf8string_len = attrs->xattrs->val.len;
+	return 0;
+}
+
+/**
+ * @brief Get Extended Attribute
+ *
+ * The function fetches the extended attr of the object.
+ *
+ * @param[in]  obj_hdl  Input object to query
+ * @param[in]  xa_name  Input xattr name
+ * @param[out] xa_value Output xattr value
+ */
+
+static fsal_status_t getxattrs(struct fsal_obj_handle *obj_hdl,
+				xattrvalue4 *xa_name,
+				xattrvalue4 *xa_value)
+{
+	int rc = 0;
+	int errsv = 0;
+	fsal_status_t status;
+	rgw_xattrlist attrs = {NULL, 1};
+
+	struct rgw_export *export =
+		container_of(op_ctx->fsal_export,
+			     struct rgw_export, export);
+	struct rgw_handle *handle =
+		container_of(obj_hdl, struct rgw_handle, handle);
+
+	attrs.xattrs->key.val =
+		xa_name->utf8string_val;
+	attrs.xattrs->key.len = xa_name->utf8string_len;
+
+	rc = rgw_getxattrs(export->rgw_fs, handle->rgw_fh, &attrs,
+			   getxattr_cb, &xa_value,
+			   RGW_GETXATTR_FLAG_NONE);
+
+	if (rc < 0) {
+		errsv = errno;
+		LogDebug(COMPONENT_FSAL,
+			 "GETEXATTRS returned rc %d errsv %d",
+			 rc, errsv);
+
+		if (errsv == ERANGE) {
+			status = fsalstat(ERR_FSAL_TOOSMALL, 0);
+			goto out;
+		}
+		if (errsv == ENODATA) {
+			status = fsalstat(ERR_FSAL_NOENT, 0);
+			goto out;
+		}
+		status = fsalstat(posix2fsal_error(errsv), errsv);
+		goto out;
+	}
+
+	LogDebug(COMPONENT_FSAL,
+		 "GETXATTRS returned value %s length %d rc %d",
+		 xa_value->utf8string_val,
+		 xa_value->utf8string_len, rc);
+
+	status = fsalstat(ERR_FSAL_NO_ERROR, 0);
+
+out:
+	return status;
+}
+
+/**
+ * @brief Set Extended Attribute
+ *
+ * The function sets the extended attr of the object.
+ *
+ * @param[in]  obj_hdl  The input object
+ * @param[in]  sa_type  Input xattr type
+ * @param[in]  xa_name  Input xattr name
+ * @param[in]  xa_value Input xattr value to be set
+ */
+
+
+static fsal_status_t setxattrs(struct fsal_obj_handle *obj_hdl,
+				setxattr_option4 sa_type,
+				xattrvalue4 *xa_name,
+				xattrvalue4 *xa_value)
+{
+	int rc = 0;
+	int errsv = 0;
+	fsal_status_t status = {0, 0};
+	rgw_xattrlist attrs = {NULL, 1};
+
+	struct rgw_export *export =
+		container_of(op_ctx->fsal_export,
+			     struct rgw_export, export);
+	struct rgw_handle *handle = container_of(obj_hdl, struct rgw_handle,
+						 handle);
+
+	attrs.xattrs->key.val =
+		xa_name->utf8string_val;
+	attrs.xattrs->key.len = xa_name->utf8string_len;
+	attrs.xattrs->val.val =
+		xa_value->utf8string_val;
+	attrs.xattrs->val.len = xa_value->utf8string_len;
+
+	rc = rgw_setxattrs(export->rgw_fs, handle->rgw_fh, &attrs,
+			   RGW_SETXATTR_FLAG_NONE);
+
+	if (rc < 0) {
+		errsv = errno;
+		LogDebug(COMPONENT_FSAL,
+			 "SETXATTRS returned rc %d errsv %d",
+			 rc, errsv);
+		status = fsalstat(posix2fsal_error(errsv), errsv);
+		goto out;
+	}
+	status = fsalstat(ERR_FSAL_NO_ERROR, 0);
+
+out:
+	return status;
+}
+
+/**
+ * @brief Remove Extended Attribute
+ *
+ * This function removes an extended attribute of the object.
+ *
+ * @param[in]  obj_hdl  The input object
+ * @param[in]  xa_name  Input xattr name to be removed
+ *
+ * @return FSAL status.
+ */
+
+static fsal_status_t removexattrs(struct fsal_obj_handle *obj_hdl,
+				xattrvalue4 *xa_name)
+{
+	int rc = 0;
+	int errsv = 0;
+	fsal_status_t status = {0, 0};
+	rgw_xattrlist attrs = {NULL, 1};
+
+	struct rgw_export *export =
+		container_of(op_ctx->fsal_export,
+			     struct rgw_export, export);
+	struct rgw_handle *handle =
+		container_of(obj_hdl, struct rgw_handle, handle);
+
+	attrs.xattr_cnt = 1;
+	attrs.xattrs->key.val =
+		xa_name->utf8string_val;
+	attrs.xattrs->key.len = xa_name->utf8string_len;
+
+	rc = rgw_rmxattrs(export->rgw_fs,
+			  handle->rgw_fh,
+			  &attrs, RGW_RMXATTR_FLAG_NONE);
+
+	if (rc < 0) {
+		errsv = errno;
+		LogDebug(COMPONENT_FSAL,
+			 "REMOVEXATTRS returned rc %d errsv %d",
+			 rc, errsv);
+		status = fsalstat(posix2fsal_error(errsv), errsv);
+		goto out;
+	}
+	status = fsalstat(ERR_FSAL_NO_ERROR, 0);
+
+out:
+	return status;
+}
+
+struct lxattr_cb_arg {
+	count4 max;
+	xattrlist4 *names;
+};
+
+static int lsxattr_cb(rgw_xattrlist *attrs, void *cb_arg, uint32_t flag)
+{
+	struct lxattr_cb_arg *cb = (struct lxattr_cb_arg *)cb_arg;
+
+	cb->names->xl4_entries[cb->names->xl4_count].utf8string_val =
+		gsh_strldup(attrs->xattrs[0].key.val, attrs->xattrs[0].key.len,
+		&cb->names->xl4_entries[cb->names->xl4_count].utf8string_len);
+
+	cb->names->xl4_count++;
+
+	if (cb->names->xl4_count == cb->max) {
+		flag = RGW_LSXATTR_FLAG_STOP;
+		return flag;
+	}
+
+	return 0;
+}
+
+/**
+ * @brief List Extended Attribute
+ *
+ * The function fetches the list of extended attribute of the object.
+ *
+ * @param[in]		obj_hdl		The input object
+ * @param[in]		la_maxcount	Input max number of bytes for names
+ * @param[in,out]	la_cookie	In/out cookie
+ * @param[out]		lr_eof		Output eof set if no more xattrs
+ * @param[out]		lr_names	Output list of xattr names this buffer
+ *					size is double the size of la_maxcount
+ *					to allow for component4 overhead
+ *
+ * @return FSAL status.
+ */
+
+static fsal_status_t listxattrs(struct fsal_obj_handle *obj_hdl,
+				count4 la_maxcount,
+				nfs_cookie4 *la_cookie,
+				bool_t *lr_eof,
+				xattrlist4 *lr_names)
+{
+	int rc = 0;
+	int errsv = 0;
+	fsal_status_t status = {0, 0};
+	struct lxattr_cb_arg attrs;
+
+	attrs.names = lr_names;
+	attrs.max = la_maxcount;
+
+	struct rgw_export *export =
+		container_of(op_ctx->fsal_export,
+			     struct rgw_export, export);
+	struct rgw_handle *handle =
+		container_of(obj_hdl, struct rgw_handle, handle);
+
+	rc = rgw_lsxattrs(export->rgw_fs, handle->rgw_fh,
+			  NULL /* filter_prefix not yet implemented in RGW */,
+			  lsxattr_cb, &attrs, RGW_LSXATTR_FLAG_NONE);
+
+	if (rc < 0) {
+		errsv = errno;
+		LogDebug(COMPONENT_FSAL,
+			 "LISTXATTRS returned rc %d errsv %d",
+			 rc, errsv);
+		status = fsalstat(posix2fsal_error(errsv), errsv);
+		goto out;
+	}
+
+	/* Setting true unconditionally since cookie isn't implemented yet */
+	*lr_eof = true;
+
+	status = fsalstat(ERR_FSAL_NO_ERROR, 0);
+
+out:
+	return status;
+}
+
+#endif
+
 /**
  * @brief Override functions in ops vector
  *
@@ -1683,6 +1939,12 @@ void handle_ops_init(struct fsal_obj_ops *ops)
 	ops->release = release;
 	ops->merge = rgw_merge;
 	ops->lookup = lookup;
+#ifdef USE_FSAL_RGW_XATTRS
+	ops->getxattrs = getxattrs;
+	ops->setxattrs = setxattrs;
+	ops->removexattrs = removexattrs;
+	ops->listxattrs = listxattrs;
+#endif
 	ops->mkdir = rgw_fsal_mkdir;
 	ops->readdir = rgw_fsal_readdir;
 #if HAVE_DIRENT_OFFSETOF
