@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -265,18 +266,15 @@ fsal_status_t vfs_merge(struct fsal_obj_handle *orig_hdl,
 				    obj_handle);
 
 		/* This can block over an I/O operation. */
-		PTHREAD_RWLOCK_wrlock(&orig_hdl->obj_lock);
-
-		status = merge_share(&orig->u.file.share, &dupe->u.file.share);
-
-		PTHREAD_RWLOCK_unlock(&orig_hdl->obj_lock);
+		status = merge_share(orig_hdl, &orig->u.file.share,
+				     &dupe->u.file.share);
 	}
 
 	return status;
 }
 
 static fsal_status_t fetch_attrs(struct vfs_fsal_obj_handle *myself,
-				 int my_fd, struct attrlist *attrs)
+				 int my_fd, struct fsal_attrlist *attrs)
 {
 	struct stat stat;
 	int retval = 0;
@@ -338,6 +336,10 @@ static fsal_status_t fetch_attrs(struct vfs_fsal_obj_handle *myself,
 	}
 
 	posix2fsal_attributes_all(&stat, attrs);
+
+	/* Get correct fsid from the fsal_filesystem, it may
+	* not be the device major/minor from stat.
+	*/
 	attrs->fsid = myself->obj_handle.fs->fsid;
 
 	if (myself->sub_ops && myself->sub_ops->getattrs) {
@@ -359,9 +361,9 @@ static fsal_status_t vfs_open2_by_handle(struct fsal_obj_handle *obj_hdl,
 					 struct state_t *state,
 					 fsal_openflags_t openflags,
 					 enum fsal_create_mode createmode,
-					 struct attrlist *attrib_set,
+					 struct fsal_attrlist *attrib_set,
 					 fsal_verifier_t verifier,
-					 struct attrlist *attrs_out,
+					 struct fsal_attrlist *attrs_out,
 					 bool *caller_perm_check)
 {
 	fsal_status_t status = {0, 0};
@@ -467,7 +469,7 @@ static fsal_status_t vfs_open2_by_handle(struct fsal_obj_handle *obj_hdl,
 
 	if (createmode >= FSAL_EXCLUSIVE || truncated) {
 		/* Refresh the attributes */
-		struct attrlist attrs;
+		struct fsal_attrlist attrs;
 		attrmask_t attrs_mask = ATTR_ATIME | ATTR_MTIME;
 
 		if (attrs_out)
@@ -483,7 +485,9 @@ static fsal_status_t vfs_open2_by_handle(struct fsal_obj_handle *obj_hdl,
 
 			if (createmode >= FSAL_EXCLUSIVE &&
 			    createmode != FSAL_EXCLUSIVE_9P &&
-			    !check_verifier_attrlist(&attrs, verifier)) {
+			    !check_verifier_attrlist(&attrs, verifier,
+						     obj_hdl->fs->trunc_verif)
+			   ) {
 				/* Verifier didn't match, return EEXIST */
 				status = fsalstat(posix2fsal_error(EEXIST),
 						  EEXIST);
@@ -603,10 +607,10 @@ fsal_status_t vfs_open2(struct fsal_obj_handle *obj_hdl,
 			fsal_openflags_t openflags,
 			enum fsal_create_mode createmode,
 			const char *name,
-			struct attrlist *attrib_set,
+			struct fsal_attrlist *attrib_set,
 			fsal_verifier_t verifier,
 			struct fsal_obj_handle **new_obj,
-			struct attrlist *attrs_out,
+			struct fsal_attrlist *attrs_out,
 			bool *caller_perm_check)
 {
 	int posix_flags = 0;
@@ -634,7 +638,8 @@ fsal_status_t vfs_open2(struct fsal_obj_handle *obj_hdl,
 
 	if (createmode >= FSAL_EXCLUSIVE) {
 		/* Now fixup attrs for verifier if exclusive create */
-		set_common_verifier(attrib_set, verifier);
+		set_common_verifier(attrib_set, verifier,
+				    obj_hdl->fs->trunc_verif);
 	}
 
 	if (name == NULL) {
@@ -653,11 +658,11 @@ fsal_status_t vfs_open2(struct fsal_obj_handle *obj_hdl,
 
 #ifdef ENABLE_VFS_DEBUG_ACL
 	if (createmode != FSAL_NO_CREATE) {
-		/* Need to ammend attributes for inherited ACL, these will be
+		/* Need to amend attributes for inherited ACL, these will be
 		 * set later. We also need to test for permission to create
 		 * since there might be an ACL.
 		 */
-		struct attrlist attrs;
+		struct fsal_attrlist attrs;
 		fsal_accessflags_t access_type;
 
 		access_type = FSAL_MODE_MASK_SET(FSAL_W_OK) |
@@ -729,7 +734,7 @@ fsal_status_t vfs_open2(struct fsal_obj_handle *obj_hdl,
 	/* Become the user because we are creating an object in this dir.
 	 */
 	if (createmode != FSAL_NO_CREATE)
-		if (!vfs_set_credentials(op_ctx->creds, obj_hdl->fsal)) {
+		if (!vfs_set_credentials(&op_ctx->creds, obj_hdl->fsal)) {
 			status = posix2fsal_status(EPERM);
 			goto direrr;
 		}
@@ -889,6 +894,9 @@ fsal_status_t vfs_open2(struct fsal_obj_handle *obj_hdl,
 	*new_obj = &hdl->obj_handle;
 
 	if (created && attrib_set->valid_mask != 0) {
+
+retry_attr:
+
 		/* Set attributes using our newly opened file descriptor as the
 		 * share_fd if there are any left to set (mode and truncate
 		 * have already been handled).
@@ -899,17 +907,15 @@ fsal_status_t vfs_open2(struct fsal_obj_handle *obj_hdl,
 		 * Note if we have ENABLE_VFS_DEBUG_ACL an inherited ACL might
 		 * be part of the attributes we are setting here.
 		 */
-		status = (*new_obj)->obj_ops->setattr2(*new_obj,
-						      false,
-						      state,
-						      attrib_set);
+		status = (*new_obj)->obj_ops->setattr2(*new_obj, false,
+						       state, attrib_set);
 
 		if (FSAL_IS_ERROR(status))
 			goto fileerr;
 
 		if (attrs_out != NULL) {
 			status = (*new_obj)->obj_ops->getattrs(*new_obj,
-							      attrs_out);
+							       attrs_out);
 			if (FSAL_IS_ERROR(status) &&
 			    (attrs_out->request_mask & ATTR_RDATTR_ERR) == 0) {
 				/* Get attributes failed and caller expected
@@ -918,6 +924,29 @@ fsal_status_t vfs_open2(struct fsal_obj_handle *obj_hdl,
 				 */
 				goto fileerr;
 			}
+
+			LogFullDebug(COMPONENT_FSAL,
+				     "Set atime %llx %llx mtime %llx %llx",
+				     (long long) attrs_out->atime.tv_sec,
+				     (long long) attrs_out->atime.tv_nsec,
+				     (long long) attrs_out->mtime.tv_sec,
+				     (long long) attrs_out->mtime.tv_nsec);
+
+			if ((createmode >= FSAL_EXCLUSIVE) &&
+			    (!(*new_obj)->fs->trunc_verif) &&
+			    ((attrs_out->atime.tv_sec !=
+						attrib_set->atime.tv_sec) ||
+			     (attrs_out->mtime.tv_sec !=
+						attrib_set->mtime.tv_sec))) {
+				LogInfo(COMPONENT_FSAL,
+					"Verifier was not stored correctly for filesystem %s, trying again with truncated verifier",
+					(*new_obj)->fs->path);
+				(*new_obj)->fs->trunc_verif = true;
+				FSAL_UNSET_MASK(attrib_set->valid_mask,
+						ATTR_ATIME | ATTR_MTIME);
+				set_common_verifier(attrib_set, verifier, true);
+				goto retry_attr;
+			}
 		}
 	} else if (attrs_out != NULL) {
 		/* Since we haven't set any attributes other than what was set
@@ -925,6 +954,10 @@ fsal_status_t vfs_open2(struct fsal_obj_handle *obj_hdl,
 		 * we used to create the fsal_obj_handle.
 		 */
 		posix2fsal_attributes_all(&stat, attrs_out);
+
+		/* Get correct fsid from the fsal_filesystem, it may
+		* not be the device major/minor from stat.
+		*/
 		attrs_out->fsid = myself->obj_handle.fs->fsid;
 	}
 
@@ -979,7 +1012,7 @@ fsal_status_t vfs_open2(struct fsal_obj_handle *obj_hdl,
  * @brief Re-open a file that may be already opened
  *
  * This function supports changing the access mode of a share reservation and
- * thus should only be called with a share state. The state_lock must be held.
+ * thus should only be called with a share state. The st_lock must be held.
  *
  * This MAY be used to open a file the first time if there is no need for
  * open by name or create semantics. One example would be 9P lopen.
@@ -1084,7 +1117,6 @@ fsal_status_t find_fd(int *fd,
 		      bool open_for_locks)
 {
 	struct vfs_fsal_obj_handle *myself;
-	struct vfs_filesystem *vfs_fs;
 	struct vfs_fd temp_fd = {
 			FSAL_O_CLOSED, PTHREAD_RWLOCK_INITIALIZER, -1 };
 	struct vfs_fd *out_fd = &temp_fd;
@@ -1093,7 +1125,6 @@ fsal_status_t find_fd(int *fd,
 	bool reusing_open_state_fd = false;
 
 	myself = container_of(obj_hdl, struct vfs_fsal_obj_handle, obj_handle);
-	vfs_fs = myself->obj_handle.fs->private_data;
 
 	fsal2posix_openflags(openflags, &posix_flags);
 
@@ -1102,7 +1133,7 @@ fsal_status_t find_fd(int *fd,
 	case SOCKET_FILE:
 	case CHARACTER_FILE:
 	case BLOCK_FILE:
-		rc = vfs_open_by_handle(vfs_fs,
+		rc = vfs_open_by_handle(myself->obj_handle.fs,
 					myself->u.unopenable.dir,
 					O_PATH | O_NOACCESS,
 					&status.major);
@@ -1321,13 +1352,6 @@ void vfs_write2(struct fsal_obj_handle *obj_hdl,
 	fsal_openflags_t openflags = FSAL_O_WRITE;
 	struct vfs_fd *vfs_fd = NULL;
 
-	if (write_arg->info != NULL) {
-		/* Currently we don't support WRITE_PLUS */
-		done_cb(obj_hdl, fsalstat(ERR_FSAL_NOTSUPP, 0), write_arg,
-			caller_arg);
-		return;
-	}
-
 	if (obj_hdl->fsal != obj_hdl->fs->fsal) {
 		LogDebug(COMPONENT_FSAL,
 			 "FSAL %s operation for handle belonging to FSAL %s, return EXDEV",
@@ -1359,7 +1383,7 @@ void vfs_write2(struct fsal_obj_handle *obj_hdl,
 		goto out;
 	}
 
-	if (!vfs_set_credentials(op_ctx->creds, obj_hdl->fsal)) {
+	if (!vfs_set_credentials(&op_ctx->creds, obj_hdl->fsal)) {
 		retval = EPERM;
 		status = fsalstat(ERR_FSAL_PERM, EPERM);
 		goto out;
@@ -1403,6 +1427,168 @@ void vfs_write2(struct fsal_obj_handle *obj_hdl,
 	done_cb(obj_hdl, status, write_arg, caller_arg);
 }
 
+/**
+ * @brief Seek to data or hole
+ *
+ * This function seek to data or hole in a file.
+ *
+ * @param[in]     obj_hdl   File on which to operate
+ * @param[in]     state     state_t to use for this operation
+ * @param[in,out] info      Information about the data
+ *
+ * @return FSAL status.
+ */
+
+#ifdef __USE_GNU
+fsal_status_t vfs_seek2(struct fsal_obj_handle *obj_hdl,
+			struct state_t *state,
+			struct io_info *info)
+{
+	struct vfs_fsal_obj_handle *myself;
+	off_t ret = 0, offset = info->io_content.hole.di_offset;
+	int what = 0;
+	bool has_lock = false;
+	bool closefd = false;
+	fsal_openflags_t openflags = FSAL_O_ANY;
+	fsal_status_t status = { ERR_FSAL_NO_ERROR, 0 };
+	int my_fd = -1;
+	struct fsal_attrlist attrs;
+
+	myself = container_of(obj_hdl, struct vfs_fsal_obj_handle, obj_handle);
+
+	/* Get a usable file descriptor */
+	status = find_fd(&my_fd, obj_hdl, false, state, openflags,
+		&has_lock, &closefd, false);
+
+	if (FSAL_IS_ERROR(status))
+		goto out;
+
+	fsal_prepare_attrs(&attrs,
+			   (op_ctx->fsal_export->exp_ops.fs_supported_attrs(
+							op_ctx->fsal_export)
+				& ~(ATTR_ACL | ATTR4_FS_LOCATIONS)));
+
+	status = fetch_attrs(myself, my_fd, &attrs);
+
+	fsal_release_attrs(&attrs);
+
+	if (FSAL_IS_ERROR(status)) {
+		goto out;
+	}
+
+	/* RFC7862 15.11.3,
+	 * If the sa_offset is beyond the end of the file,
+	 * then SEEK MUST return NFS4ERR_NXIO. */
+	if (offset >= attrs.filesize) {
+		status = posix2fsal_status(ENXIO);
+		goto out;
+	}
+
+	if (info->io_content.what == NFS4_CONTENT_DATA) {
+		what = SEEK_DATA;
+	} else if (info->io_content.what == NFS4_CONTENT_HOLE) {
+		what = SEEK_HOLE;
+	} else {
+		status = fsalstat(ERR_FSAL_UNION_NOTSUPP, 0);
+		goto out;
+	}
+
+	ret = lseek(my_fd, offset, what);
+
+	if (ret < 0) {
+		if (errno == ENXIO) {
+			info->io_eof = TRUE;
+		} else {
+			status = posix2fsal_status(errno);
+		}
+		goto out;
+	} else {
+		info->io_eof = (ret >= attrs.filesize);
+		info->io_content.hole.di_offset = ret;
+	}
+
+ out:
+
+	if (closefd) {
+		LogFullDebug(COMPONENT_FSAL,
+			     "Closing Opened fd %d", my_fd);
+		close(my_fd);
+	}
+
+	if (has_lock)
+		PTHREAD_RWLOCK_unlock(&obj_hdl->obj_lock);
+
+	return status;
+}
+#endif
+
+/**
+ * @brief Reserve/Deallocate space in a region of a file
+ *
+ * @param[in] obj_hdl File to which bytes should be allocated
+ * @param[in] state   open stateid under which to do the allocation
+ * @param[in] offset  offset at which to begin the allocation
+ * @param[in] length  length of the data to be allocated
+ * @param[in] allocate Should space be allocated or deallocated?
+ *
+ * @return FSAL status.
+ */
+
+#ifdef FALLOC_FL_PUNCH_HOLE
+fsal_status_t vfs_fallocate(struct fsal_obj_handle *obj_hdl,
+			    struct state_t *state, uint64_t offset,
+			    uint64_t length, bool allocate)
+{
+	int ret = 0;
+	bool has_lock = false;
+	bool closefd = false;
+	fsal_openflags_t openflags = FSAL_O_WRITE;
+	fsal_status_t status = { ERR_FSAL_NO_ERROR, 0 };
+	int my_fd = -1;
+
+	/* Get a usable file descriptor */
+	status = find_fd(&my_fd, obj_hdl, false, state, openflags,
+		&has_lock, &closefd, false);
+
+	if (FSAL_IS_ERROR(status))
+		goto out;
+
+	if (!vfs_set_credentials(&op_ctx->creds, obj_hdl->fsal)) {
+		status = posix2fsal_status(EPERM);
+		goto out;
+	}
+
+	ret = fallocate(my_fd,
+			allocate
+				? 0
+				: FALLOC_FL_KEEP_SIZE | FALLOC_FL_PUNCH_HOLE,
+			offset, length);
+
+	if (ret < 0) {
+		ret = errno;
+		LogFullDebug(COMPONENT_FSAL,
+			     "fallocate returned %s (%d)",
+			     strerror(ret), ret);
+		status = posix2fsal_status(ret);
+	}
+
+	vfs_restore_ganesha_credentials(obj_hdl->fsal);
+
+ out:
+
+	if (closefd) {
+		LogFullDebug(COMPONENT_FSAL,
+			     "Closing Opened fd %d", my_fd);
+		close(my_fd);
+	}
+
+	if (has_lock)
+		PTHREAD_RWLOCK_unlock(&obj_hdl->obj_lock);
+
+	return status;
+}
+#endif
+
 /**
  * @brief Commit written data
  *
@@ -1447,7 +1633,7 @@ fsal_status_t vfs_commit2(struct fsal_obj_handle *obj_hdl,
 
 	if (!FSAL_IS_ERROR(status)) {
 
-		if (!vfs_set_credentials(op_ctx->creds, obj_hdl->fsal)) {
+		if (!vfs_set_credentials(&op_ctx->creds, obj_hdl->fsal)) {
 			retval = EPERM;
 			status = fsalstat(ERR_FSAL_PERM, EPERM);
 			goto out;
@@ -1685,7 +1871,7 @@ fsal_status_t vfs_lock_op2(struct fsal_obj_handle *obj_hdl,
  */
 
 fsal_status_t vfs_getattr2(struct fsal_obj_handle *obj_hdl,
-			   struct attrlist *attrs)
+			   struct fsal_attrlist *attrs)
 {
 	struct vfs_fsal_obj_handle *myself;
 	fsal_status_t status = {0, 0};
@@ -1771,7 +1957,7 @@ fsal_status_t vfs_getattr2(struct fsal_obj_handle *obj_hdl,
 fsal_status_t vfs_setattr2(struct fsal_obj_handle *obj_hdl,
 			   bool bypass,
 			   struct state_t *state,
-			   struct attrlist *attrib_set)
+			   struct fsal_attrlist *attrib_set)
 {
 	struct vfs_fsal_obj_handle *myself;
 	fsal_status_t status = {0, 0};
@@ -1805,7 +1991,7 @@ fsal_status_t vfs_setattr2(struct fsal_obj_handle *obj_hdl,
 	if (FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_MODE) &&
 	    !FSAL_TEST_MASK(attrib_set->valid_mask, ATTR_ACL)) {
 		/* Set ACL from MODE */
-		struct attrlist attrs;
+		struct fsal_attrlist attrs;
 
 		fsal_prepare_attrs(&attrs, ATTR_ACL);
 
@@ -1999,6 +2185,12 @@ fsal_status_t vfs_setattr2(struct fsal_obj_handle *obj_hdl,
 			timebuf[1].tv_sec = 0;
 			timebuf[1].tv_nsec = UTIME_OMIT;
 		}
+
+		LogFullDebug(COMPONENT_FSAL,
+			     "Setting atime %lx %lx mtime %lx %lx",
+			     timebuf[0].tv_sec, timebuf[0].tv_nsec,
+			     timebuf[1].tv_sec, timebuf[1].tv_nsec);
+
 		if (vfs_unopenable_type(obj_hdl->type))
 			retval = vfs_utimesat(my_fd, myself->u.unopenable.name,
 					      timebuf, AT_SYMLINK_NOFOLLOW);
