@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -44,6 +45,7 @@
 #include "fsal_convert.h"
 #include "fsal_handle_syscalls.h"
 #include "FSAL/fsal_commonlib.h"
+#include "FSAL/fsal_localfs.h"
 #include "vfs_methods.h"
 #include <os/subr.h>
 #include "subfsal.h"
@@ -58,9 +60,8 @@ int vfs_fsal_open(struct vfs_fsal_obj_handle *hdl,
 		  int openflags,
 		  fsal_errors_t *fsal_error)
 {
-	struct vfs_filesystem *vfs_fs = hdl->obj_handle.fs->private_data;
-
-	return vfs_open_by_handle(vfs_fs, hdl->handle, openflags, fsal_error);
+	return vfs_open_by_handle(hdl->obj_handle.fs, hdl->handle,
+				  openflags, fsal_error);
 }
 
 /**
@@ -146,8 +147,9 @@ struct vfs_fsal_obj_handle *alloc_handle(int dirfd,
 	hdl->obj_handle.fs = fs;
 
 	LogFullDebug(COMPONENT_FSAL,
-		     "Creating object %p for file %s of type %s",
-		     hdl, path, object_file_type_to_str(hdl->obj_handle.type));
+		     "Creating object %p for file %s of type %s on filesystem %p %s",
+		     hdl, path, object_file_type_to_str(hdl->obj_handle.type),
+		     fs, fs->path);
 
 	if (hdl->obj_handle.type == REGULAR_FILE) {
 		hdl->u.file.fd.fd = -1;	/* no open on this yet */
@@ -203,7 +205,7 @@ struct vfs_fsal_obj_handle *alloc_handle(int dirfd,
 }
 
 static fsal_status_t populate_fs_locations(struct vfs_fsal_obj_handle *hdl,
-					   struct attrlist *attrs_out)
+					   struct fsal_attrlist *attrs_out)
 {
 	fsal_status_t status;
 	uint64 hash;
@@ -219,7 +221,7 @@ static fsal_status_t populate_fs_locations(struct vfs_fsal_obj_handle *hdl,
 
 	if (FSAL_TEST_MASK(attrs_out->valid_mask, ATTR4_FS_LOCATIONS)) {
 		/* on a referral the filesystem id has to change
-		 * it get's calculated via a hash from the referral
+		 * it gets calculated via a hash from the referral
 		 * and stored in the fsid object of the fsal_obj_handle
 		 */
 
@@ -258,8 +260,8 @@ static fsal_status_t check_filesystem(struct vfs_fsal_obj_handle *parent_hdl,
 	fsal_dev_t dev;
 	struct fsal_filesystem *fs = NULL;
 	fsal_status_t status = { ERR_FSAL_NO_ERROR, 0 };
-	struct vfs_fsal_export *myexp_hdl =
-	    container_of(op_ctx->fsal_export, struct vfs_fsal_export, export);
+
+again:
 
 	retval = fstatat(dirfd, path, stat, AT_SYMLINK_NOFOLLOW);
 
@@ -267,6 +269,17 @@ static fsal_status_t check_filesystem(struct vfs_fsal_obj_handle *parent_hdl,
 		retval = errno;
 		LogDebug(COMPONENT_FSAL, "Failed to open stat %s: %s", path,
 			 msg_fsal_err(posix2fsal_error(retval)));
+
+		if (errno == EAGAIN) {
+			/* autofs can cause EAGAIN if the file system is not
+			 * yet mounted.
+			 */
+			/** @todo FSF: should we retry forever
+			 *             is there any other reason for EAGAIN?
+			 */
+			goto again;
+		}
+
 		status = posix2fsal_status(retval);
 		goto out;
 	}
@@ -275,6 +288,8 @@ static fsal_status_t check_filesystem(struct vfs_fsal_obj_handle *parent_hdl,
 
 	fs = parent_hdl->obj_handle.fs;
 
+	LogFilesystem("Parent", "", fs);
+
 	if ((dev.minor == parent_hdl->dev.minor) &&
 	    (dev.major == parent_hdl->dev.major)) {
 		/* Filesystem is ok */
@@ -290,16 +305,12 @@ static fsal_status_t check_filesystem(struct vfs_fsal_obj_handle *parent_hdl,
 			PRIu64".%"PRIu64" - reloading filesystems to find it",
 			path, dev.major, dev.minor);
 
-		retval = reload_posix_filesystems(op_ctx->ctx_export->fullpath,
-						  parent_hdl->obj_handle.fsal,
-						  op_ctx->fsal_export,
-						  vfs_claim_filesystem,
-						  vfs_unclaim_filesystem,
-						  &myexp_hdl->root_fs);
+		retval = populate_posix_file_systems(path);
 
 		if (retval != 0) {
-			LogFullDebug(COMPONENT_FSAL,
-				     "resolve_posix_filesystem failed");
+			LogCrit(COMPONENT_FSAL,
+				"populate_posix_file_systems returned %s (%d)",
+				strerror(retval), retval);
 			status = posix2fsal_status(EXDEV);
 			goto out;
 		}
@@ -312,13 +323,20 @@ static fsal_status_t check_filesystem(struct vfs_fsal_obj_handle *parent_hdl,
 			status = posix2fsal_status(EXDEV);
 			goto out;
 		} else {
+			/* We have now found the file system, since it was
+			 * just added, fs->fsal will be NULL so the next if
+			 * block below will be executed to claim the file system
+			 * and carry on.
+			 */
 			LogInfo(COMPONENT_FSAL,
-				"Filesystem %s has been added to export %d:%s",
+				"Filesystem %s will be added to export %d:%s",
 				fs->path, op_ctx->ctx_export->export_id,
-				op_ctx_export_path(op_ctx->ctx_export));
+				op_ctx_export_path(op_ctx));
 		}
 	}
 
+	LogFilesystem("XDEV", "", fs);
+
 	if (fs->fsal == NULL) {
 		/* The filesystem wasn't claimed, it must have been added after
 		 * we created this export. Go ahead and try to get it claimed.
@@ -327,12 +345,13 @@ static fsal_status_t check_filesystem(struct vfs_fsal_obj_handle *parent_hdl,
 			"Lookup of %s crosses filesystem boundary to unclaimed file system %s - attempt to claim it",
 			path, fs->path);
 
-		retval = claim_posix_filesystems(op_ctx->ctx_export->fullpath,
+		retval = claim_posix_filesystems(CTX_FULLPATH(op_ctx),
 						 parent_hdl->obj_handle.fsal,
 						 op_ctx->fsal_export,
 						 vfs_claim_filesystem,
 						 vfs_unclaim_filesystem,
-						 &myexp_hdl->root_fs);
+						 &op_ctx->fsal_export->root_fs,
+						 stat);
 
 		if (retval != 0) {
 			LogFullDebug(COMPONENT_FSAL,
@@ -367,7 +386,7 @@ out:
 static fsal_status_t lookup_with_fd(struct vfs_fsal_obj_handle *parent_hdl,
 				    int dirfd, const char *path,
 				    struct fsal_obj_handle **handle,
-				    struct attrlist *attrs_out)
+				    struct fsal_attrlist *attrs_out)
 {
 	struct vfs_fsal_obj_handle *hdl;
 	int retval;
@@ -434,6 +453,11 @@ static fsal_status_t lookup_with_fd(struct vfs_fsal_obj_handle *parent_hdl,
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes_all(&stat, attrs_out);
+
+		/* Get correct fsid from the fsal_filesystem, it may not be
+		 * the device major/minor from stat.
+		 */
+		attrs_out->fsid = hdl->obj_handle.fs->fsid;
 	}
 
 	hdl->obj_handle.fsid = hdl->obj_handle.fs->fsid;
@@ -470,7 +494,7 @@ static fsal_status_t lookup_with_fd(struct vfs_fsal_obj_handle *parent_hdl,
 
 static fsal_status_t lookup(struct fsal_obj_handle *parent,
 			    const char *path, struct fsal_obj_handle **handle,
-			    struct attrlist *attrs_out)
+			    struct fsal_attrlist *attrs_out)
 {
 	struct vfs_fsal_obj_handle *parent_hdl;
 	fsal_errors_t fsal_error = ERR_FSAL_NO_ERROR;
@@ -486,6 +510,8 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 		return fsalstat(ERR_FSAL_NOTDIR, 0);
 	}
 
+	LogFilesystem("About to check FSALs", "", parent->fs);
+
 	if (parent->fsal != parent->fs->fsal) {
 		LogDebug(COMPONENT_FSAL,
 			 "FSAL %s operation for handle belonging to FSAL %s, return EXDEV",
@@ -513,9 +539,9 @@ static fsal_status_t lookup(struct fsal_obj_handle *parent,
 }
 
 static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
-			     const char *name, struct attrlist *attrib,
+			     const char *name, struct fsal_attrlist *attrib,
 			     struct fsal_obj_handle **handle,
-			     struct attrlist *attrs_out)
+			     struct fsal_attrlist *attrs_out)
 {
 	struct vfs_fsal_obj_handle *myself, *hdl;
 	int dir_fd;
@@ -525,7 +551,7 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 	int retval = 0;
 	int flags = O_PATH | O_NOACCESS;
 #ifdef ENABLE_VFS_DEBUG_ACL
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 	fsal_accessflags_t access_type;
 #endif /* ENABLE_VFS_DEBUG_ACL */
 	vfs_file_handle_t *fh = NULL;
@@ -601,7 +627,7 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 	/* Become the user because we are creating an object in this dir.
 	 */
 
-	if (!vfs_set_credentials(op_ctx->creds, dir_hdl->fsal)) {
+	if (!vfs_set_credentials(&op_ctx->creds, dir_hdl->fsal)) {
 		retval = EPERM;
 		status = posix2fsal_status(retval);
 		goto direrr;
@@ -685,6 +711,11 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 			 * to create the fsal_obj_handle.
 			 */
 			posix2fsal_attributes_all(&stat, attrs_out);
+
+			/* Get correct fsid from the fsal_filesystem, it may
+			 * not be the device major/minor from stat.
+			 */
+			attrs_out->fsid = hdl->obj_handle.fs->fsid;
 		}
 	}
 
@@ -704,9 +735,9 @@ static fsal_status_t makedir(struct fsal_obj_handle *dir_hdl,
 static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 			      const char *name,
 			      object_file_type_t nodetype,	/* IN */
-			      struct attrlist *attrib,
+			      struct fsal_attrlist *attrib,
 			      struct fsal_obj_handle **handle,
-			      struct attrlist *attrs_out)
+			      struct fsal_attrlist *attrs_out)
 {
 	struct vfs_fsal_obj_handle *myself, *hdl;
 	int dir_fd = -1;
@@ -717,7 +748,7 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 	dev_t unix_dev = 0;
 	int flags = O_PATH | O_NOACCESS;
 #ifdef ENABLE_VFS_DEBUG_ACL
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 	fsal_accessflags_t access_type;
 #endif /* ENABLE_VFS_DEBUG_ACL */
 	vfs_file_handle_t *fh = NULL;
@@ -813,7 +844,7 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 		goto direrr;
 	}
 
-	if (!vfs_set_credentials(op_ctx->creds, dir_hdl->fsal)) {
+	if (!vfs_set_credentials(&op_ctx->creds, dir_hdl->fsal)) {
 		retval = EPERM;
 		status = posix2fsal_status(retval);
 		goto direrr;
@@ -893,6 +924,11 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 			 * to create the fsal_obj_handle.
 			 */
 			posix2fsal_attributes_all(&stat, attrs_out);
+
+			/* Get correct fsid from the fsal_filesystem, it may
+			 * not be the device major/minor from stat.
+			 */
+			attrs_out->fsid = hdl->obj_handle.fs->fsid;
 		}
 	}
 
@@ -921,9 +957,9 @@ static fsal_status_t makenode(struct fsal_obj_handle *dir_hdl,
 
 static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 				 const char *name, const char *link_path,
-				 struct attrlist *attrib,
+				 struct fsal_attrlist *attrib,
 				 struct fsal_obj_handle **handle,
-				 struct attrlist *attrs_out)
+				 struct fsal_attrlist *attrs_out)
 {
 	struct vfs_fsal_obj_handle *myself, *hdl;
 	int dir_fd = -1;
@@ -932,7 +968,7 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 	int retval = 0;
 	int flags = O_PATH | O_NOACCESS;
 #ifdef ENABLE_VFS_DEBUG_ACL
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 	fsal_accessflags_t access_type;
 #endif /* ENABLE_VFS_DEBUG_ACL */
 	vfs_file_handle_t *fh = NULL;
@@ -1003,7 +1039,7 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 
 	/* Become the user because we are creating an object in this dir.
 	 */
-	if (!vfs_set_credentials(op_ctx->creds, dir_hdl->fsal)) {
+	if (!vfs_set_credentials(&op_ctx->creds, dir_hdl->fsal)) {
 		retval = EPERM;
 		status = posix2fsal_status(retval);
 		goto direrr;
@@ -1083,6 +1119,11 @@ static fsal_status_t makesymlink(struct fsal_obj_handle *dir_hdl,
 			 * to create the fsal_obj_handle.
 			 */
 			posix2fsal_attributes_all(&stat, attrs_out);
+
+			/* Get correct fsid from the fsal_filesystem, it may
+			 * not be the device major/minor from stat.
+			 */
+			attrs_out->fsid = hdl->obj_handle.fs->fsid;
 		}
 	}
 
@@ -1343,7 +1384,7 @@ static fsal_status_t read_dirents(struct fsal_obj_handle *dir_hdl,
 #endif
 		for (bpos = 0; bpos < nread;) {
 			struct fsal_obj_handle *hdl;
-			struct attrlist attrs;
+			struct fsal_attrlist attrs;
 			enum fsal_dir_result cb_rc;
 
 			if (!to_vfs_dirent(buf, bpos, dentryp, baseloc))
@@ -1380,7 +1421,7 @@ static fsal_status_t read_dirents(struct fsal_obj_handle *dir_hdl,
 				goto done;
 			}
 
-			/* callback to cache inode */
+			/* callback to MDCACHE */
 			cb_rc = cb(dentryp->vd_name, hdl, &attrs, dir_state,
 				(fsal_cookie_t) dentryp->vd_offset);
 
@@ -1454,7 +1495,7 @@ static fsal_status_t renamefile(struct fsal_obj_handle *obj_hdl,
 	/* Become the user because we are creating/removing objects
 	 * in these dirs which messes with quotas and perms.
 	 */
-	if (!vfs_set_credentials(op_ctx->creds, obj_hdl->fsal)) {
+	if (!vfs_set_credentials(&op_ctx->creds, obj_hdl->fsal)) {
 		retval = EPERM;
 		fsal_error = posix2fsal_error(retval);
 		goto out;
@@ -1516,7 +1557,6 @@ struct closefd vfs_fsal_open_and_stat(struct fsal_export *exp,
 	struct closefd cfd = { .fd = -1, .close_fd = false };
 	int retval = 0;
 	const char *func = "unknown";
-	struct vfs_filesystem *vfs_fs = myself->obj_handle.fs->private_data;
 	int open_flags;
 
 	fsal2posix_openflags(flags, &open_flags);
@@ -1525,7 +1565,7 @@ struct closefd vfs_fsal_open_and_stat(struct fsal_export *exp,
 	case SOCKET_FILE:
 	case CHARACTER_FILE:
 	case BLOCK_FILE:
-		cfd.fd = vfs_open_by_handle(vfs_fs,
+		cfd.fd = vfs_open_by_handle(myself->obj_handle.fs,
 					    myself->u.unopenable.dir,
 					    O_PATH | O_NOACCESS,
 					    fsal_error);
@@ -1657,7 +1697,7 @@ static fsal_status_t file_unlink(struct fsal_obj_handle *dir_hdl,
 		goto errout;
 	}
 
-	if (!vfs_set_credentials(op_ctx->creds, dir_hdl->fsal)) {
+	if (!vfs_set_credentials(&op_ctx->creds, dir_hdl->fsal)) {
 		retval = EPERM;
 		fsal_error = posix2fsal_error(retval);
 		goto errout;
@@ -1726,8 +1766,9 @@ static fsal_status_t handle_to_wire(const struct fsal_obj_handle *obj_hdl,
 	return fsalstat(ERR_FSAL_NO_ERROR, 0);
 }
 
-/*
- * release
+/**
+ * @brief release object handle
+ *
  * release our export first so they know we are gone
  */
 
@@ -1820,7 +1861,7 @@ void vfs_handle_ops_init(struct fsal_obj_ops *ops)
 
 fsal_status_t vfs_lookup_path(struct fsal_export *exp_hdl,
 			      const char *path, struct fsal_obj_handle **handle,
-			      struct attrlist *attrs_out)
+			      struct fsal_attrlist *attrs_out)
 {
 	int dir_fd = -1;
 	struct stat stat;
@@ -1893,6 +1934,11 @@ fsal_status_t vfs_lookup_path(struct fsal_export *exp_hdl,
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes_all(&stat, attrs_out);
+
+		/* Get correct fsid from the fsal_filesystem, it may
+		* not be the device major/minor from stat.
+		*/
+		attrs_out->fsid = hdl->obj_handle.fs->fsid;
 	}
 
 	/* if it is a directory and the sticky bit is set
@@ -1936,6 +1982,7 @@ fsal_status_t vfs_check_handle(struct fsal_export *exp_hdl,
 	int retval = 0;
 	struct fsal_fsid__ fsid;
 	enum fsid_type fsid_type;
+	bool fslocked = false;
 
 	*fs = NULL;
 
@@ -1949,8 +1996,27 @@ fsal_status_t vfs_check_handle(struct fsal_export *exp_hdl,
 
 	retval = vfs_extract_fsid(fh, &fsid_type, &fsid);
 
+
 	if (retval == 0) {
-		*fs = lookup_fsid(&fsid, fsid_type);
+		/* Since the root_fs is the most likely one, and it can't
+		 * change, we can check without locking.
+		 */
+		if (fsal_fs_compare_fsid(fsid_type,
+					 &fsid,
+					 exp_hdl->root_fs->fsid_type,
+					 &exp_hdl->root_fs->fsid) == 0) {
+			/* This is the root_fs of the export, all good. */
+			*fs = exp_hdl->root_fs;
+		} else {
+			/* Must lookup fs and check that it is exported by
+			 * the export.
+			 */
+			PTHREAD_RWLOCK_rdlock(&fs_lock);
+			fslocked = true;
+
+			*fs = lookup_fsid_locked(&fsid, fsid_type);
+		}
+
 		if (*fs == NULL) {
 			LogInfo(COMPONENT_FSAL,
 				"Could not map fsid=0x%016"PRIx64
@@ -1960,6 +2026,7 @@ fsal_status_t vfs_check_handle(struct fsal_export *exp_hdl,
 			fsal_error = posix2fsal_error(retval);
 			goto errout;
 		}
+
 		if (((*fs)->fsal != exp_hdl->fsal) && !(*dummy)) {
 			LogInfo(COMPONENT_FSAL,
 				"fsid=0x%016"PRIx64".0x%016"PRIx64
@@ -1971,6 +2038,17 @@ fsal_status_t vfs_check_handle(struct fsal_export *exp_hdl,
 			goto errout;
 		}
 
+		/* If we had to lookup fs, then we must check that it is
+		 * exported by the export.
+		 */
+		if (fslocked && !(*dummy) &&
+		    !is_filesystem_exported(*fs, exp_hdl)) {
+			/* We've got a handle with a spoofed fsid/export_id */
+			retval = ESTALE;
+			fsal_error = posix2fsal_error(retval);
+			goto errout;
+		}
+
 		LogDebug(COMPONENT_FSAL,
 			 "Found filesystem %s for handle for FSAL %s",
 			 (*fs)->path,
@@ -1983,12 +2061,14 @@ fsal_status_t vfs_check_handle(struct fsal_export *exp_hdl,
 	}
 
  errout:
+	if (fslocked)
+		PTHREAD_RWLOCK_unlock(&fs_lock);
 	return fsalstat(fsal_error, retval);
 }
 
 /* create_handle
  * Does what original FSAL_ExpandHandle did (sort of)
- * returns a ref counted handle to be later used in cache_inode etc.
+ * returns a ref counted handle to be later used in mdcache etc.
  * NOTE! you must release this thing when done with it!
  * BEWARE! Thanks to some holes in the *AT syscalls implementation,
  * we cannot get an fd on an AF_UNIX socket, nor reliably on block or
@@ -2001,7 +2081,7 @@ fsal_status_t vfs_check_handle(struct fsal_export *exp_hdl,
 fsal_status_t vfs_create_handle(struct fsal_export *exp_hdl,
 				struct gsh_buffdesc *hdl_desc,
 				struct fsal_obj_handle **handle,
-				struct attrlist *attrs_out)
+				struct fsal_attrlist *attrs_out)
 {
 	fsal_status_t status;
 	struct vfs_fsal_obj_handle *hdl;
@@ -2027,8 +2107,7 @@ fsal_status_t vfs_create_handle(struct fsal_export *exp_hdl,
 		fd = -1;
 		retval = stat(fs->path, &obj_stat);
 	} else {
-		fd = vfs_open_by_handle(fs->private_data, fh, flags,
-					&fsal_error);
+		fd = vfs_open_by_handle(fs, fh, flags, &fsal_error);
 
 		if (fd < 0) {
 			#ifdef __FreeBSD__
@@ -2081,6 +2160,11 @@ fsal_status_t vfs_create_handle(struct fsal_export *exp_hdl,
 
 	if (attrs_out != NULL) {
 		posix2fsal_attributes_all(&obj_stat, attrs_out);
+
+		/* Get correct fsid from the fsal_filesystem, it may
+		* not be the device major/minor from stat.
+		*/
+		attrs_out->fsid = hdl->obj_handle.fs->fsid;
 	}
 
 	*handle = &hdl->obj_handle;
