@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -68,6 +69,7 @@ static struct vfs_acl_entry *vfs_acl_lookup(struct gsh_buffdesc *key)
 	struct vfs_acl_entry key_entry;
 	struct avltree_node *node;
 
+	memset(&key_entry, 0, sizeof(key_entry));
 	key_entry.fa_key = *key;
 	node = avltree_lookup(&key_entry.fa_node, &vfs_acl_tree);
 	if (!node)
@@ -128,7 +130,7 @@ void vfs_acl_release(struct gsh_buffdesc *key)
 
 fsal_status_t vfs_sub_getattrs(struct vfs_fsal_obj_handle *vfs_hdl,
 			       int fd, attrmask_t request_mask,
-			       struct attrlist *attrib)
+			       struct fsal_attrlist *attrib)
 {
 	fsal_status_t fsal_st = {ERR_FSAL_NO_ERROR, 0};
 
@@ -200,7 +202,7 @@ fsal_status_t vfs_sub_getattrs(struct vfs_fsal_obj_handle *vfs_hdl,
 
 fsal_status_t vfs_sub_setattrs(struct vfs_fsal_obj_handle *vfs_hdl,
 			       int fd, attrmask_t request_mask,
-			       struct attrlist *attrib)
+			       struct fsal_attrlist *attrib)
 {
 #ifdef ENABLE_VFS_DEBUG_ACL
 	struct vfs_acl_entry *fa;
