@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -60,12 +61,13 @@
 #ifdef USE_LTTNG
 #include "gsh_lttng/mdcache.h"
 #endif
+#include "sys_resource.h"
 
 /**
  *
  * @file mdcache_lru.c
  * @author Matt Benjamin <matt@linuxbox.com>
- * @brief Constant-time cache inode cache management implementation
+ * @brief Constant-time MDCACHE cache management implementation
  */
 
 /**
@@ -75,16 +77,16 @@
  * based on LRU.  Some ideas are taken from 2Q [Johnson and Shasha 1994]
  * and MQ [Zhou, Chen, Li 2004].  In this system, cache management does
  * interact with cache entry lifecycle, but the lru queue is not a garbage
- * collector. Most imporantly, cache management operations execute in constant
+ * collector. Most importantly, cache management operations execute in constant
  * time, as expected with LRU (and MQ).
  *
  * Cache entries in use by a currently-active protocol request (or other
- * operation) have a positive refcount, and threfore should not be present
+ * operation) have a positive refcount, and therefore should not be present
  * at the cold end of an lru queue if the cache is well-sized.
  *
  * As noted below, initial references to cache entries may only be granted
- * under the cache inode hash table latch.  Likewise, entries must first be
- * made unreachable to the cache inode hash table, then independently reach
+ * under the MDCACHE hash table latch.  Likewise, entries must first be
+ * made unreachable to the MDCACHE hash table, then independently reach
  * a refcnt of 0, before they may be disposed or recycled.
  */
 
@@ -560,8 +562,8 @@ mdcache_lru_clean(mdcache_entry_t *entry)
 		 * we can use the first_export_id to get a valid export for
 		 * a reaping case.
 		 */
-		struct root_op_context ctx;
-		struct req_op_context *saved_ctx = op_ctx;
+		struct req_op_context op_context;
+		bool used_ctx = false;
 		int32_t export_id;
 		struct gsh_export *export;
 
@@ -603,8 +605,9 @@ mdcache_lru_clean(mdcache_entry_t *entry)
 				     PRIi32,
 				     export_id);
 
-			init_root_op_context(&ctx, export, export->fsal_export,
-					     0, 0, UNKNOWN_REQUEST);
+			init_op_context_simple(&op_context, export,
+					       export->fsal_export);
+			used_ctx = true;
 		} else {
 			/* We MUST have a valid op_ctx based on the conditions
 			 * we could get here. first_export_id coild be -1 or it
@@ -635,12 +638,11 @@ mdcache_lru_clean(mdcache_entry_t *entry)
 		       );
 		entry->sub_handle = NULL;
 
-		if (op_ctx != saved_ctx) {
+		if (used_ctx) {
 			/* We had to use our own op_ctx, clean it up and revert
 			 * to the saved op_ctx.
 			 */
-			put_gsh_export(op_ctx->ctx_export);
-			op_ctx = saved_ctx;
+			release_op_context();
 		}
 	}
 
@@ -660,7 +662,7 @@ mdcache_lru_clean(mdcache_entry_t *entry)
 	PTHREAD_RWLOCK_destroy(&entry->content_lock);
 	PTHREAD_RWLOCK_destroy(&entry->attr_lock);
 
-	state_hdl_cleanup(entry->obj_handle.state_hdl);
+	state_hdl_cleanup(entry->obj_handle.state_hdl, entry->obj_handle.type);
 
 	if (entry->obj_handle.type == DIRECTORY)
 		pthread_spin_destroy(&entry->fsobj.fsdir.spin);
@@ -718,56 +720,51 @@ lru_reap_impl(enum lru_q_id qid)
 		QUNLOCK(qlane);
 
 		if (unlikely(refcnt != (LRU_SENTINEL_REFCOUNT + 1))) {
-			/* cant use it. */
+			/* can't use it. */
 			adjust_lru_root_object(entry);
 			mdcache_put(entry);
 			continue;
 		}
 		/* potentially reclaimable */
 		/* entry must be unreachable from CIH when recycled */
-		if (cih_latch_entry(&entry->fh_hk.key, &latch, CIH_GET_WLOCK,
-				    __func__, __LINE__)) {
-			QLOCK(qlane);
-			refcnt = atomic_fetch_int32_t(&entry->lru.refcnt);
-			/* there are two cases which permit reclaim,
-			 * entry is:
-			 * 1. reachable but unref'd (refcnt==2)
-			 * 2. unreachable, being removed (plus refcnt==0)
-			 *  for safety, take only the former
-			 */
-			if (LRU_ENTRY_RECLAIMABLE(entry, refcnt)) {
-				/* it worked */
-				struct lru_q *q = lru_queue_of(entry);
+		cih_latch_entry(&entry->fh_hk.key, &latch, CIH_GET_WLOCK,
+				    __func__, __LINE__);
+		QLOCK(qlane);
+		refcnt = atomic_fetch_int32_t(&entry->lru.refcnt);
+		/* there are two cases which permit reclaim,
+		 * entry is:
+		 * 1. reachable but unref'd (refcnt==2)
+		 * 2. unreachable, being removed (plus refcnt==0)
+		 *  for safety, take only the former
+		 */
+		if (LRU_ENTRY_RECLAIMABLE(entry, refcnt)) {
+			/* it worked */
+			struct lru_q *q = lru_queue_of(entry);
 
 #ifdef USE_LTTNG
-				tracepoint(mdcache, mdc_lru_reap, __func__,
-					   __LINE__, &entry->obj_handle,
-					   entry->lru.refcnt);
+			tracepoint(mdcache, mdc_lru_reap, __func__,
+				   __LINE__, &entry->obj_handle,
+				   entry->lru.refcnt);
 #endif
-				LRU_DQ_SAFE(lru, q);
-				entry->lru.qid = LRU_ENTRY_NONE;
-				QUNLOCK(qlane);
-				cih_remove_latched(entry, &latch,
-						   CIH_REMOVE_UNLOCK);
-				/* Note, we're not releasing our ref here.
-				 * cih_remove_latched() called
-				 * mdcache_lru_unref(), which released the
-				 * sentinal ref, leaving just the one ref we
-				 * took earlier.  Returning this as is leaves it
-				 * with a ref of 1 (ie, just the sentinal ref)
-				 * */
-				goto out;
-			}
-			cih_hash_release(&latch);
+			LRU_DQ_SAFE(lru, q);
+			entry->lru.qid = LRU_ENTRY_NONE;
 			QUNLOCK(qlane);
-			/* return the ref we took above--unref deals
-			 * correctly with reclaim case */
-			mdcache_lru_unref(entry);
-		} else {
-			/* ! QLOCKED but needs to be Unref'ed */
-			mdcache_lru_unref(entry);
-			continue;
+			cih_remove_latched(entry, &latch,
+					   CIH_REMOVE_UNLOCK);
+			/* Note, we're not releasing our ref here.
+			 * cih_remove_latched() called
+			 * mdcache_lru_unref(), which released the
+			 * sentinel ref, leaving just the one ref we
+			 * took earlier.  Returning this as is leaves it
+			 * with a ref of 1 (ie, just the sentinel ref)
+			 * */
+			goto out;
 		}
+		cih_hash_release(&latch);
+		QUNLOCK(qlane);
+		/* return the ref we took above--unref deals
+		 * correctly with reclaim case */
+		mdcache_lru_unref(entry);
 	}			/* foreach lane */
 
 	/* ! reclaimable */
@@ -1044,11 +1041,11 @@ mdcache_lru_cleanup_push(mdcache_entry_t *entry)
  * @brief Push an entry to the cleanup queue that may be unexported
  * for out-of-line cleanup
  *
- * This routine is used to try pushing a cache inode into the cleanup
+ * This routine is used to try pushing a cache entry into the cleanup
  * queue. If the entry ends up with another LRU reference before this
  * is accomplished, then don't push it to cleanup.
  *
- * This will be used when unexporting an export. Any cache inode entry
+ * This will be used when unexporting an export. Any cache entry
  * that only belonged to that export is a candidate for cleanup.
  * However, it is possible the entry is still accessible via another
  * export, and an LRU reference might be gained before we can lock the
@@ -1064,40 +1061,43 @@ mdcache_lru_cleanup_try_push(mdcache_entry_t *entry)
 	struct lru_q_lane *qlane = &LRU[lru->lane];
 	cih_latch_t latch;
 
-	if (cih_latch_entry(&entry->fh_hk.key, &latch, CIH_GET_WLOCK,
-			    __func__, __LINE__)) {
-		QLOCK(qlane);
+	cih_latch_entry(&entry->fh_hk.key, &latch, CIH_GET_WLOCK,
+			    __func__, __LINE__);
+	QLOCK(qlane);
 
-		/* Take the attr lock, so we can check that this entry is still
-		 * not in any export */
-		PTHREAD_RWLOCK_rdlock(&entry->attr_lock);
+	/* Take the attr lock, so we can check that this entry is still
+	 * not in any export */
+	PTHREAD_RWLOCK_rdlock(&entry->attr_lock);
 
-		if (glist_empty(&entry->export_list)) {
-			/* it worked */
-			struct lru_q *q = lru_queue_of(entry);
+	/* Make sure that the entry is not reaped by the time this
+	 * thread got the QLOCK
+	 */
+	if (glist_empty(&entry->export_list) &&
+	    entry->lru.qid != LRU_ENTRY_NONE) {
+		/* it worked */
+		struct lru_q *q = lru_queue_of(entry);
 
-			LRU_DQ_SAFE(lru, q);
-			entry->lru.qid = LRU_ENTRY_CLEANUP;
-			atomic_set_uint32_t_bits(&entry->lru.flags,
-						 LRU_CLEANUP);
-			/* Note: we didn't take a ref here, so the only ref left
-			 * in this callpath is the one owned by
-			 * mdcache_unexport().  When it unref's, that may free
-			 * this object; otherwise, it will be freed when the
-			 * last op is done, as it's unreachable. */
+		LRU_DQ_SAFE(lru, q);
+		entry->lru.qid = LRU_ENTRY_CLEANUP;
+		atomic_set_uint32_t_bits(&entry->lru.flags,
+					 LRU_CLEANUP);
+		/* Note: we didn't take a ref here, so the only ref left
+		 * in this callpath is the one owned by
+		 * mdcache_unexport().  When it unref's, that may free
+		 * this object; otherwise, it will be freed when the
+		 * last op is done, as it's unreachable. */
 
-			/* It's safe to drop the attr lock here, as we have the
-			 * latch, so no one can look up the entry */
-			PTHREAD_RWLOCK_unlock(&entry->attr_lock);
-			QUNLOCK(qlane);
-			cih_remove_latched(entry, &latch, CIH_REMOVE_NONE);
-		} else {
-			PTHREAD_RWLOCK_unlock(&entry->attr_lock);
-			QUNLOCK(qlane);
-		}
-
-		cih_hash_release(&latch);
+		/* It's safe to drop the attr lock here, as we have the
+		 * latch, so no one can look up the entry */
+		PTHREAD_RWLOCK_unlock(&entry->attr_lock);
+		QUNLOCK(qlane);
+		cih_remove_latched(entry, &latch, CIH_REMOVE_NONE);
+	} else {
+		PTHREAD_RWLOCK_unlock(&entry->attr_lock);
+		QUNLOCK(qlane);
 	}
+
+	cih_hash_release(&latch);
 }
 
 /**
@@ -1142,8 +1142,7 @@ static inline size_t lru_run_lane(size_t lane, uint64_t *const totalclosed)
 	 * the iteration also adjusts glist and (in particular) glistn */
 	glist_for_each_safe(qlane->iter.glist, qlane->iter.glistn, &q->q) {
 		struct lru_q *q;
-		struct root_op_context ctx;
-		struct req_op_context *saved_ctx = op_ctx;
+		struct req_op_context op_context;
 		int32_t export_id;
 		struct gsh_export *export;
 
@@ -1196,8 +1195,8 @@ static inline size_t lru_run_lane(size_t lane, uint64_t *const totalclosed)
 		   refcnt);
 #endif
 
-		init_root_op_context(&ctx, export, export->fsal_export, 0, 0,
-				     UNKNOWN_REQUEST);
+		init_op_context_simple(&op_context, export,
+				       export->fsal_export);
 
 		/* check refcnt in range */
 		if (unlikely(refcnt > 2)) {
@@ -1237,8 +1236,7 @@ static inline size_t lru_run_lane(size_t lane, uint64_t *const totalclosed)
 
 next_lru:
 		QLOCK(qlane); /* QLOCKED */
-		put_gsh_export(export);
-		op_ctx = saved_ctx;
+		release_op_context();
 
 		++workdone;
 	} /* for_each_safe lru */
@@ -1334,10 +1332,12 @@ lru_run(struct fridgethr_context *ctx)
 
 	SetNameFunction("cache_lru");
 
+	/* fds_avg is the average *range* of fds instead absolute average
+	 * e.g. high water(10) and low water(5), the *range* between these is
+	 * `5` then avg of *range* should be `2.5`
+	 */
 	fds_avg = (lru_state.fds_hiwat - lru_state.fds_lowat) / 2;
 
-	extremis = atomic_fetch_size_t(&open_fd_count) > lru_state.fds_hiwat;
-
 	LogFullDebug(COMPONENT_CACHE_INODE_LRU, "LRU awakes.");
 
 	if (!woke) {
@@ -1361,17 +1361,29 @@ lru_run(struct fridgethr_context *ctx)
 
 	currentopen = atomic_fetch_size_t(&open_fd_count);
 
+	/* avoid to do more atomic fetch we could check with currentopen */
+	extremis = currentopen > lru_state.fds_hiwat;
+
 	if (currentopen < lru_state.fds_lowat) {
 		LogDebug(COMPONENT_CACHE_INODE_LRU,
 			 "FD count is %zd and low water mark is %d: not reaping.",
-			 atomic_fetch_size_t(&open_fd_count),
-			 lru_state.fds_lowat);
+			 currentopen, lru_state.fds_lowat);
 		if (atomic_fetch_uint32_t(&lru_state.fd_state) > FD_LOW) {
 			LogEvent(COMPONENT_CACHE_INODE_LRU,
 				 "Return to normal fd reaping.");
 			atomic_store_uint32_t(&lru_state.fd_state, FD_LOW);
 		}
 	} else {
+		/* Set state first because that does not realted
+		 * with following code flow
+		 */
+		if (currentopen < lru_state.fds_hiwat &&
+		    atomic_fetch_uint32_t(&lru_state.fd_state) == FD_LIMIT) {
+			LogEvent(COMPONENT_CACHE_INODE_LRU,
+				 "Count of fd is below high water mark.");
+			atomic_store_uint32_t(&lru_state.fd_state, FD_MIDDLE);
+		}
+
 		/* The count of open file descriptors before this run
 		   of the reaper. */
 		size_t formeropen = atomic_fetch_size_t(&open_fd_count);
@@ -1381,17 +1393,11 @@ lru_run(struct fridgethr_context *ctx)
 		size_t workpass = 0;
 		time_t curr_time = time(NULL);
 
-		if (currentopen < lru_state.fds_hiwat &&
-		    atomic_fetch_uint32_t(&lru_state.fd_state) == FD_LIMIT) {
-			LogEvent(COMPONENT_CACHE_INODE_LRU,
-				 "Count of fd is below high water mark.");
-			atomic_store_uint32_t(&lru_state.fd_state, FD_MIDDLE);
-		}
-
 		if ((curr_time >= lru_state.prev_time) &&
 		    (curr_time - lru_state.prev_time < fridgethr_getwait(ctx)))
 			threadwait = curr_time - lru_state.prev_time;
 
+		/* means increase rate or 1 */
 		fdratepersec = ((curr_time <= lru_state.prev_time) ||
 				(formeropen < lru_state.prev_fd_count))
 			? 1 : (formeropen - lru_state.prev_fd_count) /
@@ -1402,6 +1408,9 @@ lru_run(struct fridgethr_context *ctx)
 			     fdratepersec, formeropen,
 			     ((uint64_t) (curr_time - lru_state.prev_time)));
 
+		/* `extremis` setup here again would be more exactly */
+		extremis = formeropen > lru_state.fds_hiwat;
+
 		if (extremis) {
 			LogDebug(COMPONENT_CACHE_INODE_LRU,
 				 "Open FDs over high water mark, reapring aggressively.");
@@ -1426,6 +1435,7 @@ lru_run(struct fridgethr_context *ctx)
 		} while (extremis && (workpass >= lru_state.per_lane_work)
 			 && (totalwork < lru_state.biggest_window));
 
+		/* means latest current open fds after reap */
 		currentopen = atomic_fetch_size_t(&open_fd_count);
 		if (extremis &&
 		    ((currentopen > formeropen)
@@ -1699,7 +1709,7 @@ void init_fds_limit(void)
 	};
 
 	/* Find out the system-imposed file descriptor limit */
-	if (getrlimit(RLIMIT_NOFILE, &rlim) != 0) {
+	if (get_open_file_limit(&rlim) != 0) {
 		code = errno;
 		LogCrit(COMPONENT_CACHE_INODE_LRU,
 			"Call to getrlimit failed with error %d. This should not happen.  Assigning default of %d.",
@@ -1957,7 +1967,7 @@ mdcache_entry_t *mdcache_lru_get(struct fsal_obj_handle *sub_handle)
  * else, insert into LRU of L1, so that a single ref promotes to the MRU of L1.
  *
  * @param [in] entry  Entry to insert.
- * @param [in] reason Reason we're inserting
+ * @param [in] reason The Reason we're inserting
  */
 void mdcache_lru_insert(mdcache_entry_t *entry, mdc_reason_t reason)
 {
@@ -2078,6 +2088,12 @@ _mdcache_lru_unref(mdcache_entry_t *entry, uint32_t flags, const char *func,
 			goto out;
 		}
 
+		/*
+		 * The cih table holds a non-weak reference, so entry should no
+		 * longer be in it.
+		 */
+		assert(!entry->fh_hk.inavl);
+
 		/* Really zero.  Remove entry and mark it as dead. */
 		q = lru_queue_of(entry);
 		if (q) {
@@ -2280,8 +2296,7 @@ void mdc_lru_map_dirent(mdcache_dir_entry_t *dirent)
 		/* Move to MRU */
 		dmap = avltree_container_of(node, mdcache_dmap_entry_t, node);
 		now(&dmap->timestamp);
-		glist_del(&dmap->lru_entry);
-		glist_add_tail(&exp->dirent_map.lru, &dmap->lru_entry);
+		glist_move_tail(&exp->dirent_map.lru, &dmap->lru_entry);
 		PTHREAD_MUTEX_unlock(&exp->dirent_map.mtx);
 		return;
 	}
