@@ -1,3 +1,29 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
+/*
+ * vim:expandtab:shiftwidth=8:tabstop=8:
+ *
+ * Copyright CEA/DAM/DIF  (2008)
+ * contributeur : Philippe DENIEL   philippe.deniel@cea.fr
+ *                Thomas LEIBOVICI  thomas.leibovici@cea.fr
+ *
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 3 of the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+ *
+ * -------------
+ */
+
 /**
  * @addtogroup FSAL
  * @{
@@ -431,11 +457,12 @@ static void fsal_print_access_by_acl(int naces, int ace_number,
  * @return ERR_FSAL_NO_ERROR, ERR_FSAL_ACCESS, or ERR_FSAL_NO_ACE
  */
 
-static fsal_status_t fsal_check_access_acl(struct user_cred *creds,
-					   fsal_aceperm_t v4mask,
-					   fsal_accessflags_t *allowed,
-					   fsal_accessflags_t *denied,
-					   struct attrlist *p_object_attributes)
+static fsal_status_t fsal_check_access_acl(
+				struct user_cred *creds,
+				fsal_aceperm_t v4mask,
+				fsal_accessflags_t *allowed,
+				fsal_accessflags_t *denied,
+				struct fsal_attrlist *p_object_attributes)
 {
 	fsal_aceperm_t missing_access;
 	fsal_aceperm_t tperm;
@@ -546,7 +573,7 @@ static fsal_status_t fsal_check_access_acl(struct user_cred *creds,
 		ace_number += 1;
 
 		LogFullDebug(COMPONENT_NFS_V4_ACL,
-			     "ace numnber: %d ace type 0x%X perm 0x%X flag 0x%X who %u",
+			     "ace number: %d ace type 0x%X perm 0x%X flag 0x%X who %u",
 			     ace_number, pace->type, pace->perm, pace->flag,
 			     GET_FSAL_ACE_WHO(*pace));
 
@@ -693,7 +720,7 @@ fsal_check_access_no_acl(struct user_cred *creds,
 			 fsal_accessflags_t access_type,
 			 fsal_accessflags_t *allowed,
 			 fsal_accessflags_t *denied,
-			 struct attrlist *p_object_attributes)
+			 struct fsal_attrlist *p_object_attributes)
 {
 	uid_t uid;
 	gid_t gid;
@@ -821,7 +848,7 @@ fsal_status_t fsal_test_access(struct fsal_obj_handle *obj_hdl,
 			       fsal_accessflags_t *denied,
 			       bool owner_skip)
 {
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 	fsal_status_t status;
 
 	fsal_prepare_attrs(&attrs,
@@ -834,18 +861,18 @@ fsal_status_t fsal_test_access(struct fsal_obj_handle *obj_hdl,
 	if (FSAL_IS_ERROR(status))
 		goto out;
 
-	if (owner_skip && attrs.owner == op_ctx->creds->caller_uid) {
+	if (owner_skip && attrs.owner == op_ctx->creds.caller_uid) {
 		status = fsalstat(ERR_FSAL_NO_ERROR, 0);
 		goto out;
 	}
 
 	if (IS_FSAL_ACE4_REQ(access_type) ||
 	    (attrs.acl != NULL && IS_FSAL_ACE4_MASK_VALID(access_type))) {
-		status = fsal_check_access_acl(op_ctx->creds,
+		status = fsal_check_access_acl(&op_ctx->creds,
 					       FSAL_ACE4_MASK(access_type),
 					       allowed, denied, &attrs);
 	} else {		/* fall back to use mode to check access. */
-		status = fsal_check_access_no_acl(op_ctx->creds,
+		status = fsal_check_access_no_acl(&op_ctx->creds,
 						  FSAL_MODE_MASK(access_type),
 						  allowed, denied, &attrs);
 	}
@@ -863,18 +890,30 @@ gid_t ganesha_gid;
 int ganesha_ngroups;
 gid_t *ganesha_groups;
 
+#if GSH_CAN_HOST_LOCAL_FS
 void fsal_set_credentials(const struct user_cred *creds)
 {
 	if (set_threadgroups(creds->caller_glen, creds->caller_garray) != 0)
-		LogFatal(COMPONENT_FSAL, "Could not set Context credentials");
+		LogFatal(COMPONENT_FSAL,
+			 "set_threadgroups() returned %s (%d)",
+			 strerror(errno), errno);
 	setgroup(creds->caller_gid);
 	setuser(creds->caller_uid);
 }
 
+void fsal_restore_ganesha_credentials(void)
+{
+	setuser(ganesha_uid);
+	setgroup(ganesha_gid);
+	if (set_threadgroups(ganesha_ngroups, ganesha_groups) != 0)
+		LogFatal(COMPONENT_FSAL, "Could not set Ganesha credentials");
+}
+#endif		/* GSH_CAN_HOST_LOCAL_FS */
+
 bool fsal_set_credentials_only_one_user(const struct user_cred *creds)
 {
-	if (creds->caller_uid == ganesha_uid
-		    && creds->caller_gid == ganesha_gid)
+	if (creds->caller_uid == ganesha_uid &&
+	    creds->caller_gid == ganesha_gid)
 		return true;
 	else
 		return false;
@@ -882,8 +921,9 @@ bool fsal_set_credentials_only_one_user(const struct user_cred *creds)
 
 void fsal_save_ganesha_credentials(void)
 {
-	int i;
-	char buffer[1024], *p = buffer;
+	int i, b_left;
+	char buffer[1024];
+	struct display_buffer dspbuf = {sizeof(buffer), buffer, buffer};
 
 	ganesha_uid = getuser();
 	ganesha_gid = getgroup();
@@ -899,29 +939,26 @@ void fsal_save_ganesha_credentials(void)
 		}
 	}
 
-	p += sprintf(p, "Ganesha uid=%d gid=%d ngroups=%d", (int)ganesha_uid,
-		     (int)ganesha_gid, ganesha_ngroups);
-	if (ganesha_ngroups != 0)
-		p += sprintf(p, " (");
-	for (i = 0; i < ganesha_ngroups; i++) {
-		if ((p - buffer) < (sizeof(buffer) - 10)) {
-			if (i == 0)
-				p += sprintf(p, "%d", (int)ganesha_groups[i]);
-			else
-				p += sprintf(p, " %d", (int)ganesha_groups[i]);
-		}
+	if (!isInfo(COMPONENT_FSAL))
+		return;
+
+	b_left = display_printf(&dspbuf,  "Ganesha uid=%d gid=%d ngroups=%d",
+				(int)ganesha_uid, (int)ganesha_gid,
+				ganesha_ngroups);
+
+	if (b_left > 0 && ganesha_ngroups != 0)
+		b_left = display_cat(&dspbuf, " (");
+
+	for (i = 0; b_left > 0 && i < ganesha_ngroups; i++) {
+		b_left = display_printf(&dspbuf, "%s%d",
+					i == 0 ? "" : " ",
+					(int)ganesha_groups[i]);
 	}
-	if (ganesha_ngroups != 0)
-		p += sprintf(p, ")");
+
+	if (b_left > 0 && ganesha_ngroups != 0)
+		(void) display_cat(&dspbuf, ")");
+
 	LogInfo(COMPONENT_FSAL, "%s", buffer);
 }
 
-void fsal_restore_ganesha_credentials(void)
-{
-	setuser(ganesha_uid);
-	setgroup(ganesha_gid);
-	if (set_threadgroups(ganesha_ngroups, ganesha_groups) != 0)
-		LogFatal(COMPONENT_FSAL, "Could not set Ganesha credentials");
-}
-
 /** @} */
