@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -53,7 +54,7 @@
  * @param[in]  req     SVC request related to this call
  * @param[out] res     Structure to contain the result of the call
  *
- * @retval NFS_REQ_OK if successfull
+ * @retval NFS_REQ_OK if successful
  * @retval NFS_REQ_DROP if failed but retryable
  * @retval NFS_REQ_FAILED if failed and not retryable
  *
@@ -66,7 +67,7 @@ int nfs3_lookup(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 	fsal_status_t fsal_status;
 	char *name = NULL;
 	int rc = NFS_REQ_OK;
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 	LOOKUP3resfail *resfail = &res->res_lookup3.LOOKUP3res_u.resfail;
 	LOOKUP3resok *resok = &res->res_lookup3.LOOKUP3res_u.resok;
 
