@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -88,6 +89,7 @@ static int nfs3_complete_read(struct nfs3_read_data *data)
 	if (data->rc == NFS_REQ_OK) {
 		if (!op_ctx->fsal_export->exp_ops.fs_supports(
 		     op_ctx->fsal_export, fso_compliant_eof_behavior)
+			&& nfs_param.core_param.getattrs_in_complete_read
 		    && !read_arg->end_of_file) {
 			/*
 			 * NFS requires to set the EOF flag for all reads that
@@ -97,7 +99,7 @@ static int nfs3_complete_read(struct nfs3_read_data *data)
 			 * will just see a short read and continue reading and
 			 * then get the EOF flag as 0 bytes are returned.
 			 */
-			struct attrlist attrs;
+			struct fsal_attrlist attrs;
 			fsal_status_t status;
 
 			fsal_prepare_attrs(&attrs, ATTR_SIZE);
@@ -153,7 +155,7 @@ static enum xprt_stat nfs3_read_resume(struct svc_req *req)
 	int rc;
 
 	/* Restore the op_ctx */
-	op_ctx = &reqdata->req_ctx;
+	resume_op_context(&reqdata->op_context);
 
 	/* Complete the read */
 	rc = nfs3_complete_read(data);
@@ -298,7 +300,7 @@ int nfs3_read(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 		goto putref;
 	}
 
-	/* do not exceed maxium READ offset if set */
+	/* do not exceed maximum READ offset if set */
 	if (MaxOffsetRead < UINT64_MAX) {
 		LogFullDebug(COMPONENT_NFSPROTO,
 			     "Read offset=%" PRIu64
@@ -307,7 +309,7 @@ int nfs3_read(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 
 		if ((offset + size) > MaxOffsetRead) {
 			LogEvent(COMPONENT_NFSPROTO,
-				 "A client tryed to violate max file size %"
+				 "A client tried to violate max file size %"
 				 PRIu64 " for exportid #%hu",
 				 MaxOffsetRead,
 				 op_ctx->ctx_export->export_id);
