@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -48,7 +49,7 @@
 
 fsal_errors_t nfs3_readdirplus_callback(void *opaque,
 					       struct fsal_obj_handle *obj,
-					       const struct attrlist *attr,
+					       const struct fsal_attrlist *attr,
 					       uint64_t mounted_on_fileid,
 					       uint64_t cookie,
 					       enum cb_state cb_state);
@@ -73,7 +74,7 @@ static
 nfsstat3 nfs_readdir_dot_entry(struct fsal_obj_handle *obj, const char *name,
 			       uint64_t cookie, helper_readdir_cb cb,
 			       struct nfs3_readdirplus_cb_data *tracker,
-			       struct attrlist *attrs)
+			       struct fsal_attrlist *attrs)
 {
 	struct fsal_readdir_cb_parms cb_parms;
 	fsal_status_t fsal_status;
@@ -99,7 +100,7 @@ nfsstat3 nfs_readdir_dot_entry(struct fsal_obj_handle *obj, const char *name,
  * @param[in]  req     SVC request related to this call
  * @param[out] res     Structure to contain the result of the call
  *
- * @retval NFS_REQ_OK if successfull
+ * @retval NFS_REQ_OK if successful
  * @retval NFS_REQ_DROP if failed but retryable
  * @retval NFS_REQ_FAILED if failed and not retryable
  */
@@ -109,15 +110,19 @@ int nfs3_readdirplus(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 	struct fsal_obj_handle *dir_obj = NULL;
 	uint64_t begin_cookie = 0;
 	uint64_t fsal_cookie = 0;
+	uint32_t maxcount;
+	uint32_t cfg_readdir_size = nfs_param.core_param.readdir_res_size;
 	cookieverf3 cookie_verifier;
 	unsigned int num_entries = 0;
+	uint64_t mem_avail = 0;
+	uint64_t max_mem = 0;
 	object_file_type_t dir_filetype = 0;
 	bool eod_met = false;
 	fsal_status_t fsal_status = {0, 0};
 	fsal_status_t fsal_status_gethandle = {0, 0};
 	int rc = NFS_REQ_OK;
 	struct nfs3_readdirplus_cb_data tracker;
-	struct attrlist attrs_dir, attrs_parent;
+	struct fsal_attrlist attrs_dir, attrs_parent;
 	bool use_cookie_verifier = op_ctx_export_has_option(
 					EXPORT_OPTION_USE_COOKIE_VERIFIER);
 	READDIRPLUS3resfail *resfail =
@@ -152,11 +157,19 @@ int nfs3_readdirplus(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 	 *		struct fattr3_wire attributes
 	 *	cookieverf3 cookieverf (8 or sizeof(cookieverf3))
 	 */
-	tracker.mem_avail = arg->arg_readdirplus3.maxcount
+	if (cfg_readdir_size != 0 &&
+	    cfg_readdir_size < arg->arg_readdirplus3.maxcount) {
+		maxcount = cfg_readdir_size;
+	} else {
+		maxcount = arg->arg_readdirplus3.maxcount;
+	}
+	mem_avail = maxcount
 				- BYTES_PER_XDR_UNIT
 				- BYTES_PER_XDR_UNIT
 				- sizeof(struct fattr3_wire)
 				- sizeof(cookieverf3);
+	max_mem = atomic_fetch_uint64_t(&op_ctx->ctx_export->MaxRead);
+	tracker.mem_avail = MIN(mem_avail, max_mem);
 
 	begin_cookie = arg->arg_readdirplus3.cookie;
 
@@ -309,7 +322,7 @@ int nfs3_readdirplus(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 	}
 
 	if (tracker.error != NFS3_OK) {
-		res->res_readdir3.status = tracker.error;
+		res->res_readdirplus3.status = tracker.error;
 		goto out_destroy;
 	}
 
@@ -333,7 +346,8 @@ int nfs3_readdirplus(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 				/* Oops... */
 				LogCrit(COMPONENT_NFS_READDIR,
 					"Encode of no next entry failed.");
-				res->res_readdir3.status = NFS3ERR_SERVERFAULT;
+				res->res_readdirplus3.status
+					= NFS3ERR_SERVERFAULT;
 				goto out_destroy;
 			}
 		}
@@ -344,7 +358,7 @@ int nfs3_readdirplus(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 			/* Oops... */
 			LogCrit(COMPONENT_NFS_READDIR,
 				"Encode of EOD failed.");
-			res->res_readdir3.status = NFS3ERR_SERVERFAULT;
+			res->res_readdirplus3.status = NFS3ERR_SERVERFAULT;
 			goto out_destroy;
 		}
 
@@ -425,7 +439,7 @@ void nfs3_readdirplus_free(nfs_res_t *resp)
  *
  * @param opaque [in] Pointer to a struct nfs3_readdirplus_cb_data that is
  *                    gives the location of the array and other
- *                    bookeeping information
+ *                    bookkeeping information
  * @param name [in] The filename for the current obj
  * @param handle [in] The current obj's filehandle
  * @param attrs [in] The current obj's attributes
@@ -434,7 +448,7 @@ void nfs3_readdirplus_free(nfs_res_t *resp)
 
 fsal_errors_t nfs3_readdirplus_callback(void *opaque,
 					struct fsal_obj_handle *obj,
-					const struct attrlist *attr,
+					const struct fsal_attrlist *attr,
 					uint64_t mounted_on_fileid,
 					uint64_t cookie,
 					enum cb_state cb_state)
@@ -476,10 +490,10 @@ fsal_errors_t nfs3_readdirplus_callback(void *opaque,
 
 	/* Encode the entry into the xdrmem buffer and then assure there is
 	 * space for at least two booleans (one to be false to terminate the
-	 * entry ist, the other to encode EOD or not). Note we use the special
-	 * xdr_encode_entryplus3 that uses a passed in struct attrlist rather
-	 * than name_attributes from entryplus3, though we will use the boolean
-	 * attributes_follow from the entryplus3.
+	 * entry list, the other to encode EOD or not). Note we use the special
+	 * xdr_encode_entryplus3 that uses a passed in struct fsal_attrlist
+	 * rather than name_attributes from entryplus3, though we will use the
+	 * boolean attributes_follow from the entryplus3.
 	 */
 	if (!xdr_encode_entryplus3(&tracker->xdr, &ep3, attr) ||
 	    (xdr_getpos(&tracker->xdr) + BYTES_PER_XDR_UNIT)
