@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -42,6 +43,7 @@
 #include "server_stats.h"
 #include "export_mgr.h"
 #include "nfs_creds.h"
+#include "pnfs_utils.h"
 
 #ifdef USE_LTTNG
 #include "gsh_lttng/nfs_rpc.h"
@@ -76,19 +78,19 @@ struct nfs4_op_desc {
 	 * Note that entries placed into the NFS v4.1 slot cache are marked so
 	 * the higher level operation completion will not release them. A deep
 	 * copy is made when the slot cache is replayed. If sa_cachethis
-	 * indicates a response will not be cached, the higher level operaiton
+	 * indicates a response will not be cached, the higher level operation
 	 * completion will call the free_res, HOWEVER, a shallow copy of the
 	 * SEQUENCE op and first operation responses are made. If the first
 	 * operation resulted in an error (other than NFS4_DENIED for LOCK and
 	 * LOCKT) the shallow copy preserves that error rather than replacing
 	 * it with NFS4ERR_RETRY_UNCACHED_REP. For this reason for any response
-	 * that includes dyanmically allocated data on NFS4_OK MUST check the
+	 * that includes dynamically allocated data on NFS4_OK MUST check the
 	 * response status before freeing any memory since the shallow copy will
 	 * mean the cached NFS4ERR_RETRY_UNCACHED_REP response will have copied
 	 * those pointers. It should only free data if the status is NFS4_OK
 	 * (or NFS4ERR_DENIED in the case of LOCK and LOCKT). Note that
 	 * SETCLIENTID also has dunamic data on a non-NFS4_OK status, and the
-	 * free_res function for that checks, howwever, we will never see
+	 * free_res function for that checks, however, we will never see
 	 * SETCLIENTID in NFS v4.1+, or if we do, it will get an error.
 	 *
 	 * At this time, LOCK and LOCKT are the only NFS v4.1 or v4.2 operations
@@ -107,6 +109,7 @@ struct nfs4_op_desc {
 /**
  * @brief  NFSv4 and 4.1 ops table.
  * indexed by opcode
+ * name *must* begin with OP_
  */
 
 static const struct nfs4_op_desc optabv4[] = {
@@ -743,7 +746,7 @@ enum nfs_req_result complete_op(compound_data_t *data, nfsstat4 *status,
 
 out:
 
-	server_stats_nfsv4_op_done(data->opcode, data->op_start_time, *status);
+	server_stats_nfsv4_op_done(data->opcode, &data->op_start_time, *status);
 
 	return result;
 }
@@ -751,7 +754,6 @@ out:
 enum nfs_req_result process_one_op(compound_data_t *data, nfsstat4 *status)
 {
 	const char *bad_op_state_reason = "";
-	struct timespec ts;
 	int perm_flags;
 	log_components_t alt_component = COMPONENT_NFS_V4;
 	nfs_argop4 *thisarg = &data->argarray[data->oppos];
@@ -829,8 +831,7 @@ enum nfs_req_result process_one_op(compound_data_t *data, nfsstat4 *status)
 	}
 
 	/* time each op */
-	now(&ts);
-	data->op_start_time = timespec_diff(&nfs_ServerBootTime, &ts);
+	now(&data->op_start_time);
 
 	if (data->minorversion > 0 && data->session != NULL &&
 	    data->session->fore_channel_attrs.ca_maxoperations ==
@@ -856,11 +857,10 @@ enum nfs_req_result process_one_op(compound_data_t *data, nfsstat4 *status)
 		 */
 		LogMidDebugAlt(COMPONENT_NFS_V4, COMPONENT_EXPORT,
 			       "Check export perms export = %08x req = %08x",
-			       op_ctx->export_perms->options &
+			       op_ctx->export_perms.options &
 					EXPORT_OPTION_ACCESS_MASK,
 			       perm_flags);
-		if ((op_ctx->export_perms->options &
-		     perm_flags) != perm_flags) {
+		if ((op_ctx->export_perms.options & perm_flags) != perm_flags) {
 			/* Export doesn't allow requested
 			 * access for this client.
 			 */
@@ -1061,7 +1061,8 @@ static enum xprt_stat nfs4_compound_resume(struct svc_req *req)
 	compound_data_t *data = reqdata->proc_data;
 	enum nfs_req_result result;
 
-	op_ctx = &reqdata->req_ctx;
+	/* Restore the op_ctx */
+	resume_op_context(&reqdata->op_context);
 
 	/* Start by resuming the operation that suspended. */
 	result = (optabv4[data->opcode].resume)
@@ -1074,9 +1075,10 @@ static enum xprt_stat nfs4_compound_resume(struct svc_req *req)
 		/* The request is suspended, don't touch the request in
 		 * any way because the resume may already be scheduled
 		 * and running on nother thread. The xp_resume_cb has
-		 * already been set up before we started proecessing
+		 * already been set up before we started processing
 		 * ops on this request at all.
 		 */
+		suspend_op_context();
 		return XPRT_SUSPEND;
 	}
 
@@ -1090,9 +1092,10 @@ static enum xprt_stat nfs4_compound_resume(struct svc_req *req)
 			/* The request is suspended, don't touch the request in
 			 * any way because the resume may already be scheduled
 			 * and running on nother thread. The xp_resume_cb has
-			 * already been set up before we started proecessing
+			 * already been set up before we started processing
 			 * ops on this request at all.
 			 */
+			suspend_op_context();
 			return XPRT_SUSPEND;
 		}
 	}
@@ -1101,12 +1104,8 @@ static enum xprt_stat nfs4_compound_resume(struct svc_req *req)
 
 	compound_data_Free(data);
 	/* release current active export in op_ctx. */
-	if (op_ctx->ctx_export) {
-		put_gsh_export(op_ctx->ctx_export);
-		op_ctx->ctx_export = NULL;
-		op_ctx->fsal_export = NULL;
-	}
-
+	if (op_ctx->ctx_export)
+		clear_op_context_export();
 
 	nfs_rpc_complete_async_request(reqdata, NFS_REQ_OK);
 
@@ -1239,7 +1238,7 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 	}
 
 	/* Check for too long request */
-	if (argarray_len > 100) {
+	if (argarray_len > NFS4_MAX_OPERATIONS) {
 		LogMajor(COMPONENT_NFS_V4,
 			 "A COMPOUND with too many operations (%d) was received",
 			 argarray_len);
@@ -1298,7 +1297,7 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 			 * If the COMPOUND request does not start with SEQUENCE,
 			 * and if DESTROY_SESSION is not the sole operation,
 			 * then server MUST return  NFS4ERR_NOT_ONLY_OP. See
-			 * 18.37.3 nd test DSESS9005 for details
+			 * 18.37.3 and test DSESS9005 for details
 			 */
 			if (argarray[0].argop == NFS4_OP_EXCHANGE_ID ||
 			    argarray[0].argop == NFS4_OP_CREATE_SESSION ||
@@ -1330,7 +1329,7 @@ int nfs4_Compound(nfs_arg_t *arg, struct svc_req *req, nfs_res_t *res)
 			/* The request is suspended, don't touch the request in
 			 * any way because the resume may already be scheduled
 			 * and running on nother thread. The xp_resume_cb has
-			 * already been set up before we started proecessing
+			 * already been set up before we started processing
 			 * ops on this request at all.
 			 */
 			return result;
@@ -1344,11 +1343,8 @@ out:
 	compound_data_Free(data);
 
 	/* release current active export in op_ctx. */
-	if (op_ctx->ctx_export) {
-		put_gsh_export(op_ctx->ctx_export);
-		op_ctx->ctx_export = NULL;
-		op_ctx->fsal_export = NULL;
-	}
+	if (op_ctx->ctx_export)
+		clear_op_context_export();
 
 	return drop ? NFS_REQ_DROP : NFS_REQ_OK;
 }				/* nfs4_Compound */
@@ -1436,7 +1432,13 @@ void compound_data_Free(compound_data_t *data)
 	if (data == NULL)
 		return;
 
-	/* Release refcounted cache entries */
+	/* Release refcounted cache entries. A note on current_ds and saved_ds,
+	 * If both are in use and the same, it will be released during
+	 * set_saved_entry since set_current_entry will have set current_ds to
+	 * NULL. If both are non-NULL and different, current_ds will be
+	 * release by set_current_entry and saved_ds will be released by
+	 * set_saved_entry.
+	 */
 	set_current_entry(data, NULL);
 	set_saved_entry(data, NULL);
 
@@ -1461,6 +1463,12 @@ void compound_data_Free(compound_data_t *data)
 		data->saved_export = NULL;
 	}
 
+	/* If there was a saved_pnfs_ds is present, release reference. */
+	if (data->saved_pnfs_ds != NULL) {
+		pnfs_ds_put(data->saved_pnfs_ds);
+		data->saved_pnfs_ds = NULL;
+	}
+
 	if (data->currentFH.nfs_fh4_val != NULL)
 		gsh_free(data->currentFH.nfs_fh4_val);
 
