@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -41,7 +42,7 @@
 
 int get_raddr(SVCXPRT *xprt)
 {
-	struct sockaddr_storage *ss = svc_getrpclocal(xprt);
+	sockaddr_t *ss = svc_getrpclocal(xprt);
 	int addr = 0;
 
 	if (ss == NULL)
@@ -85,6 +86,8 @@ int get_raddr(SVCXPRT *xprt)
 
 char cid_server_owner[MAXNAMLEN+1]; /* max hostname length */
 char *cid_server_scope_suffix = "_NFS-Ganesha";
+char *cid_server_scope;
+int owner_len, scope_len, ss_suffix_len;
 
 /**
  * @brief The NFS4_OP_EXCHANGE_ID operation
@@ -107,7 +110,6 @@ enum nfs_req_result nfs4_op_exchange_id(struct nfs_argop4 *op,
 	nfs_client_id_t *conf;
 	nfs_client_id_t *unconf;
 	int rc;
-	int owner_len, scope_len;
 	char *temp;
 	bool update;
 	uint32_t pnfs_flags;
@@ -142,19 +144,37 @@ enum nfs_req_result nfs4_op_exchange_id(struct nfs_argop4 *op,
 
 	if (cid_server_owner[0] == '\0') {
 		/* Set up the server owner string */
-		if (gethostname(cid_server_owner,
-				sizeof(cid_server_owner)) == -1) {
+		if (gsh_gethostname(cid_server_owner,
+				sizeof(cid_server_owner),
+				nfs_param.core_param.enable_AUTHSTATS)
+				== -1) {
 			res_EXCHANGE_ID4->eir_status = NFS4ERR_SERVERFAULT;
 			return NFS_REQ_ERROR;
 		}
+
+		owner_len = strlen(cid_server_owner);
+
+		/* use server_owner as server_scope if server_scope not
+		 * mentioned in main config file
+		 */
+		if (nfs_param.nfsv4_param.server_scope == NULL) {
+			ss_suffix_len = strlen(cid_server_scope_suffix);
+			scope_len = owner_len + ss_suffix_len;
+			cid_server_scope = gsh_malloc(scope_len + 1);
+			memcpy(cid_server_scope, cid_server_owner, owner_len);
+			memcpy(cid_server_scope + owner_len,
+					cid_server_scope_suffix,
+					ss_suffix_len + 1);
+		} else {
+			cid_server_scope = nfs_param.nfsv4_param.server_scope;
+			scope_len = strlen(cid_server_scope);
+		}
 	}
 
 	/* Now check that the response will fit. Use 0 for
 	 * eir_server_impl_id_len
 	 */
-	owner_len = strlen(cid_server_owner);
-	scope_len = strlen(cid_server_scope_suffix);
-	resp_size += RNDUP(owner_len) + RNDUP(owner_len + scope_len) + 0;
+	resp_size += RNDUP(owner_len) + RNDUP(scope_len) + 0;
 
 	res_EXCHANGE_ID4->eir_status = check_resp_room(data, resp_size);
 
@@ -406,12 +426,11 @@ enum nfs_req_result nfs4_op_exchange_id(struct nfs_argop4 *op,
 
 	res_EXCHANGE_ID4_ok->eir_server_owner.so_minor_id = 0;
 
-	temp = gsh_malloc(owner_len + scope_len + 1);
-	memcpy(temp, cid_server_owner, owner_len);
-	memcpy(temp + owner_len, cid_server_scope_suffix, scope_len + 1);
+	temp = gsh_malloc(scope_len + 1);
+	memcpy(temp, cid_server_scope, scope_len + 1);
 
 	res_EXCHANGE_ID4_ok->eir_server_scope.eir_server_scope_len =
-	    owner_len + scope_len + 1;
+		scope_len + 1;
 	res_EXCHANGE_ID4_ok->eir_server_scope.eir_server_scope_val = temp;
 
 	res_EXCHANGE_ID4_ok->eir_server_impl_id.eir_server_impl_id_len = 0;
@@ -437,7 +456,7 @@ enum nfs_req_result nfs4_op_exchange_id(struct nfs_argop4 *op,
 }
 
 /**
- * @brief free memory alocated for nfs4_op_exchange_id result
+ * @brief free memory allocated for nfs4_op_exchange_id result
  *
  * @param[in,out] resp Pointer to nfs4_op results
  */
