@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -91,7 +92,7 @@ enum nfs_req_result nfs4_op_lock(struct nfs_argop4 *op,
 	nfs_client_id_t *clientid = NULL;
 	/* Name for the lock owner */
 	state_nfs4_owner_name_t owner_name;
-	/* Description of requrested lock */
+	/* Description of requested lock */
 	fsal_lock_param_t lock_desc;
 	/* Description of conflicting lock */
 	fsal_lock_param_t conflict_desc;
@@ -104,7 +105,7 @@ enum nfs_req_result nfs4_op_lock(struct nfs_argop4 *op,
 	bool_t have_grace_ref = false;
 	int rc;
 	struct fsal_obj_handle *obj = data->current_obj;
-	bool state_lock_held = false;
+	bool st_lock_held = false;
 	uint64_t maxfilesize =
 	    op_ctx->fsal_export->exp_ops.fs_maxfilesize(op_ctx->fsal_export);
 	bool new_lock_state = false;
@@ -555,12 +556,12 @@ enum nfs_req_result nfs4_op_lock(struct nfs_argop4 *op,
 			 * new stateid, we will attempt to recycle.
 			 */
 			STATELOCK_lock(obj);
-			state_lock_held = true;
+			st_lock_held = true;
 			lock_state = nfs4_State_Get_Obj(obj, lock_owner);
 		} else {
-			/* Take the state_lock now */
+			/* Take the st_lock now */
 			STATELOCK_lock(obj);
-			state_lock_held = true;
+			st_lock_held = true;
 		}
 
 		if (lock_state == NULL) {
@@ -600,9 +601,9 @@ enum nfs_req_result nfs4_op_lock(struct nfs_argop4 *op,
 
 		}
 	} else {
-		/* Take the state_lock now */
+		/* Take the st_lock now */
 		STATELOCK_lock(obj);
-		state_lock_held = true;
+		st_lock_held = true;
 	}
 
 	if (data->minorversion == 0) {
@@ -732,8 +733,8 @@ enum nfs_req_result nfs4_op_lock(struct nfs_argop4 *op,
 	if (have_grace_ref)
 		nfs_put_grace_status();
 
-	if (state_lock_held) {
-		/* Now release the state_lock */
+	if (st_lock_held) {
+		/* Now release the st_lock */
 		STATELOCK_unlock(obj);
 	}
 
