@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -189,11 +190,11 @@ static nfsstat4 open4_validate_claim(compound_data_t *data,
  * This function finds or creates an owner to be associated with the
  * requested open state.
  *
- * @param[in]     arg      Arguments to OPEN4 operation
+ * @param[in]     op       Arguments to OPEN4 operation
  * @param[in,out] data     Compound's data
  * @param[out]    res      Response to OPEN4 operation
- * @param[in]     clientid Clientid record for this request
- * @param[out]    owner    The found/created owner owner
+ * @param[in]     clientid The cientid record for this request
+ * @param[out]    owner    The found/created owner
  *
  * @return false if error or replay (res_OPEN4->status is already set),
  *         true otherwise.
@@ -405,13 +406,13 @@ find_state:
  * Then attempt to get a LEASE lock to delegate the file
  * according to whether the client opened READ or READ/WRITE.
  *
- * @note state_lock must be held for WRITE
+ * @note st_lock must be held
  *
  * @param[in] data Compound data for this request
  * @param[in] op NFS arguments for the request
  * @param[in] open_state Open state for the inode to be delegated.
  * @param[in] openowner Open owner of the open state.
- * @param[in] client Client that will own the delegation.
+ * @param[in] client The client that will own the delegation.
  * @param[in/out] resok Delegation attempt result to be returned to client.
  * @param[in] prerecall flag for reclaims.
  */
@@ -569,7 +570,7 @@ static void do_delegation(OPEN4args *arg_OPEN4, OPEN4res *res_OPEN4,
 
 	/* Check if delegations are supported */
 	if (!deleg_supported(data->current_obj, op_ctx->fsal_export,
-			     op_ctx->export_perms, arg_OPEN4->share_access)) {
+			     &op_ctx->export_perms, arg_OPEN4->share_access)) {
 		resok->delegation.open_delegation4_u.od_whynone.ond_why =
 							WND4_NOT_SUPP_FTYPE;
 		LogFullDebug(COMPONENT_STATE, "Delegation type not supported.");
@@ -600,6 +601,7 @@ static void do_delegation(OPEN4args *arg_OPEN4, OPEN4res *res_OPEN4,
  * @param[in,out] verifier    The verifier for exclusive create
  * @param[in,out] createmode  The method of create
  * @param[in,out] sattr       The attributes to set
+ * @param[in,out] attrset     The attributes set on open
  *
  */
 
@@ -608,7 +610,8 @@ static void open4_ex_create_args(OPEN4args *arg,
 				 OPEN4res *res_OPEN4,
 				 void *verifier,
 				 enum fsal_create_mode *createmode,
-				 struct attrlist *sattr)
+				 struct fsal_attrlist *sattr,
+				 struct bitmap4 *attrset)
 {
 	createhow4 *createhow = &arg->openhow.openflag4_u.how;
 	fattr4 *arg_attrs = NULL;
@@ -668,12 +671,29 @@ static void open4_ex_create_args(OPEN4args *arg,
 			}
 		}
 
+		/* Now fill in attrset with the attributes requested to be set.
+		 */
+		memcpy(attrset, &arg_attrs->attrmask, sizeof(*attrset));
+
 		/* If owner or owner_group are set, and the credential was
 		 * squashed, then we must squash the set owner and owner_group.
 		 */
 		squash_setattr(sattr);
 	}
 
+
+	if (createhow->mode == EXCLUSIVE4 ||
+	    createhow->mode == EXCLUSIVE4_1) {
+		/* For an exclusive create we must indicate atime and mtime have
+		 * been set for the verifier. The client will use this
+		 * indication to know it must use a separate SETATTR to set
+		 * these attributes. This is especially important for
+		 * EXCLUSIVE4_1 creates that come with attributes.
+		 */
+		set_attribute_in_bitmap(attrset, FATTR4_TIME_ACCESS);
+		set_attribute_in_bitmap(attrset, FATTR4_TIME_MODIFY);
+	}
+
 	if (!(sattr->valid_mask & ATTR_MODE)) {
 		/* Make sure mode is set, even for exclusive create. */
 		sattr->mode = 0600;
@@ -684,7 +704,7 @@ static void open4_ex_create_args(OPEN4args *arg,
 /**
  * @brief NFS4_OP_OPEN processing using extended FSAL API
  *
- * This function impelments the NFS4_OP_OPEN operation, which
+ * This function implements the NFS4_OP_OPEN operation, which
  * potentially creates and opens a regular file.
  *
  * @param[in]     arg         Arguments for nfs4_op
@@ -692,6 +712,7 @@ static void open4_ex_create_args(OPEN4args *arg,
  * @param[out]    res_OPEN4   Results for nfs4_op
  * @param[out]    file_state  state_t created for this operation
  * @param[out]    new_state   Indicates if the state_t is new
+ * @param[in,out] attrset     The attrset bitmap to be passed down
  *
  */
 
@@ -701,7 +722,8 @@ static void open4_ex(OPEN4args *arg,
 		     nfs_client_id_t *clientid,
 		     state_owner_t *owner,
 		     state_t **file_state,
-		     bool *new_state)
+		     bool *new_state,
+		     struct bitmap4 *attrset)
 {
 	/* Parent directory in which to open the file. */
 	struct fsal_obj_handle *parent = NULL;
@@ -718,14 +740,14 @@ static void open4_ex(OPEN4args *arg,
 	enum fsal_create_mode createmode = FSAL_NO_CREATE;
 	/* The filename to create */
 	char *filename = NULL;
-	/* The supplied calim type */
+	/* The supplied claim type */
 	open_claim_type4 claim = arg->claim.claim;
 	fsal_verifier_t verifier;
-	struct attrlist sattr;
+	struct fsal_attrlist sattr;
 	/* Status for fsal calls */
 	fsal_status_t status = {0, 0};
 	/* The open state for the file */
-	bool state_lock_held = false;
+	bool st_lock_held = false;
 
 	/* Make sure the attributes are initialized */
 	memset(&sattr, 0, sizeof(sattr));
@@ -762,10 +784,10 @@ static void open4_ex(OPEN4args *arg,
 
 		filename = arg->claim.open_claim4_u.file.utf8string_val;
 
-		/* Set the createmode if appropriate) */
+		/* Set the createmode if appropriate), pass down attrset */
 		if (arg->openhow.opentype == OPEN4_CREATE) {
 			open4_ex_create_args(arg, data, res_OPEN4, verifier,
-					     &createmode, &sattr);
+					     &createmode, &sattr, attrset);
 
 			if (res_OPEN4->status != NFS4_OK)
 				goto out;
@@ -894,7 +916,7 @@ static void open4_ex(OPEN4args *arg,
 	if (file_obj != NULL) {
 		/* Go ahead and take the state lock now. */
 		STATELOCK_lock(file_obj);
-		state_lock_held = true;
+		st_lock_held = true;
 
 		in_obj = file_obj;
 
@@ -1016,13 +1038,13 @@ static void open4_ex(OPEN4args *arg,
 	}
 
 	if (file_obj == NULL) {
-		/* We have a new cache inode entry, take the state lock. */
+		/* We have a file object, take the state lock. */
 		file_obj = out_obj;
 		STATELOCK_lock(file_obj);
-		state_lock_held = true;
+		st_lock_held = true;
 	}
 
-	/* Now the state_lock is held for sure and we have an extra LRU
+	/* Now the st_lock is held for sure and we have an extra LRU
 	 * reference to file_obj, which is the opened file.
 	 */
 
@@ -1099,11 +1121,11 @@ static void open4_ex(OPEN4args *arg,
 					fsal_err_txt(status));
 			}
 
-			/* Need to release the state_lock before the put_ref
+			/* Need to release the st_lock before the put_ref
 			 * call.
 			 */
 			STATELOCK_unlock(file_obj);
-			state_lock_held = false;
+			st_lock_held = false;
 
 			/* Release the extra LRU reference on file_obj. */
 			file_obj->obj_ops->put_ref(file_obj);
@@ -1150,10 +1172,10 @@ static void open4_ex(OPEN4args *arg,
 			     (*file_state)->state_data.share.share_deny_prev);
 	}
 
-	if (!state_lock_held) {
-		/* Acquire state_lock before adding deleg state */
+	if (!st_lock_held) {
+		/* Acquire st_lock before adding deleg state */
 		STATELOCK_lock(file_obj);
-		state_lock_held = true;
+		st_lock_held = true;
 	}
 
 	do_delegation(arg, res_OPEN4, data, owner, *file_state, clientid);
@@ -1162,7 +1184,7 @@ static void open4_ex(OPEN4args *arg,
 	/* Release the attributes (may release an inherited ACL) */
 	fsal_release_attrs(&sattr);
 
-	if (state_lock_held) {
+	if (st_lock_held) {
 		STATELOCK_unlock(file_obj);
 	}
 
@@ -1186,7 +1208,7 @@ static void open4_ex(OPEN4args *arg,
 /**
  * @brief NFS4_OP_OPEN
  *
- * This function impelments the NFS4_OP_OPEN operation, which
+ * This function implements the NFS4_OP_OPEN operation, which
  * potentially creates and opens a regular file.
  *
  * @param[in]     op   Arguments for nfs4_op
@@ -1218,7 +1240,7 @@ enum nfs_req_result nfs4_op_open(struct nfs_argop4 *op, compound_data_t *data,
 	nfs_client_id_t *clientid = NULL;
 	/* The found or created state owner for this open */
 	state_owner_t *owner = NULL;
-	/* The supplied calim type */
+	/* The supplied claim type */
 	open_claim_type4 claim = arg_OPEN4->claim.claim;
 	/* The open state for the file */
 	state_t *file_state = NULL;
@@ -1245,7 +1267,7 @@ enum nfs_req_result nfs4_op_open(struct nfs_argop4 *op, compound_data_t *data,
 
 	/* Check export permissions if OPEN4_CREATE */
 	if ((arg_OPEN4->openhow.opentype == OPEN4_CREATE) &&
-	    ((op_ctx->export_perms->options &
+	    ((op_ctx->export_perms.options &
 	      EXPORT_OPTION_MD_WRITE_ACCESS) == 0)) {
 		res_OPEN4->status = NFS4ERR_ROFS;
 
@@ -1257,7 +1279,7 @@ enum nfs_req_result nfs4_op_open(struct nfs_argop4 *op, compound_data_t *data,
 
 	/* Check export permissions if OPEN4_SHARE_ACCESS_WRITE */
 	if (((arg_OPEN4->share_access & OPEN4_SHARE_ACCESS_WRITE) != 0) &&
-	    ((op_ctx->export_perms->options &
+	    ((op_ctx->export_perms.options &
 	      EXPORT_OPTION_WRITE_ACCESS) == 0)) {
 		res_OPEN4->status = NFS4ERR_ROFS;
 
@@ -1343,7 +1365,7 @@ enum nfs_req_result nfs4_op_open(struct nfs_argop4 *op, compound_data_t *data,
 		goto out2;
 	}
 
-	/* So we still have a reference even after we repalce the
+	/* So we still have a reference even after we replace the
 	 * current FH.
 	 */
 	obj_change = data->current_obj;
@@ -1369,24 +1391,20 @@ enum nfs_req_result nfs4_op_open(struct nfs_argop4 *op, compound_data_t *data,
 		goto out;
 	}
 
-	/* Utilize the extended FSAL APU functionality to perform the open. */
-	open4_ex(arg_OPEN4, data, res_OPEN4, clientid,
-		 owner, &file_state, &new_state);
-
-	if (res_OPEN4->status != NFS4_OK)
-		goto out;
-
+	/* Prepare the attrset attribute */
 	memset(&res_OPEN4->OPEN4res_u.resok4.attrset,
 	       0,
 	       sizeof(struct bitmap4));
 
-	if (arg_OPEN4->openhow.openflag4_u.how.mode == EXCLUSIVE4 ||
-	    arg_OPEN4->openhow.openflag4_u.how.mode == EXCLUSIVE4_1) {
-		struct bitmap4 *bits = &res_OPEN4->OPEN4res_u.resok4.attrset;
+	/* Utilize the extended FSAL API functionality to perform the open.
+	 * The usage of atime/mtime for exclusive create verifier will be
+	 * handled in open4_ex, so pass in attrset to be handled there.
+	 */
+	open4_ex(arg_OPEN4, data, res_OPEN4, clientid, owner, &file_state,
+		 &new_state, &res_OPEN4->OPEN4res_u.resok4.attrset);
 
-		set_attribute_in_bitmap(bits, FATTR4_TIME_ACCESS);
-		set_attribute_in_bitmap(bits, FATTR4_TIME_MODIFY);
-	}
+	if (res_OPEN4->status != NFS4_OK)
+		goto out;
 
 	/* If server use OPEN_CONFIRM4, set the correct flag,
 	 * but not for 4.1 */
