@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -80,9 +81,8 @@ static enum nfs_req_result nfs4_complete_read(struct nfs4_read_data *data)
 	struct fsal_io_arg *read_arg = &data->read_arg;
 
 	if (data->res_READ4->status == NFS4_OK) {
-		if (!read_arg->end_of_file) {
-			/** @todo FSF: add a config option for this behavior?
-			*/
+		if (nfs_param.core_param.getattrs_in_complete_read &&
+				!read_arg->end_of_file) {
 			/*
 			 * NFS requires to set the EOF flag for all reads that
 			 * reach the EOF, i.e., even the ones returning data.
@@ -91,7 +91,7 @@ static enum nfs_req_result nfs4_complete_read(struct nfs4_read_data *data)
 			 * will just see a short read and continue reading and
 			 * then get the EOF flag as 0 bytes are returned.
 			 */
-			struct attrlist attrs;
+			struct fsal_attrlist attrs;
 			fsal_status_t status;
 
 			fsal_prepare_attrs(&attrs, ATTR_SIZE);
@@ -252,7 +252,7 @@ static void nfs4_read_cb(struct fsal_obj_handle *obj, fsal_status_t ret,
 /**
  * @brief Read on a pNFS pNFS data server
  *
- * This function bypasses cache_inode and calls directly into the FSAL
+ * This function bypasses mdcache and calls directly into the FSAL
  * to perform a data-server read.
  *
  * @param[in]     op   Arguments for nfs41_op
@@ -293,9 +293,8 @@ static enum nfs_req_result op_dsread(struct nfs_argop4 *op,
 
 	res_READ4->READ4res_u.resok4.data.data_val = buffer;
 
-	nfs_status = data->current_ds->dsh_ops.read(
+	nfs_status = op_ctx->ctx_pnfs_ds->s_ops.dsh_read(
 				data->current_ds,
-				op_ctx,
 				&arg_READ4->stateid,
 				arg_READ4->offset,
 				arg_READ4->count,
@@ -321,7 +320,7 @@ static enum nfs_req_result op_dsread(struct nfs_argop4 *op,
 /**
  * @brief Read on a pNFS pNFS data server
  *
- * This function bypasses cache_inode and calls directly into the FSAL
+ * This function bypasses mdcache and calls directly into the FSAL
  * to perform a data-server read.
  *
  * @param[in]     op   Arguments for nfs41_op
@@ -364,9 +363,8 @@ static enum nfs_req_result op_dsread_plus(struct nfs_argop4 *op,
 
 	buffer = gsh_malloc_aligned(4096, RNDUP(arg_READ4->count));
 
-	nfs_status = data->current_ds->dsh_ops.read_plus(
+	nfs_status = op_ctx->ctx_pnfs_ds->s_ops.dsh_read_plus(
 				data->current_ds,
-				op_ctx,
 				&arg_READ4->stateid,
 				arg_READ4->offset,
 				arg_READ4->count,
@@ -495,7 +493,7 @@ static enum nfs_req_result nfs4_read(struct nfs_argop4 *op,
 		    && (state_open->state_data.share.share_access &
 		    OPEN4_SHARE_ACCESS_READ) == 0) {
 			/* Even if file is open for write, the client
-			 * may do accidently read operation (caching).
+			 * may do accidentally read operation (caching).
 			 * Because of this, READ is allowed if not
 			 * explicitly denied.  See page 112 in RFC 7530
 			 * for more details.
@@ -600,7 +598,7 @@ static enum nfs_req_result nfs4_read(struct nfs_argop4 *op,
 
 		if ((offset + size) > MaxOffsetRead) {
 			LogEvent(COMPONENT_NFS_V4,
-				 "A client tryed to violate max file size %"
+				 "A client tried to violate max file size %"
 				 PRIu64 " for exportid #%hu",
 				 MaxOffsetRead,
 				 op_ctx->ctx_export->export_id);
@@ -1046,8 +1044,7 @@ enum nfs_req_result nfs4_op_seek(struct nfs_argop4 *op,
 		if (arg_SEEK->sa_what == NFS4_CONTENT_DATA ||
 				arg_SEEK->sa_what == NFS4_CONTENT_HOLE) {
 			info.io_content.hole.di_offset = arg_SEEK->sa_offset;
-}
-		else
+		} else
 			info.io_content.adb.adb_offset = arg_SEEK->sa_offset;
 
 		fsal_status = obj->obj_ops->seek2(obj, state_found, &info);
