@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -154,7 +155,7 @@ static void nfs4_write_cb(struct fsal_obj_handle *obj, fsal_status_t ret,
 /**
  * @brief Write for a data server
  *
- * This function bypasses cache_inode and calls directly into the FSAL
+ * This function bypasses mdcache and calls directly into the FSAL
  * to perform a pNFS data server write.
  *
  * @param[in]     op    Arguments for nfs41_op
@@ -174,9 +175,8 @@ static enum nfs_req_result op_dswrite(struct nfs_argop4 *op,
 	/* NFSv4 return code */
 	nfsstat4 nfs_status = 0;
 
-	res_WRITE4->status = data->current_ds->dsh_ops.write(
+	res_WRITE4->status = op_ctx->ctx_pnfs_ds->s_ops.dsh_write(
 				data->current_ds,
-				op_ctx,
 				&arg_WRITE4->stateid,
 				arg_WRITE4->offset,
 				arg_WRITE4->data.data_len,
@@ -221,7 +221,7 @@ enum nfs_req_result nfs4_op_write(struct nfs_argop4 *op, compound_data_t *data,
 		atomic_fetch_uint64_t(&op_ctx->ctx_export->MaxWrite);
 	uint64_t MaxOffsetWrite =
 		atomic_fetch_uint64_t(&op_ctx->ctx_export->MaxOffsetWrite);
-	bool force_sync = op_ctx->export_perms->options & EXPORT_OPTION_COMMIT;
+	bool force_sync = op_ctx->export_perms.options & EXPORT_OPTION_COMMIT;
 	struct nfs4_write_data *write_data = NULL;
 	struct fsal_io_arg *write_arg;
 	/* In case we don't call write2, we indicate the I/O as already done
@@ -250,7 +250,7 @@ enum nfs_req_result nfs4_op_write(struct nfs_argop4 *op, compound_data_t *data,
 	   allows inode creation or not */
 	fsal_status = op_ctx->fsal_export->exp_ops.check_quota(
 						op_ctx->fsal_export,
-						op_ctx->ctx_export->fullpath,
+						CTX_FULLPATH(op_ctx),
 						FSAL_QUOTA_BLOCKS);
 
 	if (FSAL_IS_ERROR(fsal_status)) {
@@ -384,7 +384,7 @@ enum nfs_req_result nfs4_op_write(struct nfs_argop4 *op, compound_data_t *data,
 
 		if ((offset + size) > MaxOffsetWrite) {
 			LogEvent(COMPONENT_NFS_V4,
-				 "A client tryed to violate max file size %"
+				 "A client tried to violate max file size %"
 				 PRIu64 " for exportid #%hu",
 				 MaxOffsetWrite,
 				 op_ctx->ctx_export->export_id);
