@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -303,8 +304,6 @@ void dec_nlm_state_ref(state_t *state)
 
 	LogFullDebug(COMPONENT_STATE, "Free {%s}", str);
 
-	dec_state_owner_ref(state->state_owner);
-
 	put_gsh_export(state->state_export);
 
 	obj = get_state_obj_ref(state);
@@ -320,7 +319,7 @@ void dec_nlm_state_ref(state_t *state)
 
 	state->state_exp->exp_ops.free_state(state->state_exp, state);
 
-	/* Release 2 refs: our sentinal one, plus the one from
+	/* Release 2 refs: our sentinel one, plus the one from
 	 * get_state_obj_ref() */
 	obj->obj_ops->put_ref(obj);
 	obj->obj_ops->put_ref(obj);
@@ -490,8 +489,6 @@ int get_nlm_state(enum state_type state_type,
 
 	get_gsh_export_ref(state->state_export);
 
-	inc_state_owner_ref(state->state_owner);
-
 	*pstate = state;
 
 	return 0;
