@@ -1,3 +1,22 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
+/*
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 3 of the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+ *
+ */
+
 #include "config.h"
 #include "log.h"
 #include "nfs_core.h"
@@ -12,12 +31,10 @@
 #include "fsal.h"
 #include "recovery_fs.h"
 
-#define NFS_V4_OLD_DIR "v4old"
-
 char v4_recov_dir[PATH_MAX];
-int v4_recov_dir_len;
+unsigned int v4_recov_dir_len;
 char v4_old_dir[PATH_MAX];
-int v4_old_dir_len;
+unsigned int v4_old_dir_len;
 
 /**
  * @brief convert clientid opaque bytes as a hex string for mkdir purpose.
@@ -137,42 +154,41 @@ static void fs_create_clid_name(nfs_client_id_t *clientid)
 int fs_create_recov_dir(void)
 {
 	int err, root_len, dir_len, old_len, node_size = 0;
-	char *nfs_v4_recov_root = NFS_V4_RECOV_ROOT;
-	char *nfs_v4_recov_dir = NFS_V4_RECOV_DIR;
-	char *nfs_v4_old_dir = NFS_V4_OLD_DIR;
 	char node[14];
 
 	if (nfs_param.core_param.clustered) {
-		node_size = snprintf(node, sizeof(node), "node%d", g_nodeid)
-			    + 1;
+		node_size = snprintf(node, sizeof(node), "node%d", g_nodeid);
 
 		if (unlikely(node_size >= sizeof(node) || node_size < 0)) {
 			LogFatal(COMPONENT_CLIENTID,
 				 "snprintf returned unexpected %d", node_size);
 		}
-		/* Now include the NUL */
+		/* Now include the '/' */
 		node_size++;
 	}
 
-	err = mkdir(nfs_v4_recov_root, 0755);
+	err = mkdir(nfs_param.nfsv4_param.recov_root, 0755);
 	if (err == -1 && errno != EEXIST) {
 		LogEvent(COMPONENT_CLIENTID,
 			 "Failed to create v4 recovery dir (%s), errno: %s (%d)",
-			 nfs_v4_recov_root, strerror(errno), errno);
+			 nfs_param.nfsv4_param.recov_root,
+			 strerror(errno), errno);
 	}
 
-	root_len = strlen(nfs_v4_recov_root);
-	dir_len = strlen(nfs_v4_recov_dir);
-	v4_recov_dir_len = root_len + 1 + dir_len + node_size - 1;
+	root_len = strlen(nfs_param.nfsv4_param.recov_root);
+	dir_len = strlen(nfs_param.nfsv4_param.recov_dir);
+	v4_recov_dir_len = root_len + 1 + dir_len + node_size;
 
 	if (v4_recov_dir_len >= sizeof(v4_recov_dir))
 		LogFatal(COMPONENT_CLIENTID,
 			 "v4 recovery dir path (%s/%s) is to long",
-			 nfs_v4_recov_root, nfs_v4_recov_dir);
+			 nfs_param.nfsv4_param.recov_root,
+			 nfs_param.nfsv4_param.recov_dir);
 
-	memcpy(v4_recov_dir, nfs_v4_recov_root, root_len);
+	memcpy(v4_recov_dir, nfs_param.nfsv4_param.recov_root, root_len);
 	v4_recov_dir[root_len] = '/';
-	memcpy(v4_recov_dir + 1 + root_len, nfs_v4_recov_dir, dir_len + 1);
+	memcpy(v4_recov_dir + 1 + root_len,
+	       nfs_param.nfsv4_param.recov_dir, dir_len + 1);
 	dir_len = 1 + root_len + dir_len;
 
 	LogDebug(COMPONENT_CLIENTID, "v4_recov_dir=%s", v4_recov_dir);
@@ -184,18 +200,20 @@ int fs_create_recov_dir(void)
 			 v4_recov_dir, strerror(errno), errno);
 	}
 
-	root_len = strlen(nfs_v4_recov_root);
-	old_len = strlen(nfs_v4_old_dir);
-	v4_old_dir_len = root_len + 1 + dir_len + node_size - 1;
+	root_len = strlen(nfs_param.nfsv4_param.recov_root);
+	old_len = strlen(nfs_param.nfsv4_param.recov_old_dir);
+	v4_old_dir_len = root_len + 1 + dir_len + node_size;
 
 	if (v4_old_dir_len >= sizeof(v4_old_dir))
 		LogFatal(COMPONENT_CLIENTID,
 			 "v4 recovery dir path (%s/%s) is to long",
-			 nfs_v4_recov_root, nfs_v4_old_dir);
+			 nfs_param.nfsv4_param.recov_root,
+			 nfs_param.nfsv4_param.recov_old_dir);
 
-	memcpy(v4_old_dir, nfs_v4_recov_root, root_len);
+	memcpy(v4_old_dir, nfs_param.nfsv4_param.recov_root, root_len);
 	v4_old_dir[root_len] = '/';
-	memcpy(v4_old_dir + 1 + root_len, nfs_v4_old_dir, old_len + 1);
+	memcpy(v4_old_dir + 1 + root_len,
+	       nfs_param.nfsv4_param.recov_old_dir, old_len + 1);
 	old_len = 1 + root_len + old_len;
 
 	LogDebug(COMPONENT_CLIENTID, "v4_old_dir=%s", v4_old_dir);
@@ -211,8 +229,10 @@ int fs_create_recov_dir(void)
 		/* Now make the node specific directories */
 		v4_recov_dir[dir_len] = '/';
 		v4_old_dir[old_len] = '/';
-		memcpy(v4_recov_dir + 1 + dir_len, node, node_size);
-		memcpy(v4_old_dir + 1 + old_len, node, node_size);
+
+		/* Copy an extra byte to NUL terminate */
+		memcpy(v4_recov_dir + 1 + dir_len, node, node_size + 1);
+		memcpy(v4_old_dir + 1 + old_len, node, node_size + 1);
 
 		LogDebug(COMPONENT_CLIENTID, "v4_recov_dir=%s", v4_recov_dir);
 		LogDebug(COMPONENT_CLIENTID, "v4_old_dir=%s", v4_old_dir);
@@ -233,6 +253,13 @@ int fs_create_recov_dir(void)
 				 v4_old_dir, strerror(errno), errno);
 		}
 	}
+
+
+	LogInfo(COMPONENT_CLIENTID,
+		"NFSv4 Recovery Directory %s", v4_recov_dir);
+	LogInfo(COMPONENT_CLIENTID,
+		"NFSv4 Recovery Directory (old) %s", v4_old_dir);
+
 	return 0;
 }
 
@@ -300,7 +327,7 @@ void fs_add_clid(nfs_client_id_t *clientid)
  * @brief Remove the revoked file handles created under a specific
  * client-id path on the stable storage.
  *
- * @param[in] path Path of the client-id on the stable storage.
+ * @param[in] path The path of the client-id on the stable storage.
  */
 static void fs_rm_revoked_handles(char *path)
 {
@@ -374,7 +401,8 @@ static void fs_rm_clid_impl(int position,
 
 	memcpy(path, parent_path, parent_len);
 	path[parent_len] = '/';
-	memcpy(path + parent_len + 1, recov_dir + position, segment_len + 1);
+	memcpy(path + parent_len + 1, recov_dir + position, segment_len);
+	path[total_len - 1] = '\0';
 
 	/* recursively remove the directory hirerchy which represent the
 	 *clientid
@@ -412,14 +440,14 @@ void fs_rm_clid(nfs_client_id_t *clientid)
  * @brief Copy and Populate revoked delegations for this client.
  *
  * Even after delegation revoke, it is possible for the client to
- * contiue its leas and other operatoins. Sever saves revoked delegations
+ * continue its lease and other operations. Sever saves revoked delegations
  * in the memory so client will not be granted same delegation with
  * DELEG_CUR ; but it is possible that the server might reboot and has
  * no record of the delegatin. This list helps to reject delegations
  * client is obtaining through DELEG_PREV.
  *
- * @param[in] clientid Clientid that is being created.
- * @param[in] path Path of the directory structure.
+ * @param[in] clientid The clientid that is being created.
+ * @param[in] path The path of the directory structure.
  * @param[in] Target dir to copy.
  * @param[in] del Delete after populating
  */
@@ -596,7 +624,7 @@ static int fs_read_recov_clids_impl(const char *parent_path,
 
 			if ((rc == -1) && (errno != EEXIST)) {
 				LogEvent(COMPONENT_CLIENTID,
-					 "mkdir %s faied errno: %s (%d)",
+					 "mkdir %s failed errno: %s (%d)",
 					 new_path, strerror(errno), errno);
 			}
 		}
@@ -765,14 +793,14 @@ void fs_read_recov_clids_takeover(nfs_grace_start_t *gsp,
 		break;
 	case EVENT_TAKE_IP:
 		rc = snprintf(path, sizeof(path), "%s/%s/%s",
-			      NFS_V4_RECOV_ROOT, gsp->ipaddr,
-			      NFS_V4_RECOV_DIR);
+			      nfs_param.nfsv4_param.recov_root, gsp->ipaddr,
+			      nfs_param.nfsv4_param.recov_dir);
 
 		if (unlikely(rc >= sizeof(path))) {
 			LogCrit(COMPONENT_CLIENTID,
 				"Path %s/%s/%s too long",
-				NFS_V4_RECOV_ROOT, gsp->ipaddr,
-				NFS_V4_RECOV_DIR);
+				nfs_param.nfsv4_param.recov_root, gsp->ipaddr,
+				nfs_param.nfsv4_param.recov_dir);
 			return;
 		} else if (unlikely(rc < 0)) {
 			LogCrit(COMPONENT_CLIENTID,
@@ -782,13 +810,15 @@ void fs_read_recov_clids_takeover(nfs_grace_start_t *gsp,
 		break;
 	case EVENT_TAKE_NODEID:
 		rc = snprintf(path, sizeof(path), "%s/%s/node%d",
-			      NFS_V4_RECOV_ROOT, NFS_V4_RECOV_DIR,
+			      nfs_param.nfsv4_param.recov_root,
+			      nfs_param.nfsv4_param.recov_dir,
 			      gsp->nodeid);
 
 		if (unlikely(rc >= sizeof(path))) {
 			LogCrit(COMPONENT_CLIENTID,
 				"Path %s/%s/node%d too long",
-				NFS_V4_RECOV_ROOT, NFS_V4_RECOV_DIR,
+				nfs_param.nfsv4_param.recov_root,
+				nfs_param.nfsv4_param.recov_dir,
 				gsp->nodeid);
 			return;
 		} else if (unlikely(rc < 0)) {
@@ -889,7 +919,7 @@ void fs_add_revoke_fh(nfs_client_id_t *delr_clid, nfs_fh4 *delr_handle)
 
 	assert(v4_recov_dir_len < sizeof(path));
 
-	memcpy(path, v4_recov_dir, v4_recov_dir_len + 1);
+	memcpy(path, v4_recov_dir, v4_recov_dir_len);
 	pathpos = v4_recov_dir_len;
 
 	length = strlen(delr_clid->cid_recov_tag);
@@ -916,7 +946,6 @@ void fs_add_revoke_fh(nfs_client_id_t *delr_clid, nfs_fh4 *delr_handle)
 			memcpy(path + pathpos + len + 2,
 			       rhdlstr,
 			       rhdlstr_len + 1);
-			pathpos = new_pathpos;
 			fd = creat(path, 0700);
 			if (fd < 0) {
 				LogEvent(COMPONENT_CLIENTID,
