@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -294,6 +295,8 @@ state_status_t state_error_convert(fsal_status_t fsal_status)
 	case ERR_FSAL_NO_ACE:
 	case ERR_FSAL_CROSS_JUNCTION:
 	case ERR_FSAL_BADNAME:
+	case ERR_FSAL_NOXATTR:
+	case ERR_FSAL_XATTR2BIG:
 		/* These errors should be handled inside state
 		 * (or should never be seen by state)
 		 */
@@ -781,7 +784,7 @@ int display_owner(struct display_buffer *dspbuf, state_owner_t *owner)
 /**
  * @brief Acquire a reference on a state owner
  *
- * @param[in] owner Owner to acquire
+ * @param[in] owner The owner to acquire
  */
 void inc_state_owner_ref(state_owner_t *owner)
 {
@@ -806,7 +809,7 @@ void inc_state_owner_ref(state_owner_t *owner)
 /**
  * @brief Free a state owner object
  *
- * @param[in] owner Owner to free
+ * @param[in] owner The owner to free
  */
 void free_state_owner(state_owner_t *owner)
 {
@@ -856,7 +859,7 @@ void free_state_owner(state_owner_t *owner)
 /**
  * @brief Get the hash table associated with a state owner object
  *
- * @param[in] owner Owner to get associated hash table for
+ * @param[in] owner The owner to get associated hash table for
  */
 hash_table_t *get_state_owner_hash_table(state_owner_t *owner)
 {
@@ -886,7 +889,7 @@ hash_table_t *get_state_owner_hash_table(state_owner_t *owner)
 /**
  * @brief Relinquish a reference on a state owner
  *
- * @param[in] owner Owner to release
+ * @param[in] owner The owner to release
  */
 void dec_state_owner_ref(state_owner_t *owner)
 {
@@ -1216,7 +1219,7 @@ not_found:
 /**
  * @brief Acquire a reference on a state owner that might be getting freed.
  *
- * @param[in] owner Owner to acquire
+ * @param[in] owner The owner to acquire
  *
  * inc_state_owner_ref() can be called to hold a reference provided
  * someone already has a valid refcount. In other words, one can't call
@@ -1284,7 +1287,7 @@ bool hold_state_owner(state_owner_t *owner)
  * @brief Release all state on a file
  *
  * This function may not be called in any context which could hold
- * entry->state_lock.  It will now be reliably called in cleanup
+ * entry->st_lock.  It will now be reliably called in cleanup
  * processing.
  *
  * @param[in,out] obj File to be wiped
@@ -1362,18 +1365,18 @@ void dump_all_owners(void)
 
 void state_release_export(struct gsh_export *export)
 {
-	struct root_op_context root_op_context;
+	struct req_op_context op_context;
 
-	/* Initialize req_ctx */
-	init_root_op_context(&root_op_context, export, export->fsal_export,
-			     0, 0, UNKNOWN_REQUEST);
+	/* Get a ref to the export and initialize op_context */
+	get_gsh_export_ref(export);
+	init_op_context_simple(&op_context, export, export->fsal_export);
 
 	state_export_unlock_all();
 	state_export_release_nfs4_state();
 #ifdef _USE_NLM
 	state_export_unshare_all();
 #endif /* _USE_NLM */
-	release_root_op_context();
+	release_op_context();
 }
 
 /** @} */
