@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /* ----------------------------------------------------------------------------
  * Copyright CEA/DAM/DIF  (2007)
  * contributeur : Thomas LEIBOVICI  thomas.leibovici@cea.fr
@@ -298,6 +299,7 @@ static bool convert_number(struct config_node *node,
 	char *endptr;
 	int base;
 	bool signed_int = false;
+	bool zero_ok = false, inrange;
 
 	if (node->type != TYPE_TERM) {
 		config_proc_error(node, err_type,
@@ -332,38 +334,45 @@ static bool convert_number(struct config_node *node,
 	case CONFIG_INT16:
 		smin = item->u.i16.minval;
 		smax = item->u.i16.maxval;
+		zero_ok = item->u.i16.zero_ok;
 		signed_int = true;
 		break;
 	case CONFIG_UINT16:
 		mask = UINT16_MAX;
 		min = item->u.ui16.minval;
 		max = item->u.ui16.maxval;
+		zero_ok = item->u.ui16.zero_ok;
 		break;
 	case CONFIG_INT32:
 		smin = item->u.i32.minval;
 		smax = item->u.i32.maxval;
+		zero_ok = item->u.i32.zero_ok;
 		signed_int = true;
 		break;
 	case CONFIG_UINT32:
 		mask = UINT32_MAX;
 		min = item->u.ui32.minval;
 		max = item->u.ui32.maxval;
+		zero_ok = item->u.ui32.zero_ok;
 		break;
 	case CONFIG_ANON_ID:
 		/* Internal to config, anonymous id is treated as int64_t */
 		smin = item->u.i64.minval;
 		smax = item->u.i64.maxval;
+		zero_ok = item->u.i64.zero_ok;
 		signed_int = true;
 		break;
 	case CONFIG_INT64:
 		smin = item->u.i64.minval;
 		smax = item->u.i64.maxval;
+		zero_ok = item->u.i64.zero_ok;
 		signed_int = true;
 		break;
 	case CONFIG_UINT64:
 		mask = UINT64_MAX;
 		min = item->u.ui64.minval;
 		max = item->u.ui64.maxval;
+		zero_ok = item->u.ui64.zero_ok;
 		break;
 	default:
 		goto errout;
@@ -394,12 +403,17 @@ static bool convert_number(struct config_node *node,
 				  *node->u.term.op_code);
 			goto errout;
 		}
-		if (sval < smin || sval > smax) {
+
+		inrange = (sval >= smin && sval <= smax) ||
+			  (zero_ok && sval == 0);
+
+		if (!inrange) {
 			config_proc_error(node, err_type,
 				  "(%s) is out of range",
 				  node->u.term.varvalue);
 			goto errout;
 		}
+
 		val = (uint64_t) sval;
 	} else {
 		if (node->u.term.op_code != NULL &&
@@ -418,7 +432,11 @@ static bool convert_number(struct config_node *node,
 				  *node->u.term.op_code);
 			goto errout;
 		}
-		if (val < min || val > max) {
+
+		inrange = (val >= min && val <= max) ||
+			  (zero_ok && val == 0);
+
+		if (!inrange) {
 			config_proc_error(node, err_type,
 				  "(%s) is out of range",
 				  node->u.term.varvalue);
@@ -1239,7 +1257,7 @@ static int do_block_load(struct config_node *blk,
 				break;
 			case CONFIG_DEPRECATED:
 				config_proc_error(node, err_type,
-					"Deprectated parameter (%s)%s%s",
+					"Deprecated parameter (%s)%s%s",
 					item->name,
 					item->u.deprecated.message
 						? " - "
@@ -1313,7 +1331,7 @@ static int do_block_load(struct config_node *blk,
  *  This function attaches the build param struct to its link_mem.
  *  Before it does the attach, it will do validation of input if required.
  *  Returns 0 if validation passes and the attach is successful.  Otherwise,
- *  it returns an error which will trigger a release of resourcs acquired
+ *  it returns an error which will trigger a release of resources acquired
  *  by the self_struct_init.
  *
  * Both of these functions are called in the context of the link_mem parse.
@@ -1388,6 +1406,9 @@ static bool proc_block(struct config_node *node,
 				  item->name);
 		goto err_out;
 	}
+	if (item->u.blk.check && item->u.blk.check(param_struct, err_type)) {
+		goto err_out;
+	}
 	LogFullDebug(COMPONENT_CONFIG,
 		     "------ At (%s:%d): commit %s",
 		     node->filename,
@@ -1679,7 +1700,7 @@ errout:
  * @param expr   [IN] pointer to the expression to be parsed
  * @param expr_node [OUT] pointer to expression parse tree
  *
- * @return pointer to first char past parse or NULL for erros.
+ * @return pointer to first char past parse or NULL for errors.
  */
 
 static char *parse_expr(char *expr, struct expr_parse **expr_node)
@@ -1734,7 +1755,7 @@ static inline bool match_one_term(char *value, struct config_node *node)
  * once found, the token value is compared.  '*' matches anything. else
  * it is a case-insensitive string compare.
  *
- * @param blk   [IN] pointer to config_node descibing the block
+ * @param blk   [IN] pointer to config_node describing the block
  * @param expr  [IN] expr_parse node to match
  *
  * @return true if matched, else false
@@ -1883,6 +1904,7 @@ int load_config_from_node(void *tree_node,
 {
 	struct config_node *node = (struct config_node *)tree_node;
 	char *blkname = conf_blk->blk_desc.name;
+	char *altblkname = conf_blk->blk_desc.altname;
 
 	if (node == NULL) {
 		config_proc_error(NULL, err_type,
@@ -1892,7 +1914,9 @@ int load_config_from_node(void *tree_node,
 		return -1;
 	}
 	if (node->type == TYPE_BLOCK) {
-		if (strcasecmp(node->u.nterm.name, blkname) != 0) {
+		if (strcasecmp(node->u.nterm.name, blkname) != 0 &&
+		    (altblkname == NULL ||
+		     strcasecmp(node->u.nterm.name, altblkname) != 0)) {
 			config_proc_error(node, err_type,
 					  "Looking for block (%s), got (%s)",
 					  blkname, node->u.nterm.name);
@@ -1942,6 +1966,7 @@ int load_config_from_parse(config_file_t config,
 	struct config_node *node = NULL;
 	struct glist_head *ns;
 	char *blkname = conf_blk->blk_desc.name;
+	char *altblkname = conf_blk->blk_desc.altname;
 	int found = 0;
 	int prev_errs = err_type->errors;
 	void *blk_mem = NULL;
@@ -1973,7 +1998,9 @@ int load_config_from_parse(config_file_t config,
 	glist_for_each(ns, &tree->root.u.nterm.sub_nodes) {
 		node = glist_entry(ns, struct config_node, node);
 		if (node->type == TYPE_BLOCK &&
-		    strcasecmp(blkname, node->u.nterm.name) == 0) {
+		    (strcasecmp(blkname, node->u.nterm.name) == 0 ||
+		     (altblkname != NULL &&
+		      strcasecmp(altblkname, node->u.nterm.name) == 0))) {
 			if (found > 0 &&
 			    (conf_blk->blk_desc.flags & CONFIG_UNIQUE)) {
 				config_proc_error(node, err_type,
