@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: LGPL-3.0-or-later */
 /*
  *
  *
@@ -31,7 +32,7 @@
 
 /**
  * @file  fsal_commomnlib.h
- * @brief Miscelaneous FSAL common library routines
+ * @brief Miscellaneous FSAL common library routines
  */
 
 #ifndef FSAL_COMMONLIB_H
@@ -55,7 +56,7 @@ void fsal_detach_export(struct fsal_module *fsal_hdl,
 /* fsal_export common methods
  */
 
-void fsal_export_init(struct fsal_export *export);
+void fsal_export_init(struct fsal_export *);
 
 void fsal_export_stack(struct fsal_export *sub_export,
 		       struct fsal_export *super_export);
@@ -92,66 +93,6 @@ static inline bool fsal_obj_handle_is(struct fsal_obj_handle *obj_hdl,
 void fsal_pnfs_ds_init(struct fsal_pnfs_ds *pds, struct fsal_module *fsal);
 void fsal_pnfs_ds_fini(struct fsal_pnfs_ds *pds);
 
-void fsal_ds_handle_init(struct fsal_ds_handle *dsh, struct fsal_pnfs_ds *pds);
-void fsal_ds_handle_fini(struct fsal_ds_handle *dsh);
-
-int open_dir_by_path_walk(int first_fd, const char *path, struct stat *stat);
-
-struct avltree avl_fsid;
-struct avltree avl_dev;
-
-struct glist_head posix_file_systems;
-
-pthread_rwlock_t fs_lock;
-
-void free_fs(struct fsal_filesystem *fs);
-
-int populate_posix_file_systems(bool force);
-
-int reload_posix_filesystems(const char *path,
-			     struct fsal_module *fsal,
-			     struct fsal_export *exp,
-			     claim_filesystem_cb claim,
-			     unclaim_filesystem_cb unclaim,
-			     struct fsal_filesystem **root_fs);
-
-int resolve_posix_filesystem(const char *path,
-			     struct fsal_module *fsal,
-			     struct fsal_export *exp,
-			     claim_filesystem_cb claim,
-			     unclaim_filesystem_cb unclaim,
-			     struct fsal_filesystem **root_fs);
-
-void release_posix_file_systems(void);
-
-void release_posix_file_system(struct fsal_filesystem *fs);
-
-int re_index_fs_fsid(struct fsal_filesystem *fs,
-		     enum fsid_type fsid_type,
-		     struct fsal_fsid__ *fsid);
-
-int re_index_fs_dev(struct fsal_filesystem *fs,
-		    struct fsal_dev__ *dev);
-
-int change_fsid_type(struct fsal_filesystem *fs,
-		     enum fsid_type fsid_type);
-
-struct fsal_filesystem *lookup_fsid_locked(struct fsal_fsid__ *fsid,
-					   enum fsid_type fsid_type);
-struct fsal_filesystem *lookup_dev_locked(struct fsal_dev__ *dev);
-struct fsal_filesystem *lookup_fsid(struct fsal_fsid__ *fsid,
-				    enum fsid_type fsid_type);
-struct fsal_filesystem *lookup_dev(struct fsal_dev__ *dev);
-
-void unclaim_fs(struct fsal_filesystem *this);
-
-int claim_posix_filesystems(const char *path,
-			    struct fsal_module *fsal,
-			    struct fsal_export *exp,
-			    claim_filesystem_cb claim,
-			    unclaim_filesystem_cb unclaim,
-			    struct fsal_filesystem **root_fs);
-
 int encode_fsid(char *buf,
 		int max,
 		struct fsal_fsid__ *fsid,
@@ -162,7 +103,7 @@ int decode_fsid(char *buf,
 		struct fsal_fsid__ *fsid,
 		enum fsid_type fsid_type);
 
-fsal_errors_t fsal_inherit_acls(struct attrlist *attrs, fsal_acl_t *sacl,
+fsal_errors_t fsal_inherit_acls(struct fsal_attrlist *attrs, fsal_acl_t *sacl,
 			       fsal_aceflag_t inherit);
 fsal_status_t fsal_remove_access(struct fsal_obj_handle *dir_hdl,
 				 struct fsal_obj_handle *rem_hdl,
@@ -172,10 +113,12 @@ fsal_status_t fsal_rename_access(struct fsal_obj_handle *old_dir_hdl,
 				 struct fsal_obj_handle *new_dir_hdl,
 				 struct fsal_obj_handle *dst_obj_hdl,
 				 bool isdir);
-fsal_status_t fsal_mode_to_acl(struct attrlist *attrs, fsal_acl_t *sacl);
-fsal_status_t fsal_acl_to_mode(struct attrlist *attrs);
+fsal_status_t fsal_mode_to_acl(struct fsal_attrlist *attrs, fsal_acl_t *sacl);
+fsal_status_t fsal_acl_to_mode(struct fsal_attrlist *attrs);
 
-void set_common_verifier(struct attrlist *attrs, fsal_verifier_t verifier);
+void set_common_verifier(struct fsal_attrlist *attrs,
+			 fsal_verifier_t verifier,
+			 bool trunc_verif);
 
 void update_share_counters(struct fsal_share *share,
 			   fsal_openflags_t old_openflags,
@@ -185,7 +128,8 @@ fsal_status_t check_share_conflict(struct fsal_share *share,
 				   fsal_openflags_t openflags,
 				   bool bypass);
 
-fsal_status_t merge_share(struct fsal_share *orig_share,
+fsal_status_t merge_share(struct fsal_obj_handle *orig_hdl,
+			  struct fsal_share *orig_share,
 			  struct fsal_share *dupe_share);
 
 /**
@@ -266,12 +210,16 @@ static inline struct state_t *init_state(struct state_t *state,
 	return state;
 }
 
-bool check_verifier_stat(struct stat *st, fsal_verifier_t verifier);
+bool check_verifier_stat(struct stat *st,
+			 fsal_verifier_t verifier,
+			 bool trunc_verif);
 
-bool check_verifier_attrlist(struct attrlist *attrs, fsal_verifier_t verifier);
+bool check_verifier_attrlist(struct fsal_attrlist *attrs,
+			     fsal_verifier_t verifier,
+			     bool trunc_verif);
 
 bool fsal_common_is_referral(struct fsal_obj_handle *obj_hdl,
-			     struct attrlist *attrs, bool cache_attrs);
+			     struct fsal_attrlist *attrs, bool cache_attrs);
 
 fsal_status_t update_export(struct fsal_module *fsal_hdl,
 			    void *parse_node,
@@ -279,8 +227,4 @@ fsal_status_t update_export(struct fsal_module *fsal_hdl,
 			    struct fsal_export *original,
 			    struct fsal_module *updated_super);
 
-#ifdef USE_DBUS
-void dbus_cache_init(void);
-#endif
-
 #endif				/* FSAL_COMMONLIB_H */
