@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: LGPL-3.0-or-later */
 /*
  *
  *
@@ -93,58 +94,44 @@ extern uint32_t root_op_export_set;
  */
 extern int g_nodeid;
 
-/**
- * @brief Ops context for asynch and not protocol tasks that need to use
- * subsystems that depend on op_ctx.
- */
-
-struct root_op_context {
-	struct req_op_context req_ctx;
-	struct req_op_context *old_op_ctx;
-	struct user_cred creds;
-	struct export_perms export_perms;
-};
-
 extern size_t open_fd_count;
 
-static inline void init_root_op_context(struct root_op_context *ctx,
-					struct gsh_export *exp,
-					struct fsal_export *fsal_exp,
-					uint32_t nfs_vers,
-					uint32_t nfs_minorvers,
-					uint32_t req_type)
+void init_ctx_refstr(void);
+void destroy_ctx_refstr(void);
+
+void init_op_context(struct req_op_context *ctx,
+		     struct gsh_export *exp,
+		     struct fsal_export *fsal_exp,
+		     sockaddr_t *caller_data,
+		     uint32_t nfs_vers,
+		     uint32_t nfs_minorvers,
+		     enum request_type req_type);
+
+static inline void init_op_context_simple(struct req_op_context *ctx,
+					  struct gsh_export *exp,
+					  struct fsal_export *fsal_exp)
 {
-	/* Initialize req_ctx.
-	 * Note that a zeroed creds works just fine as root creds.
-	 */
-	memset(ctx, 0, sizeof(*ctx));
-	ctx->req_ctx.creds = &ctx->creds;
-	ctx->req_ctx.nfs_vers = nfs_vers;
-	ctx->req_ctx.nfs_minorvers = nfs_minorvers;
-	ctx->req_ctx.req_type = req_type;
-
-	ctx->req_ctx.ctx_export = exp;
-	ctx->req_ctx.fsal_export = fsal_exp;
-	if (fsal_exp)
-		ctx->req_ctx.fsal_module = fsal_exp->fsal;
-	else if (op_ctx)
-		ctx->req_ctx.fsal_module = op_ctx->fsal_module;
-
-	ctx->req_ctx.export_perms = &ctx->export_perms;
-	ctx->export_perms.set = root_op_export_set;
-	ctx->export_perms.options = root_op_export_options;
-
-	ctx->old_op_ctx = op_ctx;
-	op_ctx = &ctx->req_ctx;
+	init_op_context(ctx, exp, fsal_exp, NULL, 0, 0, UNKNOWN_REQUEST);
 }
 
-static inline void release_root_op_context(void)
-{
-	struct root_op_context *ctx;
+void release_op_context(void);
+void suspend_op_context(void);
+void resume_op_context(struct req_op_context *ctx);
+void set_op_context_export_fsal(struct gsh_export *exp,
+				struct fsal_export *fsal_exp);
+void clear_op_context_export(void);
+void save_op_context_export_and_clear(struct saved_export_context *saved);
+void save_op_context_export_and_set_export(struct saved_export_context *saved,
+					   struct gsh_export *exp);
+void restore_op_context_export(struct saved_export_context *saved);
+void discard_op_context_export(struct saved_export_context *saved);
+void set_op_context_pnfs_ds(struct fsal_pnfs_ds *pds);
 
-	ctx = container_of(op_ctx, struct root_op_context, req_ctx);
-	op_ctx = ctx->old_op_ctx;
-}
+#define set_op_context_export(exp_expr) \
+	do { \
+		struct gsh_export *ex = exp_expr; \
+		set_op_context_export_fsal(ex, ex ? ex->fsal_export : NULL); \
+	} while (0)
 
 /******************************************************
  *                Structure used to define a fsal
@@ -233,10 +220,10 @@ void start_fsals(void);
 void display_fsinfo(struct fsal_module *fsal);
 
 int display_attrlist(struct display_buffer *dspbuf,
-		     struct attrlist *attr, bool is_obj);
+		     struct fsal_attrlist *attr, bool is_obj);
 
 void log_attrlist(log_components_t component, log_levels_t level,
-		  const char *reason, struct attrlist *attr, bool is_obj,
+		  const char *reason, struct fsal_attrlist *attr, bool is_obj,
 		  char *file, int line, char *function);
 
 #define LogAttrlist(component, level, reason, attr, is_obj)                  \
@@ -263,7 +250,7 @@ enum cb_state {
 typedef fsal_errors_t (*helper_readdir_cb)
 	(void *opaque,
 	 struct fsal_obj_handle *obj,
-	 const struct attrlist *attr,
+	 const struct fsal_attrlist *attr,
 	 uint64_t mounted_on_fileid,
 	 uint64_t cookie,
 	 enum cb_state cb_state);
@@ -276,7 +263,7 @@ typedef fsal_errors_t (*helper_readdir_cb)
  * track of the structure it is filling, space used, and so forth.
  *
  * This function should return true if the entry has been added to the
- * caller's responde, or false if the structure is fulled and the
+ * caller's response, or false if the structure is fulled and the
  * structure has not been added.
  */
 
@@ -285,13 +272,13 @@ struct fsal_readdir_cb_parms {
 	const char *name;	/*< Dir entry name */
 	bool attr_allowed;	/*< True if caller has perm to getattr */
 	bool in_result;		/*< true if the entry has been added to the
-				 *< caller's responde, or false if the
+				 *< caller's response, or false if the
 				 *< structure is filled and the entry has not
 				 *< been added. */
 };
 
 fsal_status_t fsal_setattr(struct fsal_obj_handle *obj, bool bypass,
-			   struct state_t *state, struct attrlist *attr);
+			   struct state_t *state, struct fsal_attrlist *attr);
 
 /**
  *
@@ -322,18 +309,20 @@ fsal_status_t fsal_readlink(struct fsal_obj_handle *obj,
 fsal_status_t fsal_lookup(struct fsal_obj_handle *parent,
 			  const char *name,
 			  struct fsal_obj_handle **obj,
-			  struct attrlist *attrs_out);
+			  struct fsal_attrlist *attrs_out);
+fsal_status_t fsal_lookup_path(const char *path,
+			       struct fsal_obj_handle **dirobj);
 fsal_status_t fsal_lookupp(struct fsal_obj_handle *obj,
 			   struct fsal_obj_handle **parent,
-			   struct attrlist *attrs_out);
+			   struct fsal_attrlist *attrs_out);
 fsal_status_t fsal_create(struct fsal_obj_handle *parent,
 			  const char *name,
 			  object_file_type_t type,
-			  struct attrlist *attrs,
+			  struct fsal_attrlist *attrs,
 			  const char *link_content,
 			  struct fsal_obj_handle **obj,
-			  struct attrlist *attrs_out);
-void fsal_create_set_verifier(struct attrlist *sattr, uint32_t verf_hi,
+			  struct fsal_attrlist *attrs_out);
+void fsal_create_set_verifier(struct fsal_attrlist *sattr, uint32_t verf_hi,
 			      uint32_t verf_lo);
 bool fsal_create_verify(struct fsal_obj_handle *obj, uint32_t verf_hi,
 			uint32_t verf_lo);
@@ -352,16 +341,16 @@ fsal_status_t fsal_open2(struct fsal_obj_handle *in_obj,
 			 fsal_openflags_t openflags,
 			 enum fsal_create_mode createmode,
 			 const char *name,
-			 struct attrlist *attr,
+			 struct fsal_attrlist *attr,
 			 fsal_verifier_t verifier,
 			 struct fsal_obj_handle **obj,
-			 struct attrlist *attrs_out);
+			 struct fsal_attrlist *attrs_out);
 fsal_status_t fsal_reopen2(struct fsal_obj_handle *obj,
 			   struct state_t *state,
 			   fsal_openflags_t openflags,
 			   bool check_permission);
 fsal_status_t get_optional_attrs(struct fsal_obj_handle *obj_hdl,
-				 struct attrlist *attrs_out);
+				 struct fsal_attrlist *attrs_out);
 /**
  * @brief Close a file
  *
@@ -425,14 +414,14 @@ fsal_status_t fsal_verify2(struct fsal_obj_handle *obj,
 			   fsal_verifier_t verifier);
 
 /**
- * @brief Pepare an attrlist for fetching attributes.
+ * @brief Prepare an fsal_attrlist for fetching attributes.
  *
- * @param[in,out] attrs   The attrlist to work with
+ * @param[in,out] attrs   The fsal_attrlist to work with
  * @param[in]             The mask to use for the fetch
  *
  */
 
-static inline void fsal_prepare_attrs(struct attrlist *attrs,
+static inline void fsal_prepare_attrs(struct fsal_attrlist *attrs,
 				      attrmask_t request_mask)
 {
 	memset(attrs, 0, sizeof(*attrs));
@@ -440,13 +429,13 @@ static inline void fsal_prepare_attrs(struct attrlist *attrs,
 }
 
 /**
- * @brief Release any extra resources from an attrlist.
+ * @brief Release any extra resources from an fsal_attrlist.
  *
- * @param[in] attrs   The attrlist to work with
+ * @param[in] attrs   The fsal_attrlist to work with
  *
  */
 
-static inline void fsal_release_attrs(struct attrlist *attrs)
+static inline void fsal_release_attrs(struct fsal_attrlist *attrs)
 {
 	if (attrs->acl != NULL) {
 		int acl_status;
@@ -480,14 +469,15 @@ static inline void fsal_release_attrs(struct attrlist *attrs)
  * If ACL is requested in dest->request_mask, then ACL reference is acquired,
  * otherwise acl pointer is set to NULL.
  *
- * @param[in,out] dest       The attrlist to receive the copy (mask must be set)
- * @param[in]     src        The attrlist to make a copy of
+ * @param[in,out] dest       The fsal_attrlist to receive the copy (mask must be
+ *                           set)
+ * @param[in]     src        The fsal_attrlist to make a copy of
  * @param[in]     pass_refs  If true, pass the ACL reference to dest.
  *
  */
 
-static inline void fsal_copy_attrs(struct attrlist *dest,
-				   struct attrlist *src,
+static inline void fsal_copy_attrs(struct fsal_attrlist *dest,
+				   struct fsal_attrlist *src,
 				   bool pass_refs)
 {
 	attrmask_t save_request_mask = dest->request_mask;
@@ -561,7 +551,7 @@ static inline void fsal_copy_attrs(struct attrlist *dest,
 static inline changeid4
 fsal_get_changeid4(struct fsal_obj_handle *obj)
 {
-	struct attrlist attrs;
+	struct fsal_attrlist attrs;
 	fsal_status_t status;
 	changeid4 change;
 
@@ -597,7 +587,7 @@ enum fsal_create_mode nfs3_createmode_to_fsal(createmode3 createmode)
  * is not open in a mode usable by the caller.
  *
  * The caller may pass FSAL_O_ANY to indicate any mode of open (RDONLY,
- * WRONLY, or RDWR is useable - often just to fetch attributes or something).
+ * WRONLY, or RDWR is usable - often just to fetch attributes or something).
  *
  * @param[in] fd_openflags The openflags describing the fd
  * @param[in] to_openflags The openflags describing the desired mode
@@ -622,7 +612,7 @@ static inline bool not_open_usable(fsal_openflags_t fd_openflags,
  * is open in a mode usable by the caller.
  *
  * The caller may pass FSAL_O_ANY to indicate any mode of open (RDONLY,
- * WRONLY, or RDWR is useable - often just to fetch attributes or something).
+ * WRONLY, or RDWR is usable - often just to fetch attributes or something).
  *
  * Note that this function is not just an inversion of the above function
  * because O_SYNC is not considered.
@@ -679,5 +669,12 @@ extern void fsal_write(struct fsal_obj_handle *obj_hdl,
 		       struct fsal_io_arg *arg,
 		       struct async_process_data *data);
 
+fsal_status_t fsal_listxattr_helper(const char *buf,
+				    size_t listlen,
+				    uint32_t maxbytes,
+				    nfs_cookie4 *lxa_cookie,
+				    bool_t *lxr_eof,
+				    xattrlist4 *lxr_names);
+
 #endif				/* !FSAL_H */
 /** @} */
