@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: LGPL-3.0-or-later */
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -55,11 +56,24 @@
 
 typedef enum protos {
 	P_NFS,			/*< NFS, of course. */
+#ifdef _USE_NFS3
 	P_MNT,			/*< Mount (for v3) */
+#endif
+#ifdef _USE_NLM
 	P_NLM,			/*< NLM (for v3) */
+#endif
+#ifdef _USE_RQUOTA
 	P_RQUOTA,		/*< RQUOTA (for v3) */
+#endif
+#ifdef USE_NFSACL3
+	P_NFSACL,		/*< NFSACL (for v3) */
+#endif
+#ifdef RPC_VSOCK
 	P_NFS_VSOCK,		/*< NFS over vmware, qemu vmci sockets */
+#endif
+#ifdef _USE_NFS_RDMA
 	P_NFS_RDMA,		/*< NFS over RPC/RDMA */
+#endif
 	P_COUNT			/*< Number of protocols */
 } protos;
 
@@ -115,7 +129,7 @@ typedef enum protos {
 #define DRC_TCP_RECYCLE_EXPIRE_S 600	/* 10m */
 
 /**
- * @brief Default value for core_param.drc.tcp.checkstum
+ * @brief Default value for core_param.drc.tcp.checksum
  */
 #define DRC_TCP_CHECKSUM true
 
@@ -189,16 +203,16 @@ typedef enum protos {
 /**
  * @brief Support NFSv3 and NFSv4.
  */
+#ifdef _USE_NFS3
 #define CORE_OPTION_ALL_NFS_VERS (CORE_OPTION_NFSV3 | CORE_OPTION_NFSV4)
+#else
+#define CORE_OPTION_ALL_NFS_VERS CORE_OPTION_NFSV4
+#endif
 
-/**
- * @brief Support all protocols
- */
-#define CORE_OPTION_ALL_VERS (CORE_OPTION_NFSV3 |			\
-				CORE_OPTION_NFSV4 |			\
-				CORE_OPTION_NFS_VSOCK |			\
-				CORE_OPTION_NFS_RDMA |			\
-				CORE_OPTION_9P)
+#define UDP_LISTENER_NONE	0
+#define UDP_LISTENER_ALL	0x00000001
+#define UDP_LISTENER_MOUNT	0x00000002
+#define UDP_LISTENER_MASK (UDP_LISTENER_ALL | UDP_LISTENER_MOUNT)
 
 typedef struct nfs_core_param {
 	/** An array of port numbers, one for each protocol.  Set by
@@ -206,7 +220,7 @@ typedef struct nfs_core_param {
 	uint16_t port[P_COUNT];
 	/** The IPv4 or IPv6 address to which to bind for our
 	    listening port.  Set by the Bind_Addr option. */
-	struct sockaddr_storage bind_addr;
+	sockaddr_t bind_addr;
 	/** An array of RPC program numbers.  The correct values, by
 	    default, they may be set to incorrect values with the
 	    NFS_Program, MNT_Program, NLM_Program, and
@@ -317,6 +331,9 @@ typedef struct nfs_core_param {
 		uint32_t max_recv_buffer_size;
 		/** Idle timeout (seconds).  Defaults to 5m */
 		uint32_t idle_timeout_s;
+		/** TIRPC ioq min simultaneous io threads.  Defaults to
+		    2 and settable by rpc_ioq_thrdmin. */
+		uint32_t ioq_thrd_min;
 		/** TIRPC ioq max simultaneous io threads.  Defaults to
 		    200 and settable by RPC_Ioq_ThrdMax. */
 		uint32_t ioq_thrd_max;
@@ -339,28 +356,38 @@ typedef struct nfs_core_param {
 	    Defaults to CORE_OPTION_ALL_VERS and is settable with
 	    NFS_Protocols (as a comma-separated list of 3 and 4.) */
 	unsigned int core_options;
-	/** Whether to use the supplied name rather than the IP
-	    address in NSM operations.  Settable with
-	    NSM_Use_Caller_Name. */
-	bool nsm_use_caller_name;
 	/** Whether this Ganesha is part of a cluster of Ganeshas.
 	    This is somewhat vendor-specific and should probably be
 	    moved somewhere else.  Settable with Clustered. */
 	bool clustered;
+#ifdef _USE_NLM
 	/** Whether to support the Network Lock Manager protocol.
 	    Defaults to true and is settable with Enable_NLM. */
 	bool enable_NLM;
+	/** Whether to use the supplied name rather than the IP
+	    address in NSM operations.  Settable with
+	    NSM_Use_Caller_Name. */
+	bool nsm_use_caller_name;
+#endif
+#ifdef _USE_RQUOTA
 	/** Whether to support the Remote Quota protocol.  Defaults
 	    to true and is settable with Enable_RQUOTA. */
 	bool enable_RQUOTA;
+#endif
+#ifdef USE_NFSACL3
+	/* Whether to support the POSIX ACL. Defaults to false. */
+	bool enable_NFSACL;
+#endif
 	/** Whether to collect NFS stats.  Defaults to true. */
 	bool enable_NFSSTATS;
 	/** Whether to use fast stats.  Defaults to false. */
 	bool enable_FASTSTATS;
 	/** Whether to collect FSAL stats.  Defaults to false. */
 	bool enable_FSALSTATS;
+#ifdef _USE_NFS3
 	/** Whether to collect NFSv3 Detailed stats.  Defaults to false. */
 	bool enable_FULLV3STATS;
+#endif
 	/** Whether to collect NFSv4 Detailed stats.  Defaults to false. */
 	bool enable_FULLV4STATS;
 	/** Whether to collect Auth related stats. Defaults to false. */
@@ -395,7 +422,7 @@ typedef struct nfs_core_param {
 	    mounts. */
 	bool mount_path_pseudo;
 	/** Whether to disable UDP listeners */
-	bool enable_UDP;
+	uint32_t enable_UDP;
 	/** DBus name prefix. Required if one wants to run multiple ganesha
 	    instances on single host. The prefix should be different for every
 	    ganesha instance. If this is set, dbus name will be
@@ -408,6 +435,16 @@ typedef struct nfs_core_param {
 	    environment where users are part of large number of groups.
 	*/
 	uint32_t max_uid_to_grp_reqs;
+	/** Enable v3 filehandle to be used for v4 */
+	bool enable_v3_fh_for_v4;
+	/** Readdir response size, default is 0(use maxcount from nfs request)
+	*  range 4K-32K
+	*/
+	uint32_t readdir_res_size;
+	/** Whether to call getattrs in nfs4_complete_read and
+		nfs3_complete_read.
+		Defaults to true and settable by Getattrs_In_Complete_Read. */
+	bool getattrs_in_complete_read;
 } nfs_core_parameter_t;
 
 /** @} */
@@ -477,12 +514,21 @@ typedef struct nfs_version4_parameter {
 	/** The NFS grace period.  Defaults to
 	    GRACE_PERIOD_DEFAULT and is settable with Grace_Period. */
 	uint32_t grace_period;
+	/** The eir_server_scope for lock recovery. Defaults to NULL
+	    and is settable with server_scope. */
+	char *server_scope;
 	/** Domain to use if we aren't using the nfsidmap.  Defaults
 	    to DOMAINNAME_DEFAULT and is set with DomainName. */
 	char *domainname;
 	/** Path to the idmap configuration file.  Defaults to
 	    IDMAPCONF_DEFAULT, settable with IdMapConf */
 	char *idmapconf;
+	/** Full path to recovery root directory */
+	char *recov_root;
+	/** Name of recovery directory */
+	char *recov_dir;
+	/** Name of recovery old dir (for legacy recovery_fs only */
+	char *recov_old_dir;
 	/** Whether to use local password (PAM, on Linux) rather than
 	    nfsidmap.  Defaults to false if nfsidmap support is
 	    compiled in and true if it isn't.  Settable with
@@ -515,6 +561,8 @@ typedef struct nfs_version4_parameter {
 	/** whether to skip utf8 validation. defaults to false and settable
 	     with enforce_utf8_validation. */
 	bool enforce_utf8_vld;
+	/** Max number of Client IDs allowed on the system */
+	uint32_t max_client_ids;
 } nfs_version4_parameter_t;
 
 /** @} */
