@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: LGPL-3.0-or-later */
 /*
  *
  *
@@ -92,6 +93,13 @@ typedef struct state_t			state_t;
  */
 #define STATE_ERR_MAX 100
 
+/**
+ * @brief Maximum number of operation in a compound request
+ *
+ * We cap the number of operation to this value for all V4 versions
+ */
+#define NFS4_MAX_OPERATIONS 100
+
 /**
  * @brief Indicate that lock extends to the entire range of the file
  *
@@ -186,7 +194,7 @@ struct nfs41_session {
 	sockaddr_t connections[NFS41_MAX_CONNECTIONS];
 
 	nfs_client_id_t *clientid_record;	/*< Client record
-						   correspinding to ID */
+						   corresponding to ID */
 	clientid4 clientid;	/*< Clientid owning this session */
 	uint32_t cb_program;	/*< Callback program ID */
 	uint32_t flags;		/*< Flags pertaining to this session */
@@ -306,7 +314,7 @@ struct state_9p_fid {
 
 /* @brief Per client, per file stats */
 struct cf_deleg_stats {
-	time_t cfd_rs_time;                   /* time when the client responsed
+	time_t cfd_rs_time;                   /* time when the client responded
 						 NFS4_OK for a recall. */
 	time_t cfd_r_time;               /* time of the recall attempt */
 };
@@ -376,7 +384,7 @@ extern char all_ones[OTHERSIZE];
  * Each state is identified by stateid and represents some resource or
  * set of resources.
  *
- * The lists are protected by the state_lock
+ * The lists are protected by the st_lock
  */
 
 struct state_t {
@@ -484,6 +492,15 @@ extern hash_table_t *ht_nlm_client;
 
 /**
  * @brief NSM (rpc.statd) state for a given client.
+ *
+ * All nsm clients have a caller name.
+ *
+ * When nsm_use_caller_name is true, the caller name comes from the NLM
+ * requests.
+ *
+ * When nsm_use_caller_name is false, the caller name is the string form of
+ * the IP address. If the IP address is IPv4 encapsulated in IPv6, the name
+ * will be the simple IPv4 version of the address.
  */
 
 typedef struct state_nsm_client_t {
@@ -509,7 +526,7 @@ struct state_nlm_client_t {
 	xprt_type_t slc_client_type;	/*< The transport type to this
 					   client */
 	int32_t slc_refcount;	/*< Reference count for disposal */
-	struct sockaddr_storage slc_server_addr; /*< local addr when request is
+	sockaddr_t slc_server_addr; /*< local addr when request is
 						     made */
 	int32_t slc_nlm_caller_name_len;	/*< Length of client name */
 	char *slc_nlm_caller_name;	/*< Client name */
@@ -534,7 +551,7 @@ typedef struct state_nlm_owner_t {
 #ifdef _USE_9P
 typedef struct state_9p_owner_t {
 	u32 proc_id;		/*< PID on the client */
-	struct sockaddr_storage client_addr;	/*< Network address of client */
+	sockaddr_t client_addr;	/*< Network address of client */
 } state_9p_owner_t;
 #endif				/* _USE_9P */
 
@@ -748,7 +765,7 @@ struct nfs_client_record_t {
 				     disjoint. Linux client stupidity
 				     forces us to do so. */
 	int cr_client_val_len;  /*< Length of owner */
-	char cr_client_val[];   /*< Suplied co_owner */
+	char cr_client_val[];   /*< Supplied co_owner */
 };
 
 extern pool_t *client_id_pool;
@@ -897,21 +914,21 @@ typedef struct cbgetattr_rsp  cbgetattr_t;
 struct state_file {
 	/** File owning state */
 	struct fsal_obj_handle *obj;
-	/** NFSv4 states on this file. Protected by state_lock */
+	/** NFSv4 states on this file. Protected by st_lock */
 	struct glist_head list_of_states;
-	/** Layout recalls on this file. Protected by state_lock */
+	/** Layout recalls on this file. Protected by st_lock */
 	struct glist_head layoutrecall_list;
-	/** Pointers for lock list. Protected by state_lock */
+	/** Pointers for lock list. Protected by st_lock */
 	struct glist_head lock_list;
-	/** Pointers for NLM share list. Protected by state_lock */
+	/** Pointers for NLM share list. Protected by st_lock */
 	struct glist_head nlm_share_list;
-	/** true iff write delegated. Protected by state_lock */
+	/** true iff write delegated. Protected by st_lock */
 	bool write_delegated;
-	/** client holding write_deleg. Protected by state_lock */
+	/** client holding write_deleg. Protected by st_lock */
 	nfs_client_id_t *write_deleg_client;
-	/** Delegation statistics. Protected by state_lock */
+	/** Delegation statistics. Protected by st_lock */
 	struct file_deleg_stats fdeleg_stats;
-	/* cbgetattr stats. Protected by state_lock */
+	/* cbgetattr stats. Protected by st_lock */
 	cbgetattr_t cbgetattr;
 	uint32_t anon_ops;   /* number of anonymous operations
 			      * happening at the moment which
@@ -926,10 +943,12 @@ struct state_file {
  */
 struct state_dir {
 	/** If this is a junction, the export this node points to.
-	 * Protected by state_lock. */
+	 * Protected by jct_lock. */
 	struct gsh_export *junction_export;
-	/** List of exports that have this cache inode as their root.
-	 * Protected by state_lock */
+	/** gsh_refstr to stash copy of the export's pseudopath */
+	struct gsh_refstr *jct_pseudopath;
+	/** List of exports that have this object as their root.
+	 * Protected by jct_lock */
 	struct glist_head export_roots;
 	/** There is one export root reference counted for each export
 	    for which this entry is a root for. This field is used
@@ -937,16 +956,54 @@ struct state_dir {
 	int32_t exp_root_refcount;
 };
 
+/**
+ *  @brief Associate lock state or export junction information with an object.
+ *
+ * LOCK ORDERING
+ *
+ * The state handle has two overlapping locks for different purposes.
+ *
+ * The st_lock is used to protect byte range locks, opens, and such for regular
+ * files. It is a mutex since there is no parallelism benefit to it being a
+ * rwlock.
+ *
+ * The jct_lock is used to protect export junction information for directories.
+ * It is a rwlock since most of the time junctions are being looked at not
+ * modified.
+ *
+ * Both of these locks are often used in conjunction with the export->lock, but
+ * the rules of lock order are different.
+ *
+ * For st_lock, the export->lock MUST NOT be held, the export->lock is often
+ * taken while already holding st_lock, so the order is:
+ *
+ * st_lock THEN export->lock
+ *
+ * For jct_lock, there is one place where the export->lock is already held,
+ * so it makes sense for the order to be:
+ *
+ * export->lock THEN jct_lock.
+ *
+ */
 struct state_hdl {
-	/** Lock protecting state */
-	pthread_rwlock_t state_lock;
+	union {
+		/** Lock protecting state */
+		pthread_mutex_t st_lock;
+		/** Lock protecting export junctions */
+		pthread_rwlock_t jct_lock;
+	};
 	bool no_cleanup;
 	union {
+		/** Details about the state held for a regular file. */
 		struct state_file	file;
+		/** Details about an export junction. */
 		struct state_dir	dir;
 	};
 };
 
+#define JCT_PSEUDOPATH(shdl) (shdl->dir.jct_pseudopath ? \
+	shdl->dir.jct_pseudopath->gr_val : NULL)
+
 /**
  * @brief Description of a layout segment
  */
