@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: LGPL-3.0-or-later */
 /*
  *
  *
@@ -80,27 +81,27 @@ void free_state_owner(state_owner_t *owner);
 	} while (0)
 
 /**
- * @brief Acquire exclusive state_lock and set no_cleanup=true
+ * @brief Acquire exclusive st_lock and set no_cleanup=true
  *
- * @param[in,out] obj the object whose state_hdl->state_lock is to be
+ * @param[in,out] obj the object whose state_hdl->st_lock is to be
  *		      acquired and state_hdl->no_cleanup needs to be set
  */
 #define STATELOCK_lock(obj)						\
 	do {								\
-		PTHREAD_RWLOCK_wrlock(&(obj)->state_hdl->state_lock);	\
+		PTHREAD_MUTEX_lock(&(obj)->state_hdl->st_lock);	\
 		(obj)->state_hdl->no_cleanup = true;			\
 	} while (0)
 
 /**
- * @brief Drop state_lock and set no_cleanup=false
+ * @brief Drop st_lock and set no_cleanup=false
  *
- * @param[in,out] obj the object whose state_hdl->state_lock is to be
+ * @param[in,out] obj the object whose state_hdl->st_lock is to be
  *		      dropped and state_hdl->no_cleanup needs to be cleared
  */
 #define STATELOCK_unlock(obj)						\
 	do {								\
 		(obj)->state_hdl->no_cleanup = false;			\
-		PTHREAD_RWLOCK_unlock(&(obj)->state_hdl->state_lock);	\
+		PTHREAD_MUTEX_unlock(&(obj)->state_hdl->st_lock);	\
 	} while (0)
 
 state_owner_t *get_state_owner(care_t care, state_owner_t *pkey,
@@ -128,9 +129,9 @@ static inline void state_hdl_init(struct state_hdl *ostate,
 				  struct fsal_obj_handle *obj)
 {
 	memset(ostate, 0, sizeof(*ostate));
-	PTHREAD_RWLOCK_init(&ostate->state_lock, NULL);
 	switch (type) {
 	case REGULAR_FILE:
+		PTHREAD_MUTEX_init(&ostate->st_lock, NULL);
 		glist_init(&ostate->file.list_of_states);
 		glist_init(&ostate->file.layoutrecall_list);
 		glist_init(&ostate->file.lock_list);
@@ -138,6 +139,7 @@ static inline void state_hdl_init(struct state_hdl *ostate,
 		ostate->file.obj = obj;
 		break;
 	case DIRECTORY:
+		PTHREAD_RWLOCK_init(&ostate->jct_lock, NULL);
 		glist_init(&ostate->dir.export_roots);
 		break;
 	default:
@@ -150,9 +152,19 @@ static inline void state_hdl_init(struct state_hdl *ostate,
  *
  * @param[in] state_hdl	State handle to clean up
  */
-static inline void state_hdl_cleanup(struct state_hdl *state_hdl)
+static inline void state_hdl_cleanup(struct state_hdl *state_hdl,
+				     object_file_type_t type)
 {
-	PTHREAD_RWLOCK_destroy(&state_hdl->state_lock);
+	switch (type) {
+	case REGULAR_FILE:
+		PTHREAD_MUTEX_destroy(&state_hdl->st_lock);
+		break;
+	case DIRECTORY:
+		PTHREAD_RWLOCK_destroy(&state_hdl->jct_lock);
+		break;
+	default:
+		break;
+	}
 }
 
 /*****************************************************************************
@@ -166,8 +178,7 @@ int compare_9p_owner(state_owner_t *owner1, state_owner_t *owner2);
 
 int display_9p_owner(struct display_buffer *dspbuf, state_owner_t *owner);
 
-state_owner_t *get_9p_owner(struct sockaddr_storage *client_addr,
-			    uint32_t proc_id);
+state_owner_t *get_9p_owner(sockaddr_t *client_addr, uint32_t proc_id);
 
 int Init_9p_hash(void);
 #endif
@@ -181,13 +192,23 @@ int Init_9p_hash(void);
 void free_nsm_client(state_nsm_client_t *client);
 
 /* These refcount functions must not be called holding the ssc_mutex */
-void inc_nsm_client_ref(state_nsm_client_t *client);
-void dec_nsm_client_ref(state_nsm_client_t *client);
+void _inc_nsm_client_ref(state_nsm_client_t *client,
+			 char *file, int line, char *function);
+
+#define inc_nsm_client_ref(client) \
+	_inc_nsm_client_ref(client, \
+	(char *) __FILE__, __LINE__, (char *) __func__)
+
+void _dec_nsm_client_ref(state_nsm_client_t *client,
+			 char *file, int line, char *function);
+
+#define dec_nsm_client_ref(client) \
+	_dec_nsm_client_ref(client, \
+	(char *) __FILE__, __LINE__, (char *) __func__)
 
 int display_nsm_client(struct display_buffer *dspbuf, state_nsm_client_t *key);
 
-state_nsm_client_t *get_nsm_client(care_t care, SVCXPRT *xprt,
-				   char *caller_name);
+state_nsm_client_t *get_nsm_client(care_t care, char *caller_name);
 void inc_nlm_client_ref(state_nlm_client_t *client);
 void dec_nlm_client_ref(state_nlm_client_t *client);
 
@@ -671,23 +692,6 @@ static inline struct fsal_obj_handle *get_state_obj_ref(state_t *state)
 	return obj;
 }
 
-static inline struct gsh_export *get_state_export_ref(state_t *state)
-{
-	struct gsh_export *export = NULL;
-
-	PTHREAD_MUTEX_lock(&state->state_mutex);
-
-	if (state->state_export != NULL &&
-	    export_ready(state->state_export)) {
-		get_gsh_export_ref(state->state_export);
-		export = state->state_export;
-	}
-
-	PTHREAD_MUTEX_unlock(&state->state_mutex);
-
-	return export;
-}
-
 static inline bool state_same_export(state_t *state, struct gsh_export *export)
 {
 	bool same = false;
@@ -940,12 +944,12 @@ static inline bool obj_is_junction(struct fsal_obj_handle *obj)
 	if (obj->type != DIRECTORY)
 		return false;
 
-	PTHREAD_RWLOCK_rdlock(&obj->state_hdl->state_lock);
+	PTHREAD_RWLOCK_rdlock(&obj->state_hdl->jct_lock);
 
 	if ((obj->state_hdl->dir.junction_export != NULL ||
 	     atomic_fetch_int32_t(&obj->state_hdl->dir.exp_root_refcount) != 0))
 		res = true;
-	PTHREAD_RWLOCK_unlock(&obj->state_hdl->state_lock);
+	PTHREAD_RWLOCK_unlock(&obj->state_hdl->jct_lock);
 
 	return res;
 }
