@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -117,7 +118,7 @@ enum log_flag_index_t {
 	LF_PROGNAME,		/*< Ganesha program name field. */
 	LF_PID,			/*< Ganesha process identifier. */
 	LF_THREAD_NAME,		/*< Name of active thread logging message. */
-	LF_FILE_NAME,		/*< Source file name message occured in. */
+	LF_FILE_NAME,		/*< Source file name message occurred in. */
 	LF_LINE_NUM,		/*< Source line number message occurred in. */
 	LF_FUNCTION_NAME,	/*< Function name message occurred in. */
 	LF_COMPONENT,		/*< Log component. */
@@ -233,6 +234,7 @@ log_header_t max_headers = LH_COMPONENT;
 
 char const_log_str[LOG_BUFF_LEN] = "\0";
 char date_time_fmt[MAX_TD_FMT_LEN] = "\0";
+static bool disp_utc_timestamp;
 
 typedef struct loglev {
 	char *str;
@@ -276,17 +278,13 @@ __thread char *clientip = NULL;
 
 /* threads keys */
 #define LogChanges(format, args...) \
-	do { \
-		if (component_log_level[COMPONENT_LOG] == \
-		    NIV_FULL_DEBUG) \
-			DisplayLogComponentLevel(COMPONENT_LOG, \
-						 __FILE__, \
-						 __LINE__, \
-						 __func__, \
-						 NIV_NULL, \
-						 "LOG: " format, \
-						 ## args); \
-	} while (0)
+	DisplayLogComponentLevel(COMPONENT_LOG, \
+				 __FILE__, \
+				 __LINE__, \
+				 __func__, \
+				 NIV_NULL, \
+				 "LOG: " format, \
+				 ## args)
 
 struct cleanup_list_element *cleanup_list;
 
@@ -309,7 +307,7 @@ void Cleanup(void)
 void Fatal(void)
 {
 	Cleanup();
-	exit(2);
+	_exit(2);
 }
 
 #ifdef _DONT_HAVE_LOCALTIME_R
@@ -418,33 +416,7 @@ void SetClientIP(char *ip_str)
 	clientip = ip_str;
 }
 
-/* Installs a signal handler */
-static void ArmSignal(int signal, void (*action) ())
-{
-	struct sigaction act;
-
-	/* Placing fields of struct sigaction */
-	act.sa_flags = 0;
-	act.sa_handler = action;
-	sigemptyset(&act.sa_mask);
-
-	if (sigaction(signal, &act, NULL) == -1)
-		LogCrit(COMPONENT_LOG, "Failed to arm signal %d, error %d (%s)",
-			signal, errno, strerror(errno));
-}				/* ArmSignal */
-
-/*
- * Five functions to manage debug level throug signal
- * handlers.
- *
- * InitLogging
- * IncrementLevelDebug
- * DecrementLevelDebug
- * SetLevelDebug
- * ReturnLevelDebug
- */
-
-static void _SetLevelDebug(int level_to_set)
+static void SetLevelDebug(int level_to_set)
 {
 	int i;
 
@@ -461,16 +433,14 @@ static void _SetLevelDebug(int level_to_set)
 	}
 }				/* _SetLevelDebug */
 
-static void SetLevelDebug(int level_to_set)
-{
-	_SetLevelDebug(level_to_set);
-
-	LogChanges("Setting log level for all components to %s",
-		   ReturnLevelInt(component_log_level[COMPONENT_ALL]));
-}
+uint32_t rpc_debug_flags = TIRPC_DEBUG_FLAG_ERROR |
+			   TIRPC_DEBUG_FLAG_WARN |
+			   TIRPC_DEBUG_FLAG_EVENT;
 
 static void SetNTIRPCLogLevel(int level_to_set)
 {
+	uint32_t old = ntirpc_pp.debug_flags;
+
 	switch (level_to_set) {
 	case NIV_NULL:
 	case NIV_FATAL:
@@ -485,6 +455,7 @@ static void SetNTIRPCLogLevel(int level_to_set)
 					TIRPC_DEBUG_FLAG_WARN;
 		break;
 	case NIV_EVENT:
+	case NIV_INFO:
 		ntirpc_pp.debug_flags = TIRPC_DEBUG_FLAG_ERROR |
 					TIRPC_DEBUG_FLAG_WARN |
 					TIRPC_DEBUG_FLAG_EVENT;
@@ -492,6 +463,7 @@ static void SetNTIRPCLogLevel(int level_to_set)
 	case NIV_DEBUG:
 	case NIV_MID_DEBUG:
 		/* set by log_conf_commit() */
+		ntirpc_pp.debug_flags = rpc_debug_flags;
 		break;
 	case NIV_FULL_DEBUG:
 		ntirpc_pp.debug_flags = 0xFFFFFFFF; /* enable all flags */
@@ -503,6 +475,9 @@ static void SetNTIRPCLogLevel(int level_to_set)
 
 	if (!tirpc_control(TIRPC_SET_DEBUG_FLAGS, &ntirpc_pp.debug_flags))
 		LogCrit(COMPONENT_CONFIG, "Setting nTI-RPC debug_flags failed");
+	else if (old != ntirpc_pp.debug_flags)
+		LogChanges("Changed RPC_Debug_Flags from %"PRIx32" to %"PRIx32,
+			   old, ntirpc_pp.debug_flags);
 }
 
 void SetComponentLogLevel(log_components_t component, int level_to_set)
@@ -512,15 +487,6 @@ void SetComponentLogLevel(log_components_t component, int level_to_set)
 	assert(level_to_set < NB_LOG_LEVEL);
 	assert(component != COMPONENT_ALL);
 
-	if (LogComponents[component].comp_env_set) {
-		LogWarn(COMPONENT_CONFIG,
-			"LOG %s level %s from config is ignored because %s was set in environment",
-			LogComponents[component].comp_name,
-			ReturnLevelInt(level_to_set),
-			ReturnLevelInt(component_log_level[component]));
-		return;
-	}
-
 	if (component_log_level[component] == level_to_set)
 		return;
 
@@ -534,27 +500,6 @@ void SetComponentLogLevel(log_components_t component, int level_to_set)
 		SetNTIRPCLogLevel(level_to_set);
 }
 
-static inline int ReturnLevelDebug(void)
-{
-	return component_log_level[COMPONENT_ALL];
-}				/* ReturnLevelDebug */
-
-static void IncrementLevelDebug(void)
-{
-	_SetLevelDebug(ReturnLevelDebug() + 1);
-
-	LogChanges("SIGUSR1 Increasing log level for all components to %s",
-		   ReturnLevelInt(component_log_level[COMPONENT_ALL]));
-}				/* IncrementLevelDebug */
-
-static void DecrementLevelDebug(void)
-{
-	_SetLevelDebug(ReturnLevelDebug() - 1);
-
-	LogChanges("SIGUSR2 Decreasing log level for all components to %s",
-		   ReturnLevelInt(component_log_level[COMPONENT_ALL]));
-}				/* DecrementLevelDebug */
-
 void set_const_log_str(void)
 {
 	struct display_buffer dspbuf = { sizeof(const_log_str),
@@ -660,39 +605,6 @@ void set_const_log_str(void)
 		date_time_fmt[strlen(date_time_fmt) - 1] = '\0';
 }
 
-static void set_logging_from_env(void)
-{
-	char *env_value;
-	int newlevel, component, oldlevel;
-
-	for (component = COMPONENT_ALL; component < COMPONENT_COUNT;
-	     component++) {
-		env_value = getenv(LogComponents[component].comp_name);
-		if (env_value == NULL)
-			continue;
-		newlevel = ReturnLevelAscii(env_value);
-		if (newlevel == -1) {
-			LogCrit(COMPONENT_LOG,
-				"Environment variable %s exists, but the value %s is not a valid log level.",
-				LogComponents[component].comp_name,
-				env_value);
-			continue;
-		}
-		oldlevel = component_log_level[component];
-		if (component == COMPONENT_ALL)
-			_SetLevelDebug(newlevel);
-		else
-			SetComponentLogLevel(component, newlevel);
-		LogComponents[component].comp_env_set = true;
-		LogChanges(
-		     "Using environment variable to switch log level for %s from %s to %s",
-		     LogComponents[component].comp_name,
-		     ReturnLevelInt(oldlevel),
-		     ReturnLevelInt(newlevel));
-	}
-
-}				/* InitLogging */
-
 /**
  *
  * @brief Finds a log facility by name
@@ -821,7 +733,7 @@ void release_log_facility(const char *name)
 	if (facility == NULL) {
 		PTHREAD_RWLOCK_unlock(&log_rwlock);
 		LogCrit(COMPONENT_LOG,
-			 "Attempting release of non-existant log facility (%s)",
+			 "Attempting release of non-existent log facility (%s)",
 			 name);
 		return;
 	}
@@ -846,7 +758,7 @@ void release_log_facility(const char *name)
 /**
  * @brief Enable the named logger
  *
- * Enabling a logger adds it to the list of facilites that will be
+ * Enabling a logger adds it to the list of facilities that will be
  * used to report messages.
  *
  * @param name [IN] the name of the logger to enable
@@ -1181,13 +1093,13 @@ void init_logging(const char *log_path, const int debug_level)
 				 "Enable error (%s) for SYSLOG logging!",
 				 strerror(-rc));
 	}
-	if (debug_level >= 0)
+
+	if (debug_level >= 0) {
+		LogChanges("Setting log level for all components to %s",
+			   ReturnLevelInt(debug_level));
 		SetLevelDebug(debug_level);
-
-	set_logging_from_env();
-
-	ArmSignal(SIGUSR1, IncrementLevelDebug);
-	ArmSignal(SIGUSR2, DecrementLevelDebug);
+		original_log_level = debug_level;
+	}
 }
 
 /*
@@ -1319,7 +1231,10 @@ int display_timeval(struct display_buffer *dspbuf, struct timeval *tv)
 	if (logfields->datefmt == TD_NONE && logfields->timefmt == TD_NONE)
 		fmt = "%c ";
 
-	Localtime_r(&tm, &the_date);
+	if (disp_utc_timestamp)
+		gmtime_r(&tm, &the_date);
+	else
+		Localtime_r(&tm, &the_date);
 
 	/* Earlier we build the date/time format string in
 	 * date_time_fmt, now use that to format the time and/or date.
@@ -1485,7 +1400,7 @@ void display_log_component_level(log_components_t component, const char *file,
 	struct display_buffer dsp_log = {sizeof(log_buffer),
 					 log_buffer, log_buffer};
 
-	/* Build up the messsage and capture the various positions in it. */
+	/* Build up the message and capture the various positions in it. */
 	b_left = display_log_header(&dsp_log);
 
 	if (b_left > 0)
@@ -1577,8 +1492,18 @@ static log_levels_t default_log_levels[] = {
 	[COMPONENT_NFS_MSK] = NIV_EVENT
 };
 
+/* Active set of log levels */
 log_levels_t *component_log_level = default_log_levels;
 
+/* Original log level set by -N or otherwise code default */
+log_levels_t original_log_level = NIV_EVENT;
+
+/* Default log level setby LOG { default_log_level }, setting to NB_LOG_LEVEL
+ * indicates it has not been specified in the config (in which case we fall
+ * back to original_log_level.
+ */
+log_levels_t default_log_level = NB_LOG_LEVEL;
+
 struct log_component_info LogComponents[COMPONENT_COUNT] = {
 	[COMPONENT_ALL] = {
 		.comp_name = "COMPONENT_ALL",
@@ -1768,9 +1693,9 @@ static bool dbus_prop_set(log_components_t component, DBusMessageIter *arg)
 	}
 
 	if (component == COMPONENT_ALL) {
-		_SetLevelDebug(log_level);
-		LogChanges("Dbus set log level for all components to %s",
+		LogChanges("Dbus setting log level for all components to %s",
 			   level_code);
+		SetLevelDebug(log_level);
 	} else {
 		LogChanges("Dbus set log level for %s from %s to %s.",
 			   LogComponents[component].comp_name,
@@ -1929,8 +1854,9 @@ struct logger_config {
 	struct glist_head facility_list;
 	struct logfields *logfields;
 	log_levels_t *comp_log_level;
-	log_levels_t default_level;
+	log_levels_t default_log_level;
 	uint32_t rpc_debug_flags;
+	bool disp_utc_timestamp;
 };
 
 /**
@@ -2235,22 +2161,11 @@ static int component_commit(void *node, void *link_mem, void *self_struct,
 	struct logger_config *logger;
 	log_levels_t *log_level = self_struct;
 
-	if (log_level[COMPONENT_ALL] != NB_LOG_LEVEL) {
-		SetLevelDebug(log_level[COMPONENT_ALL]);
-	} else {
-		int comp;
+	/* Save the log levels in logger for later use if all is well */
+	logger = container_of(log_lvls, struct logger_config, comp_log_level);
+
+	logger->comp_log_level = log_level;
 
-		logger = container_of(log_lvls,
-				      struct logger_config,
-				      comp_log_level);
-		if (logger->default_level == NB_LOG_LEVEL)
-			logger->default_level = NIV_EVENT;
-		for (comp = COMPONENT_LOG; comp < COMPONENT_COUNT; comp++)
-			if (log_level[comp] == NB_LOG_LEVEL)
-				log_level[comp] = logger->default_level;
-		log_level[COMPONENT_ALL] = NIV_NULL;
-		logger->comp_log_level = log_level;
-	}
 	return 0;
 }
 
@@ -2426,6 +2341,53 @@ static void *log_conf_init(void *link_mem, void *self_struct)
 	return NULL;
 }
 
+static void apply_logger_config_levels(struct logger_config *logger)
+{
+	enum log_components comp;
+	bool has_levels = logger->comp_log_level != NULL;
+	log_levels_t log_level_all = has_levels
+				? logger->comp_log_level[COMPONENT_ALL]
+				: NB_LOG_LEVEL;
+
+	/* Handle Default_Log_Level */
+	if (logger->default_log_level != default_log_level) {
+		/* Default log level has changed */
+
+		LogChanges("Changing Default_Log_Level from %s to %s",
+			   ReturnLevelInt(default_log_level),
+			   ReturnLevelInt(logger->default_log_level));
+
+		default_log_level = logger->default_log_level;
+	}
+
+	for (comp = COMPONENT_LOG; comp < COMPONENT_COUNT; comp++) {
+		log_levels_t level;
+
+		if (log_level_all != NB_LOG_LEVEL) {
+			/* COMPONENT { ALL } was set, so use that to override
+			 * all log levels.
+			 */
+			level = log_level_all;
+		} else if (has_levels &&
+			   logger->comp_log_level[comp] != NB_LOG_LEVEL) {
+			/* Individual component level was set, use it */
+			level = logger->comp_log_level[comp];
+		} else if (default_log_level != NB_LOG_LEVEL) {
+			/* No COMPONENT was set and Default_Log_Level is set use
+			 * it.
+			 */
+			level = default_log_level;
+		} else {
+			/* Nothing has been set, revert to original log level
+			 * from -N command line option or code default.
+			 */
+			level = original_log_level;
+		}
+
+		SetComponentLogLevel(comp, level);
+	}
+}
+
 static int log_conf_commit(void *node, void *link_mem, void *self_struct,
 			   struct config_error_type *err_type)
 {
@@ -2524,6 +2486,7 @@ static int log_conf_commit(void *node, void *link_mem, void *self_struct,
  done:
 		(void)facility_init(&logger->facility_list, conf);
 	}
+
 	if (errcnt == 0) {
 		if (logger->logfields != NULL) {
 			LogEvent(COMPONENT_CONFIG,
@@ -2540,21 +2503,18 @@ static int log_conf_commit(void *node, void *link_mem, void *self_struct,
 			/* rebuild const_log_str with new format params. */
 			set_const_log_str();
 		}
-		if (logger->comp_log_level != NULL) {
-			LogEvent(COMPONENT_CONFIG,
-				 "Switching to new component log levels");
-			if (component_log_level == default_log_levels) {
-				/* First time change from default log level */
-				component_log_level = logger->comp_log_level;
-				logger->comp_log_level = NULL;
-			} else {
-				/* update exisiting component_log_level */
-				memcpy(component_log_level,
-				  logger->comp_log_level,
-				  (sizeof(log_levels_t) * COMPONENT_COUNT));
-			}
-		}
-		ntirpc_pp.debug_flags = logger->rpc_debug_flags;
+
+		/* Apply any changes to Default_Log_Level or COMPONENTS */
+		apply_logger_config_levels(logger);
+
+		if (ntirpc_pp.debug_flags != logger->rpc_debug_flags)
+			LogChanges("Changing custom RPC_Debug_Flags from %"
+				   PRIx32" to %"PRIx32,
+				   rpc_debug_flags,
+				   logger->rpc_debug_flags);
+
+		disp_utc_timestamp = logger->disp_utc_timestamp;
+		rpc_debug_flags = logger->rpc_debug_flags;
 		SetNTIRPCLogLevel(component_log_level[COMPONENT_TIRPC]);
 	} else {
 		if (logger->logfields != NULL) {
@@ -2567,16 +2527,19 @@ static int log_conf_commit(void *node, void *link_mem, void *self_struct,
 			gsh_free(lf);
 		}
 	}
+
 	if (logger->comp_log_level != NULL)
 		gsh_free(logger->comp_log_level);
+
 	logger->logfields = NULL;
 	logger->comp_log_level = NULL;
+
 	return errcnt;
 }
 
 static struct config_item logging_params[] = {
-	CONF_ITEM_TOKEN("Default_log_level", NB_LOG_LEVEL, log_levels,
-			 logger_config, default_level),
+	CONF_ITEM_TOKEN("Default_Log_Level", NB_LOG_LEVEL, log_levels,
+			 logger_config, default_log_level),
 	CONF_ITEM_UI32("RPC_Debug_Flags", 0, UINT32_MAX,
 		       TIRPC_DEBUG_FLAG_DEFAULT,
 		       logger_config, rpc_debug_flags),
@@ -2589,6 +2552,8 @@ static struct config_item logging_params[] = {
 	CONF_ITEM_BLOCK("Components", component_levels,
 			component_init, component_commit,
 			logger_config, comp_log_level),
+	CONF_ITEM_BOOL("Display_UTC_Timestamp", false,
+			logger_config, disp_utc_timestamp),
 	CONFIG_EOL
 };
 
@@ -2678,7 +2643,7 @@ bool _ratelimit(struct ratelimit_state *rs, int *missed)
 	bool ret;
 	time_t now;
 
-	/* If we fail to acquire the mutex, then we are alreday busy,
+	/* If we fail to acquire the mutex, then we are already busy,
 	 * so don't log message (aka return false)
 	 */
 	if (pthread_mutex_trylock(&rs->mutex))
