@@ -1,3 +1,28 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
+/*
+ * Copyright IBM Corporation, 2012
+ *  Contributor: Frank Filz <ffilz@us.ibm.com>
+ *
+ * --------------------------
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public License
+ * as published by the Free Software Foundation; either version 3 of
+ * the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+ * 02110-1301 USA
+ *
+ *
+ */
+
 #include "config.h"
 #include "abstract_mem.h"
 #include "fsal_types.h"
@@ -57,9 +82,13 @@ fsal_ace_t *nfs4_ace_alloc(int nace)
 	return ace;
 }
 
-fsal_acl_t *nfs4_acl_alloc()
+fsal_acl_t *nfs4_acl_alloc(void)
 {
-	return pool_alloc(fsal_acl_pool);
+	fsal_acl_t *acl = pool_alloc(fsal_acl_pool);
+
+	PTHREAD_RWLOCK_init(&acl->lock, NULL);
+
+	return acl;
 }
 
 void nfs4_ace_free(fsal_ace_t *ace)
@@ -80,6 +109,8 @@ void nfs4_acl_free(fsal_acl_t *acl)
 	if (acl->aces)
 		nfs4_ace_free(acl->aces);
 
+	PTHREAD_RWLOCK_destroy(&acl->lock);
+
 	pool_free(fsal_acl_pool, acl);
 }
 
@@ -138,19 +169,6 @@ fsal_acl_t *nfs4_acl_new_entry(fsal_acl_data_t *acldata,
 
 	/* Adding the entry in the cache */
 	acl = nfs4_acl_alloc();
-	if (pthread_rwlock_init(&(acl->lock), NULL) != 0) {
-		nfs4_acl_free(acl);
-		LogCrit(COMPONENT_NFS_V4_ACL,
-			"New ACL RW lock init returned %d (%s)", errno,
-			strerror(errno));
-		*status = NFS_V4_ACL_INIT_ENTRY_FAILED;
-
-		nfs4_ace_free(acldata->aces);
-		hashtable_releaselatched(fsal_acl_hash, &latch);
-
-		return NULL;
-	}
-
 	acl->naces = acldata->naces;
 	acl->aces = acldata->aces;
 	acl->ref = 1;		/* We give out one reference */
