@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * vim:noexpandtab:shiftwidth=8:tabstop=8:
  *
@@ -57,42 +58,91 @@
  * @brief Core configuration parameters
  */
 
+static struct config_item_list udp_listener_type[] = {
+	CONFIG_LIST_TOK("false", UDP_LISTENER_NONE),
+	CONFIG_LIST_TOK("no", UDP_LISTENER_NONE),
+	CONFIG_LIST_TOK("off", UDP_LISTENER_NONE),
+	CONFIG_LIST_TOK("true", UDP_LISTENER_ALL),
+	CONFIG_LIST_TOK("yes", UDP_LISTENER_ALL),
+	CONFIG_LIST_TOK("on", UDP_LISTENER_ALL),
+	CONFIG_LIST_TOK("mount", UDP_LISTENER_MOUNT),
+	CONFIG_LIST_EOL
+};
+
 static struct config_item_list protocols[] = {
 	CONFIG_LIST_TOK("none", CORE_OPTION_NONE),
+#ifdef _USE_NFS3
 	CONFIG_LIST_TOK("3", CORE_OPTION_NFSV3),
 	CONFIG_LIST_TOK("v3", CORE_OPTION_NFSV3),
 	CONFIG_LIST_TOK("nfs3", CORE_OPTION_NFSV3),
 	CONFIG_LIST_TOK("nfsv3", CORE_OPTION_NFSV3),
+#endif
 	CONFIG_LIST_TOK("4", CORE_OPTION_NFSV4),
 	CONFIG_LIST_TOK("v4", CORE_OPTION_NFSV4),
 	CONFIG_LIST_TOK("nfs4", CORE_OPTION_NFSV4),
 	CONFIG_LIST_TOK("nfsv4", CORE_OPTION_NFSV4),
+#ifdef RPC_VSOCK
 	CONFIG_LIST_TOK("nfsvsock", CORE_OPTION_NFS_VSOCK),
+#endif
+#ifdef _USE_NFS_RDMA
 	CONFIG_LIST_TOK("nfsrdma", CORE_OPTION_NFS_RDMA),
 	CONFIG_LIST_TOK("rpcrdma", CORE_OPTION_NFS_RDMA),
+#endif
+#ifdef _USE_9P
 	CONFIG_LIST_TOK("9p", CORE_OPTION_9P),
+#endif
 	CONFIG_LIST_EOL
 };
 
+/**
+ * @brief Support all protocols
+ */
+#ifdef _USE_NFS3
+#define DEFAULT_INCLUDES_NFSV3		CORE_OPTION_NFSV3
+#else
+#define DEFAULT_INCLUDES_NFSV3		CORE_OPTION_NONE
+#endif
+
+#define DEFAULT_INCLUDES_NFSV4		CORE_OPTION_NFSV4
+
+#define DEFAULT_PROTOCOLS  (DEFAULT_INCLUDES_NFSV3 | \
+			    DEFAULT_INCLUDES_NFSV4)
+
 static struct config_item core_params[] = {
 	CONF_ITEM_UI16("NFS_Port", 0, UINT16_MAX, NFS_PORT,
 		       nfs_core_param, port[P_NFS]),
+#ifdef _USE_NFS3
 	CONF_ITEM_UI16("MNT_Port", 0, UINT16_MAX, 0,
 		       nfs_core_param, port[P_MNT]),
+#endif
+#ifdef _USE_NLM
 	CONF_ITEM_UI16("NLM_Port", 0, UINT16_MAX, 0,
 		       nfs_core_param, port[P_NLM]),
+#endif
+#ifdef _USE_RQUOTA
 	CONF_ITEM_UI16("Rquota_Port", 0, UINT16_MAX, RQUOTA_PORT,
 		       nfs_core_param, port[P_RQUOTA]),
+#endif
 	CONF_ITEM_IP_ADDR("Bind_Addr", "0.0.0.0",
 			  nfs_core_param, bind_addr),
 	CONF_ITEM_UI32("NFS_Program", 1, INT32_MAX, NFS_PROGRAM,
 		       nfs_core_param, program[P_NFS]),
+#ifdef _USE_NFS3
 	CONF_ITEM_UI32("MNT_Program", 1, INT32_MAX, MOUNTPROG,
 				nfs_core_param, program[P_MNT]),
+#endif
+#ifdef _USE_NLM
 	CONF_ITEM_UI32("NLM_Program", 1, INT32_MAX, NLMPROG,
 		       nfs_core_param, program[P_NLM]),
+#endif
+#ifdef _USE_RQUOTA
 	CONF_ITEM_UI32("Rquota_Program", 1, INT32_MAX, RQUOTAPROG,
 		       nfs_core_param, program[P_RQUOTA]),
+#endif
+#ifdef USE_NFSACL3
+	CONF_ITEM_UI32("NFSACL_Program", 1, INT32_MAX, NFSACLPROG,
+		       nfs_core_param, program[P_NFSACL]),
+#endif
 	CONF_ITEM_DEPRECATED("Nb_Worker",
 			     "This parameter has been replaced with _9P { Nb_Worker}"
 			     ),
@@ -138,7 +188,9 @@ static struct config_item core_params[] = {
 	CONF_ITEM_UI32("MaxRPCRecvBufferSize", 1, 1048576*9,
 		       NFS_DEFAULT_RECV_BUFFER_SIZE,
 		       nfs_core_param, rpc.max_recv_buffer_size),
-	CONF_ITEM_UI32("RPC_Ioq_ThrdMax", 1, 1024*128, 200,
+	CONF_ITEM_UI32("rpc_ioq_thrdmin", 2, 1024*128, 2,
+		       nfs_core_param, rpc.ioq_thrd_min),
+	CONF_ITEM_UI32("RPC_Ioq_ThrdMax", 2, 1024*128, 200,
 		       nfs_core_param, rpc.ioq_thrd_max),
 	CONF_ITEM_UI32("RPC_GSS_Npart", 1, 1021, 13,
 		       nfs_core_param, rpc.gss.ctx_hash_partitions),
@@ -148,18 +200,26 @@ static struct config_item core_params[] = {
 		       nfs_core_param, rpc.gss.max_gc),
 	CONF_ITEM_I64("Blocked_Lock_Poller_Interval", 0, 180, 10,
 		      nfs_core_param, blocked_lock_poller_interval),
-	CONF_ITEM_LIST("NFS_Protocols", CORE_OPTION_ALL_VERS, protocols,
+	CONF_ITEM_LIST("NFS_Protocols", DEFAULT_PROTOCOLS, protocols,
 		       nfs_core_param, core_options),
-	CONF_ITEM_LIST("Protocols", CORE_OPTION_ALL_VERS, protocols,
+	CONF_ITEM_LIST("Protocols", DEFAULT_PROTOCOLS, protocols,
 		       nfs_core_param, core_options),
-	CONF_ITEM_BOOL("NSM_Use_Caller_Name", false,
-		       nfs_core_param, nsm_use_caller_name),
 	CONF_ITEM_BOOL("Clustered", true,
 		       nfs_core_param, clustered),
+#ifdef _USE_NLM
 	CONF_ITEM_BOOL("Enable_NLM", true,
 		       nfs_core_param, enable_NLM),
+	CONF_ITEM_BOOL("NSM_Use_Caller_Name", false,
+		       nfs_core_param, nsm_use_caller_name),
+#endif
+#ifdef _USE_RQUOTA
 	CONF_ITEM_BOOL("Enable_RQUOTA", true,
 		       nfs_core_param, enable_RQUOTA),
+#endif
+#ifdef USE_NFSACL3
+	CONF_ITEM_BOOL("Enable_NFSACL", false,
+		       nfs_core_param, enable_NFSACL),
+#endif
 	CONF_ITEM_BOOL("Enable_TCP_keepalive", true,
 		       nfs_core_param, enable_tcp_keepalive),
 	CONF_ITEM_UI32("TCP_KEEPCNT", 0, 255, 0,
@@ -174,8 +234,10 @@ static struct config_item core_params[] = {
 		       nfs_core_param, enable_FASTSTATS),
 	CONF_ITEM_BOOL("Enable_FSAL_Stats", false,
 		       nfs_core_param, enable_FSALSTATS),
+#ifdef _USE_NFS3
 	CONF_ITEM_BOOL("Enable_FULLV3_Stats", false,
 		       nfs_core_param, enable_FULLV3STATS),
+#endif
 	CONF_ITEM_BOOL("Enable_FULLV4_Stats", false,
 		       nfs_core_param, enable_FULLV4STATS),
 	CONF_ITEM_BOOL("Enable_AUTH_Stats", false,
@@ -194,12 +256,18 @@ static struct config_item core_params[] = {
 		       nfs_core_param, fsid_device),
 	CONF_ITEM_BOOL("mount_path_pseudo", false,
 		       nfs_core_param, mount_path_pseudo),
-	CONF_ITEM_BOOL("Enable_UDP", true,
-		       nfs_core_param, enable_UDP),
+	CONF_ITEM_ENUM_BITS("Enable_UDP", UDP_LISTENER_ALL, UDP_LISTENER_MASK,
+		       udp_listener_type, nfs_core_param, enable_UDP),
 	CONF_ITEM_STR("Dbus_Name_Prefix", 1, 255, NULL,
 		       nfs_core_param, dbus_name_prefix),
 	CONF_ITEM_UI32("Max_Uid_To_Group_Reqs", 0, INT32_MAX, 0,
 		       nfs_core_param, max_uid_to_grp_reqs),
+	CONF_ITEM_BOOL("Enable_V3fh_Validation_For_V4", false,
+		       nfs_core_param, enable_v3_fh_for_v4),
+	CONF_ITEM_UI32("Readdir_Res_Size", 4096, 32768, 0,
+		       nfs_core_param, readdir_res_size),
+	CONF_ITEM_BOOL("Getattrs_In_Complete_Read", true,
+			   nfs_core_param, getattrs_in_complete_read),
 	CONFIG_EOL
 };
 
@@ -270,10 +338,12 @@ static struct config_item_list recovery_backend_types[] = {
 static struct config_item version4_params[] = {
 	CONF_ITEM_BOOL("Graceless", false,
 		       nfs_version4_parameter, graceless),
-	CONF_ITEM_UI32("Lease_Lifetime", 0, 120, LEASE_LIFETIME_DEFAULT,
+	CONF_ITEM_UI32("Lease_Lifetime", 0, 180, LEASE_LIFETIME_DEFAULT,
 		       nfs_version4_parameter, lease_lifetime),
-	CONF_ITEM_UI32("Grace_Period", 0, 180, GRACE_PERIOD_DEFAULT,
+	CONF_ITEM_UI32("Grace_Period", 0, 270, GRACE_PERIOD_DEFAULT,
 		       nfs_version4_parameter, grace_period),
+	CONF_ITEM_STR("Server_Scope", 1, MAXNAMLEN, NULL,
+		      nfs_version4_parameter, server_scope),
 	CONF_ITEM_STR("DomainName", 1, MAXPATHLEN, DOMAINNAME_DEFAULT,
 		      nfs_version4_parameter, domainname),
 	CONF_ITEM_PATH("IdmapConf", 1, MAXPATHLEN, IDMAPCONF_DEFAULT,
@@ -296,12 +366,20 @@ static struct config_item version4_params[] = {
 	CONF_ITEM_TOKEN("RecoveryBackend", RECOVERY_BACKEND_DEFAULT,
 			recovery_backend_types, nfs_version4_parameter,
 			recovery_backend),
+	CONF_ITEM_PATH("RecoveryRoot", 1, MAXPATHLEN, NFS_V4_RECOV_ROOT,
+		       nfs_version4_parameter, recov_root),
+	CONF_ITEM_PATH("RecoveryDir", 1, MAXNAMLEN, NFS_V4_RECOV_DIR,
+		       nfs_version4_parameter, recov_dir),
+	CONF_ITEM_PATH("RecoveryOldDir", 1, MAXNAMLEN, NFS_V4_OLD_DIR,
+		       nfs_version4_parameter, recov_old_dir),
 	CONF_ITEM_LIST("minor_versions", NFSV4_MINOR_VERSION_ALL,
 		       minor_versions, nfs_version4_parameter, minor_versions),
 	CONF_ITEM_UI32("slot_table_size", 1, 1024, NFS41_NB_SLOTS_DEF,
 		       nfs_version4_parameter, nb_slots),
 	CONF_ITEM_BOOL("Enforce_UTF8_Validation", false,
 		       nfs_version4_parameter, enforce_utf8_vld),
+	CONF_ITEM_UI32("Max_Client_Ids", 0, UINT32_MAX, 0,
+		       nfs_version4_parameter, max_client_ids),
 	CONFIG_EOL
 };
 
