@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-3.0-or-later
 /*
  * Copyright IBM Corporation, 2012
  *  Contributor: Frank Filz <ffilzlnx@mindspring.com>
@@ -77,7 +78,7 @@ char name[MAXSTR];
 char portstr[MAXSTR];
 int port;
 char line[MAXXFER];
-long int alarmtag;
+long alarmtag;
 Inode *inodes[MAXFPOS + 1];
 Fh *filehandles[MAXFPOS + 1];
 enum lock_mode lock_mode[MAXFPOS + 1];
@@ -282,12 +283,12 @@ void do_open(struct response *resp)
 		fprintf_stderr("path = '%s' name = '%s'\n", path, name);
 
 		rc = ceph_ll_walk(cmount, path, &parent, &stx, 0,
-				  AT_NO_ATTR_SYNC, cephperms);
+				  AT_STATX_DONT_SYNC, cephperms);
 
 		if (rc >= 0) {
 			rc = ceph_ll_create(cmount, parent, name, resp->r_mode,
 					    resp->r_flags, &inode, &filehandle,
-					    &stx, 0, AT_NO_ATTR_SYNC,
+					    &stx, 0, AT_STATX_DONT_SYNC,
 					    cephperms);
 
 			/* Release the parent directory. */
@@ -300,7 +301,7 @@ void do_open(struct response *resp)
 		}
 	} else {
 		rc = ceph_ll_walk(cmount, resp->r_data, &inode, &stx, 0,
-				  AT_NO_ATTR_SYNC, cephperms);
+				  AT_STATX_DONT_SYNC, cephperms);
 
 		if (rc >= 0) {
 			rc = ceph_ll_open(cmount, inode, resp->r_flags,
@@ -329,7 +330,7 @@ void do_open(struct response *resp)
 
 void do_write(struct response *resp)
 {
-	long long int rc;
+	long long rc;
 
 	if (resp->r_fpos != 0 && filehandles[resp->r_fpos] == NULL) {
 		resp->r_status = STATUS_ERRNO;
@@ -363,7 +364,7 @@ void do_write(struct response *resp)
 
 void do_read(struct response *resp)
 {
-	long long int rc;
+	long long rc;
 
 	if (resp->r_fpos != 0 && filehandles[resp->r_fpos] == NULL) {
 		resp->r_status = STATUS_ERRNO;
@@ -452,7 +453,7 @@ void cancel_work_item(struct work_item *work)
 	free_work(work);
 }
 
-static inline long long int lock_end(struct response *req)
+static inline long long lock_end(struct response *req)
 {
 	if (req->r_length == 0)
 		return LLONG_MAX;
@@ -893,8 +894,8 @@ void do_close(struct response *resp)
 
 struct test_list {
 	struct test_list *tl_next;
-	long long int tl_start;
-	long long int tl_end;
+	long long tl_start;
+	long long tl_end;
 };
 
 struct test_list *tl_head;
@@ -915,7 +916,7 @@ void remove_test_list_head(void)
 	free(item);
 }
 
-void make_test_item(long long int start, long long int end)
+void make_test_item(long long start, long long end)
 {
 	struct test_list *item = malloc(sizeof(*item));
 
@@ -935,9 +936,9 @@ void make_test_item(long long int start, long long int end)
 	tl_tail = item;
 }
 
-int list_locks(long long int start, long long int end, struct response *resp)
+int list_locks(long long start, long long end, struct response *resp)
 {
-	long long int conf_end;
+	long long conf_end;
 	struct flock lock;
 	int rc;
 
@@ -989,8 +990,8 @@ int list_locks(long long int start, long long int end, struct response *resp)
 
 void do_list(struct response *resp)
 {
-	long long int start = resp->r_start;
-	long long int length = resp->r_length;
+	long long start = resp->r_start;
+	long long length = resp->r_length;
 	int conflict = false;
 
 	if (resp->r_fpos != 0 && filehandles[resp->r_fpos] == NULL) {
